Library IEEE;
USE IEEE.std_logic_1164.all;
USE IEEE.std_logic_arith.all;
USE ieee.std_logic_unsigned.all;
USE work.pkg_081006.all; 

-------------------------------------------------------------------- 
-- Description:
-- This module implements a dual (ping-ponged) shift register for
-- receiving serial data from the AD7625, 16 Bit Analog to Digital
-- converter (ADC).  One shift register accepts the serial stream at the
-- same moment the other shift register is holding the last transmitted
-- value for the next stage of data processing.  Once a shift register
-- has filled, the rolls of the two registers switch.
--
-- The shift register operation is controlled with a 5-bit counter and
-- a 5-state state machine.  The counter's lower 4-bits count the 16
-- serial data bits shifted into the registers.  The 5th bit controls
-- which of the two shift registers is receiving the serial stream and
-- which is tied to the parallel data output of the module.
--
-- The state machine watches for the transition between registers to 
-- time when the adc_data_strb is pulsed.  This is a strobe to be used
-- by the next logic block in the data processing sequence to latch
-- the new data.
--


entity AD7625_SHIFT_REG_v3 is
	port(
		 clk           : in std_logic; -- global Logic Clock
		 reset_n       : in std_logic; -- global Logic Reset
		 adc_srdata    : in std_logic; -- AD7625 signal Data 
		 adc_srclk     : in std_logic; -- AD7625 signal DCO  
		 
		 adc_data      : out std_logic_vector (15 downto 0);
		 adc_data_strb : out std_logic -- parallel data and strobe out
		);
end AD7625_SHIFT_REG_v3;
	
architecture main of AD7625_SHIFT_REG_v3 is
	-- Shift Register declarations -----------------------------------
	signal sr1         : std_logic_vector(15 downto 0); --
	signal sr2         : std_logic_vector(15 downto 0);
	signal ena1        : std_logic;
	signal ena2        : std_logic;
	signal strobe      : std_logic;
	
	signal ena1_syncd	: std_logic;
	signal ena2_syncd	: std_logic;
	
	-- Counter/Timer declarations ------------------------------------
	signal srcount     : std_logic_vector (4 downto 0);
	
	-- Gap Detection --------------------------------------------------
	signal count	: std_logic_vector(3 downto 0);
	signal count2	: std_logic_vector(3 downto 0);


	
--===begin===begin===begin===begin===begin===begin===begin===begin===
begin -- main --------------------------------------------------------

-- Shift Registers ---------------------------------------------------
    shift_reg1: process (adc_srclk, reset_n, ena1)
      begin
		if(reset_n = '0') then 
			sr1 <= (others => '0');
		elsif rising_edge(adc_srclk) and ena1 = '1' then 
			for i in 0 to 14 loop
				sr1(i+1) <= sr1(i);
			end loop;
			sr1(0) <= adc_srdata;
        end if;
    end process;
    
    shift_reg2: process (adc_srclk, reset_n, ena2)
      begin
		if(reset_n = '0') then 
			sr2 <= (others => '0');
		elsif rising_edge(adc_srclk) and ena2 = '1' then 
			for i in 0 to 14 loop
				sr2(i+1) <= sr2(i);
			end loop;
			sr2(0) <= adc_srdata;
        end if;
    end process;
------------------------------------------------------------------------
-- Counter -------------------------------------------------------------
	srcnt: process (adc_srclk, reset_n) -- bit counter for shift register
	begin
		if(reset_n = '0')           then srcount <= (others => '0');
		elsif rising_edge(adc_srclk) then srcount <= srcount + 1;
		end if;
	end process;
				 
	ena1 <= not srcount(4); -- This is the 5th bit of the counter
	ena2 <= srcount(4);

-- Synchronize ena signals ---------------------------------------------
	sync_ena: process(clk, reset_n, ena1, ena2)
	begin
		if   (reset_n = '0') then
			ena1_syncd <= '0';
			ena2_syncd <= '0';
		elsif rising_edge(clk) then
			ena1_syncd <= ena1;
			ena2_syncd <= ena2;
		end if;
	end process;
			             
------------------------------------------------------------------------
-- Gap Detector --------------------------------------------------------	
process (clk, reset_n, adc_srclk)
begin
	if   (reset_n = '0')	then count <= (others => '0');
	elsif(adc_srclk = '1') 	then count <= (others => '0');
	elsif rising_edge(clk)	then count <= count + 1;
	end if;
end process;

process(count)
begin
	if (count = 1) then strobe <= '1';
	else				strobe <= '0';
	end if;
end process;

-- Syncronized Outputs ---------------------------------------------------
	sr_sm_outputs: process (clk)
	begin
		if rising_edge(clk) then
			adc_data_strb <= strobe;
			if (ena1_syncd = '1') then 	adc_data <= sr2;
			else                 		adc_data <= sr1;
			end if;
		end if;		
	end process; -- sclk_sm_outputs	
--------------------------------------------------------------------------


end main;

