Library IEEE;
USE IEEE.std_logic_1164.all;
USE IEEE.std_logic_unsigned.all;

-- My Definitions --------------------------------
USE work.pkg_081006.all; 
USE work.DDS_VME_CONSTANTS.all;
-------------------------------------------------------------------------------
-- MODULE DESCRIPTION, May 13, 2011
--
-- This routine takes the data in the Injection Curve Memory along with
-- new values of N_Inj_Pts and Inj_Freq and computes new injection
-- frequency data that is written to the Frequency Curve Memory.

--	The Injection Curve Memory 1024 points are written simultaneously
-- with the writting of the first 1024 points of the Frequency Curve
-- Memory.  The computation or re-computation of the first N Frequency
-- Curve Memory points lags behind the writing of the new Frequency Curve.
-- The lag exists because of the wait cycles for the floating point math.

entity Injection_Curve_Managment is 
port( 
		clk					: in std_logic;
		nreset				: in std_logic;
		Upper_Addr_To_Mem : in std_logic_vector(23 downto 0);
		Upper_WR_FC       : in std_logic;
		
		Lower_Data_From_Mem  : in std_logic_vector(31 downto 0);
		Inj_Freq			 		: in std_logic_vector(31 downto 0);
		Updated_Inj_Freq	 	: in std_logic;
		N_Inj_Pts			 	: in std_logic_vector(31 downto 0);
		Updated_N_Inj_Pts	 	: in std_logic;
		Lower_Bus_Permit	 	: in std_logic;
		
		Lower_Addr_to_Mem 	: out std_logic_vector(23 downto 0);
		Lower_Data_to_Mem	 	: out std_logic_vector(31 downto 0);
		Lower_Mem_WR_Pulse	: out std_logic;
		Lower_Bus_Request	 	: out std_logic;
		
		pretrigA, pretrigB,t1,t2,t3 : out std_logic
		
	);
end Injection_Curve_Managment;
	
architecture main of Injection_Curve_Managment is	
signal sync_ena_inj		: std_logic;

signal ADDR 	 			: std_logic_vector(23 downto 0);
signal trig_compute_1	: std_logic;
signal trig_compute_2	: std_logic;
signal trig_compute_3	: std_logic;
signal trig_compute_sum : std_logic;

signal sync_trig_A		: std_logic_vector(1 downto 0);
signal pre_trig_A			: std_logic;

signal sync_trig_B		: std_logic_vector(1 downto 0);
signal pre_trig_B			: std_logic;


signal inj_freq_reg		: std_logic_vector(31 downto 0);
signal n_inj_points		: std_logic_vector(31 downto 0);

signal count_points		: std_logic_vector(11 downto 0);
signal reset_count		: std_logic;

-- Variable Registers -------------------------------------------------
signal inj_data_i			: std_logic_vector(31 downto 0);
signal get_data_in		: std_logic;

signal inj_data_0_reg	: std_logic_vector(31 downto 0);
signal get_inj_0			: std_logic;

signal freq_step_reg		: std_logic_vector(31 downto 0);
signal get_freq_step		: std_logic;

signal delta_reg			: std_logic_vector(31 downto 0);
signal delta_0				: std_logic_vector(31 downto 0);
signal get_delta_0		: std_logic;

signal numerator			: std_logic_vector(31 downto 0);
signal denominator		: std_logic_vector(31 downto 0);

signal divide_result		: std_logic_vector(31 downto 0);

signal flt2int_result	: std_logic_vector(31 downto 0);

signal delta_i				: std_logic_vector(31 downto 0);
signal get_delta_i		: std_logic;

signal freq_curve_i		: std_logic_vector(31 downto 0);

signal freq_curve_ptr	: std_logic_vector(23 downto 0);
signal inj_curve_ptr		: std_logic_vector(23 downto 0);
signal inj_curve_flg		: std_logic;

signal incr_pointers		: std_logic;
signal load_pointers		: std_logic;

-- Pause Timer ------------------------------------
	signal pause_tmr 		: std_logic_vector(4 downto 0);
	signal inc_pause_tmr	: std_logic;
  
--mmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmm
-- Altera Math Blocks
component float2intg
	PORT
	(
		clk_en	: IN STD_LOGIC ;
		clock		: IN STD_LOGIC ;
		dataa		: IN STD_LOGIC_VECTOR (31 DOWNTO 0);
		nan		: OUT STD_LOGIC ;
		overflow	: OUT STD_LOGIC ;
		result	: OUT STD_LOGIC_VECTOR (31 DOWNTO 0);
		underflow: OUT STD_LOGIC 
	);
end component;

component intg2float
	PORT
	(
		clk_en	: IN STD_LOGIC ;
		clock		: IN STD_LOGIC ;
		dataa		: IN STD_LOGIC_VECTOR (31 DOWNTO 0);
		result	: OUT STD_LOGIC_VECTOR (31 DOWNTO 0)
	);
end component;

component sample_divider
	PORT
	(
		clk_en	: IN STD_LOGIC ;
		clock		: IN STD_LOGIC ;
		dataa		: IN STD_LOGIC_VECTOR (31 DOWNTO 0);
		datab		: IN STD_LOGIC_VECTOR (31 DOWNTO 0);
		nan		: OUT STD_LOGIC ;
		overflow	: OUT STD_LOGIC ;
		result	: OUT STD_LOGIC_VECTOR (31 DOWNTO 0);
		underflow: OUT STD_LOGIC ;
		zero		: OUT STD_LOGIC 
	);
end component;

component add_sub32
	PORT
	(
		add_sub	: IN STD_LOGIC ;
		dataa		: IN STD_LOGIC_VECTOR (31 DOWNTO 0);
		datab		: IN STD_LOGIC_VECTOR (31 DOWNTO 0);
		result	: OUT STD_LOGIC_VECTOR (31 DOWNTO 0)
	);
end component;

-- SETUP_STATE_MACHINE ---------------------------------------
type STATE_TYPE is
(init, wait_for_A, wait_for_B, wait_for_curve, idle, get_lower_bus_permit, 
 read_first_curve_point, get_first_curve_point, compute_freq_step, 
 latch_freq_step, compute_new_point, reset_tmr_1, write_new_point, 
 inc_addr_pointers, get_next_delta, read_next_curve_point, done);
signal CS, NS : STATE_TYPE;

--=====================================================================
--				B  E  G  I  N 
--=====================================================================
begin

ADDR <= Upper_Addr_To_Mem;

-----------------------------------------------------------------------
DETECT_INJECTION_FREQ_PARAM_WRITE:
process(clk, nreset, Updated_Inj_Freq, sync_trig_A)	
begin
	if   (nreset = '0') then sync_trig_A <= (others => '0');
	elsif rising_edge(clk) then	
		sync_trig_A(0) <= Updated_Inj_Freq;
		sync_trig_A(1) <= sync_trig_A(0);
	end if;
end process;
		pre_trig_A <= sync_trig_A(0) AND NOT sync_trig_A(1);

-----------------------------------------------------------------------
DETECT_N_INJECTION_POINTS_PARAM_WRITE:
process(clk, nreset, Updated_N_Inj_Pts, sync_trig_B)	
begin
	if   (nreset = '0') then sync_trig_B <= (others => '0');
	elsif rising_edge(clk) then	
		sync_trig_B(0) <= Updated_N_Inj_Pts;
		sync_trig_B(1) <= sync_trig_B(0);
	end if;
end process;
		pre_trig_B <= sync_trig_B(0) AND NOT sync_trig_B(1);

-----------------------------------------------------------------------
-- pre-trigger signals are only used at power up or reset --
pretrigA <= pre_trig_A;
pretrigB <= pre_trig_B;

-----------------------------------------------------------------------
-----------------------------------------------------------------------
DETECT_FREQ_CURVE_WRITE:
process(clk, nreset, ADDR, Upper_WR_FC)	
begin
	if   (nreset = '0') then trig_compute_1 <= '0';
	elsif rising_edge(clk) then	
		if   (ADDR = FREQUENCY_CURVE1) then	trig_compute_1 <= Upper_WR_FC;
		else											trig_compute_1 <= '0';
		end if;
	end if;
end process;


t1 <= trig_compute_1; -- Frequency Curve Changed (was written to)
t2 <= trig_compute_2; -- Number of Injection Points Changed
t3 <= trig_compute_3; -- Injection Frequency Changed

-----------------------------------------------------------------------
PARAMETER_REGISTERS:
process(clk, nreset, Inj_Freq, N_Inj_Pts)
begin
	if   (nreset = '0') then 	
				inj_freq_reg <= (others => '0');
				n_inj_points <= (others => '0');
	elsif rising_edge(clk) then	
				inj_freq_reg <= Inj_Freq;
				n_inj_points <= N_Inj_Pts;
	end if;
end process;

-----------------------------------------------------------------------
DETECT_INJ_FREQ_CHANGE:
process(clk, nreset, Inj_Freq, inj_freq_reg)
begin
	if   (nreset = '0') then trig_compute_2 <= '0';
	elsif rising_edge(clk) then	
		if   (Inj_Freq = inj_freq_reg) then trig_compute_2 <= '0';
		else											trig_compute_2 <= '1';
		end if;
	end if;
end process;

-----------------------------------------------------------------------
DETECT_N_POINTS_CHANGE:
process(clk, nreset, N_Inj_Pts, n_inj_points)
begin
	if   (nreset = '0') then trig_compute_3 <= '0';
	elsif rising_edge(clk) then	
		if   (N_Inj_Pts = n_inj_points) then trig_compute_3 <= '0';
		else								 			 trig_compute_3 <= '1';
		end if;
	end if;
end process;

trig_compute_sum <= trig_compute_1 OR trig_compute_2 OR trig_compute_3;

-----------------------------------------------------------------------
INJ_DATA_IN_BUFFER:
process(clk, nreset, get_data_in, Lower_Data_From_Mem)
begin
	if   (nreset = '0') then 	
				inj_data_i <= (others => '0');
	elsif rising_edge(clk) then	
		if   (get_data_in = '1') then inj_data_i <= Lower_Data_From_Mem;
		end if;
	end if;
end process;

-----------------------------------------------------------------------
INJ_DATA_0_REGISTER:
process(clk, nreset, get_inj_0,	inj_data_i)
begin
	if   (nreset = '0') 	then 	inj_data_0_reg 	<= (others => '0');
	elsif rising_edge(clk) 	then	
		if   (get_inj_0     = '1') then inj_data_0_reg 	<= inj_data_i;
		end if;
	end if;
end process;

-----------------------------------------------------------------------
--mmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmm
-- Computation of Freq Step and Delta[0]

--COMPUTE_DELTA_0--
SUB1:
add_sub32 port map(	add_sub 	=> '0', -- subtraction
							dataa		=>	inj_freq_reg,
							datab		=>	inj_data_0_reg,
							result	=> delta_0 );

-----------------------------------------------------------------------
--COMPUTE_FREQ_STEP--
I2F_1:
intg2float port map(	clk_en	=> '1',
							clock		=> clk,
							dataa		=> delta_0,
							result	=> numerator );

I2F_2:
intg2float port map(	clk_en	=> '1',
							clock		=> clk,
							dataa		=> N_Inj_Pts,
							result	=> denominator );

DIVIDER:
sample_divider port map(clk_en	=> '1',
								clock		=> clk,
								dataa		=> numerator,
								datab		=> denominator,
								nan		=> open,
								overflow	=> open,
								underflow=> open,
								zero		=> open,
								result	=> divide_result );

F2I:
float2intg port map(	clk_en	=> '1',
							clock		=> clk,
							dataa		=> divide_result,
							nan		=> open,
							overflow => open,
							result	=> flt2int_result );
					
-----------------------------------------------------------------------
FREQ_STEP_REGISTER:
process(clk, nreset, get_freq_step, flt2int_result)
begin
	if   (nreset = '0') 	then 	freq_step_reg 	<= (others => '0');
	elsif rising_edge(clk) 	then	
		if   (get_freq_step = '1') then freq_step_reg 	<= flt2int_result;
		end if;
	end if;
end process;

-----------------------------------------------------------------------
DELTA_REGISTER:
process(clk, nreset, get_delta_0, get_delta_i, delta_0, delta_i)
begin
	if   (nreset = '0') then 	
				delta_reg 		<= (others => '0');
				
	elsif rising_edge(clk) then	
		if   (get_delta_0   = '1') then delta_reg 	<= delta_0;
		elsif(get_delta_i   = '1') then delta_reg		<= delta_i;
		end if;
	end if;
end process;

-----------------------------------------------------------------------
--COMPUTE_NEW_FREQ_CURVE--
SUB2:
add_sub32 port map(	add_sub 	=> '0', -- subtraction
							dataa		=>	delta_reg,
							datab		=>	freq_step_reg,
							result	=> delta_i );

SUB3:
add_sub32 port map(	add_sub 	=> '1', -- addition
							dataa		=>	inj_data_i,
							datab		=>	delta_reg,
							result	=> freq_curve_i );
							
--mmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmmm
-----------------------------------------------------------------------------
PAUSE_TIMER: -- Timer used to generate setup time for FLASH writes
process (clk, inc_pause_tmr)
begin
	if rising_edge(clk) then
		if   (inc_pause_tmr = '0') then pause_tmr <= "00000";
		elsif(inc_pause_tmr = '1') then pause_tmr <= pause_tmr + 1;
		end if;
	end if;
end process;

-----------------------------------------------------------------------
-- ADDR_POINTER_COUNTER 1-------------------------------------------
SOURCE_ADDR_POINTER: countby_n -- Address Pointer to data transfer source
generic map(24,4)
port map(	clk  	 => clk,
			clrn 	 => nreset,
			ena  	 => incr_pointers,
			load 	 => load_pointers,
			step	 => "0100",
			ld_data  => INJ_CURVE_FIRST,
			flag_cnt => INJ_CURVE_LAST,
			flag_out => inj_curve_flg, 
			cnt_out  => inj_curve_ptr 
		);

-- ADDR_POINTER_COUNTER 2-------------------------------------------
DESTINATION_ADDR_POINTER: countby_n -- Address Pointer to data transfer source
generic map(24,4)
port map(	clk  	 => clk,
			clrn 	 => nreset,
			ena  	 => incr_pointers,
			load 	 => load_pointers,
			step	 => "0100",
			ld_data  => FREQUENCY_CURVE1_FIRST,
			flag_cnt => FREQUENCY_CURVE1_LAST,
			flag_out => open,
			cnt_out  =>	freq_curve_ptr 
		);

-- Count Number of Injection Points Computed --------------------------
POINT_COUNT:
process(clk, nreset, reset_count, incr_pointers)
begin
	if  (nreset = '0') then count_points <= (others => '0');
	elsif rising_edge(clk) then
		if   (reset_count   = '1') then count_points <= (others => '0');
		elsif(incr_pointers = '1') then count_points <= count_points + 1;
		end if;
	end if;
end process;

--{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{
--                S T A T E   M A C H I N E
--}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}
SM_REG: process(clk, nreset)
begin
	if(nreset='0') then CS <= init;
	elsif(rising_edge(clk))	then CS <= NS;
	end if;
end process;

SM_NEXT_STATE: 
process
(CS, pre_trig_A, pre_trig_B, trig_compute_sum, trig_compute_1, pause_tmr, 
 inj_curve_flg, inj_curve_ptr, freq_curve_i, freq_curve_ptr, count_points, 
 n_inj_points, inj_data_i, Lower_Bus_Permit )
begin
	-- defaults ----------------------
	Lower_Addr_to_Mem	<= inj_curve_ptr;
	Lower_Data_to_Mem	<= freq_curve_i;
	Lower_Mem_WR_Pulse	<= '0';
	Lower_Bus_Request	<= '1'; -- request active is the default ...
	
	get_data_in 		<= '0';
	get_inj_0			<= '0';
	get_freq_step		<= '0';
	get_delta_0			<= '0';
	get_delta_i   		<= '0';
	
	inc_pause_tmr 		<= '0';
	
	load_pointers		<= '0';
	incr_pointers		<= '0';
	
	reset_count			<= '0';
	-----------------------------------
	case CS is
		when init =>							
			Lower_Bus_Request	<= '0';
			-- Ensure parameters are setup after a reset or power up
			-- Once this has run once we return to and wait in "idle".
			if   (pre_trig_A = '1') then		NS <= wait_for_B; -- if A get B
			elsif(pre_trig_B = '1') then		NS <= wait_for_A; -- if B get A
			else								NS <= init;
			end if;
				
		when wait_for_A =>
			Lower_Bus_Request	<= '0';
			if   (pre_trig_A = '1') then		NS <= wait_for_curve;
			else										NS <= wait_for_A;
			end if;	
		
		when wait_for_B =>
			Lower_Bus_Request	<= '0';
			if   (pre_trig_B = '1') then		NS <= wait_for_curve;
			else										NS <= wait_for_B;
			end if;	
		
		----------------------------------------------------------------------	
		when wait_for_curve =>	-- wait for curve to be updated
			Lower_Bus_Request	<= '0';
			load_pointers		<= '1';
			reset_count			<= '1';
			
			if (trig_compute_1 = '1') then	NS <= get_lower_bus_permit;
			else										NS <= wait_for_curve;
			end if;	
					
		when idle =>	 -- we wait here after the first time through       
			Lower_Bus_Request	<= '0';
			load_pointers		<= '1';
			reset_count			<= '1';
			
			if (trig_compute_sum = '1') then	NS <= get_lower_bus_permit;
			else										NS <= idle;
			end if;	
			
		----------------------------------------------------------------------
		when get_lower_bus_permit =>
			Lower_Bus_Request	<= '1'; -- default value
			if (Lower_Bus_Permit = '0') then NS <= get_lower_bus_permit; --wait
			else										NS <= read_first_curve_point;
			end if;

		when read_first_curve_point =>			
			Lower_Addr_to_Mem	<= inj_curve_ptr;
			get_data_in			<= '1';
			
			-- increment the pause timer and wait for timeout
			inc_pause_tmr	<= '1'; 			
			if (pause_tmr = "00011") then		NS <= get_first_curve_point;
			else										NS <= read_first_curve_point;
			end if;
		
		when get_first_curve_point =>			NS <= compute_freq_step;
			Lower_Addr_to_Mem	<= inj_curve_ptr; -- held over
			get_inj_0			<= '1';

		when compute_freq_step =>
			-- pause for 28 clocks while computation is completed.
			inc_pause_tmr	<= '1'; 			
			if (pause_tmr = "11100") then		NS <= latch_freq_step;
			else										NS <= compute_freq_step;
			end if;
			
		when latch_freq_step =>					NS <= compute_new_point;
			get_freq_step		<= '1';
			get_delta_0			<= '1';
			
		------------------------------------------------- begin loop --	
		when compute_new_point =>
			-- pause for 3 clocks while computation is completed.
			inc_pause_tmr	<= '1'; 			
			if (pause_tmr = "00011") then		NS <= reset_tmr_1;
			else										NS <= compute_new_point;
			end if;
			
		when reset_tmr_1 =>						NS <= write_new_point;
			
		when write_new_point =>
			Lower_Addr_to_Mem	<= freq_curve_ptr;
			Lower_Mem_WR_Pulse	<= '1';
			
			if (count_points < n_inj_points) then  -- done computing --
				Lower_Data_to_Mem	<= freq_curve_i;
			else
				Lower_Data_to_Mem	<= inj_data_i;
			end if;

			-- increment the pause timer and wait for timeout
			inc_pause_tmr	<= '1'; 			
			if (pause_tmr = "00011") then		NS <= inc_addr_pointers;
			else										NS <= write_new_point;
			end if;
		
		when inc_addr_pointers =>				NS <= get_next_delta;
			incr_pointers	<= '1';
		
		when get_next_delta	=>					
			get_delta_i		<= '1';
			
			if (inj_curve_flg = '1') then 	NS <= done; 
			else										NS <= read_next_curve_point;
			end if;
		
		when read_next_curve_point =>			
			Lower_Addr_to_Mem	<= inj_curve_ptr;
			get_data_in			<= '1';
			
			-- increment the pause timer and wait for timeout
			inc_pause_tmr	<= '1'; 			
			if (pause_tmr = "00011") then		NS <= compute_new_point;
			else								NS <= read_next_curve_point;
			end if;
		
		-----------------------------------------------
		when done =>						NS <= idle;
			
		when others => 						NS <= idle;
	end case;
end process;

end;