Library IEEE;
USE IEEE.std_logic_1164.all;
USE IEEE.std_logic_unsigned.all;

-- My Definitions --------------------------------
USE work.pkg_081006.all; 
USE work.DDS_VME_CONSTANTS.all;

----------------------------------------------------------------
-- Placement of these 32 bit registers is controlled by the 
-- constant VME_CTRL_LOWER_FIRST.  This constant is defined in
-- the included file DDS_VME_CONSTANTS.vhd.

entity Mem_to_DAC is 
	port( 	
			clk					: in std_logic;
			nreset				: in std_logic;
			Match_Addr_Lower	: in std_logic_vector(19 downto 0);
			Match_Addr_Upper	: in std_logic_vector(19 downto 0);
			Division_Exp		: in std_logic_vector(15 downto 0);
			Offset				: in std_logic_vector(31 downto 0);
			
			Mem_Data_In			: in std_logic_vector(31 downto 0);
			Mem_Address			: in std_logic_vector(19 downto 0);
			
			DAC_Data_Out		: out std_logic_vector(15 downto 0);
			DAC_Strobe			: out std_logic
			
		);
	end Mem_to_DAC;
	
architecture main of Mem_to_DAC is	

	signal addr_lower		: std_logic_vector(19 downto 0);
	signal addr_upper		: std_logic_vector(19 downto 0);
	signal div_exp			: std_logic_vector(15 downto 0);
	signal crv_offset		: std_logic_vector(31 downto 0);

	component add_sub32
		PORT
		(
			add_sub		: IN STD_LOGIC ;
			dataa		: IN STD_LOGIC_VECTOR (31 DOWNTO 0);
			datab		: IN STD_LOGIC_VECTOR (31 DOWNTO 0);
			result		: OUT STD_LOGIC_VECTOR (31 DOWNTO 0)
		);
	end component;


	signal address_prev		: std_logic_vector(19 downto 0);
	signal addr_match		: std_logic;
	
	signal get_new_address	: std_logic;
	
	signal data_buffer		: std_logic_vector(15 downto 0);
	signal get_new_data		: std_logic;
	
	signal subtraction_result 	: std_logic_vector(31 downto 0);
	
	signal sr	  				: std_logic_vector(31 downto 0);
	signal sr_load				: std_logic;
	signal sr_count				: std_logic_vector(15 downto 0);
	signal sr_count_ena			: std_logic;

	-- State Machine declarations ------------------------------------
	type STATE_TYPE is 
	(idle, pause1, check_addr, pause2, load_sr, divide, get_data, strobe_dac);
	signal CS, NS  : STATE_TYPE;

--======================================================================
--               B  E  G  I  N  
------------------------------------------------------------------------       
begin

DEFAULT_INPUTS:
process(Match_Addr_Lower, Match_Addr_Upper, Division_Exp, Offset)			
begin
	if   (Match_Addr_Lower = X"00000") then
		addr_lower	<= X"20000";
		addr_upper	<= X"29000"; -- 36864 us worth of data
		div_exp		<= X"000C";
--		crv_offset	<= X"13BBBBF2";
		crv_offset	<= X"042E4921";
	else
		addr_lower	<= Match_Addr_Lower; -- VME_Ctrl_4[31..0] := X"03F910"
		addr_upper	<= Match_Addr_Upper; -- VME_Ctrl_5[31..0] := X"03F914"
		div_exp		<= Division_Exp;     -- VME_Ctrl_6[31..0] := X"03F918"
		crv_offset	<= Offset;			   -- VME_Ctrl_7[31..0] := X"03F91C"
    end if;                
end process;


LOOK_FOR_NEW_ADDRESS:
process(clk, nreset, address_prev, Mem_Address)
begin
	if   (address_prev = Mem_Address) then get_new_address <= '0';
	else								   			get_new_address <= '1';
	end if;
end process;

ADDRESS_LATCH:
process(clk, nreset, get_new_address, Mem_Address)
begin
	if   (nreset = '0') then address_prev <= (others => '0');
	elsif rising_edge(clk) then
		if   (get_new_address = '1') then address_prev <= Mem_Address;
		end if;
	end if;
end process;
	
DATA_LATCH:
process(clk, nreset, get_new_data, sr)
begin
	if   (nreset = '0') then data_buffer <= (others => '0');
	elsif rising_edge(clk) then
		if   (get_new_data = '1') then data_buffer <= sr(15 downto 0);
		end if;
	end if;
end process;

--DAC_Data_Out <= data_buffer; -- DATA OUTPUT ----------
DAC_Data_Out <= not data_buffer(15) & data_buffer(14 downto 0); -- DATA OUTPUT ----------


SUBTRACT_OFFSET: 
	add_sub32 port map(	add_sub	=>	'0', --SUBTRACT
						dataa	=>	Mem_Data_In,
						datab	=> 	crv_offset,
						result	=> 	subtraction_result 
					  );
                      
SHIFT_REGISTER_COUNT:
process(clk, nreset, sr_count_ena, sr_count)
begin
	if   (nreset = '0') then sr_count <= X"0001";
	elsif rising_edge(clk) then
		if   (sr_count_ena = '1') then 	sr_count <= sr_count + 1;
		else										sr_count <= X"0001";
		end if;
	end if;

end process;


SHIFT_REGISTER:
process(clk, nreset, sr, sr_load, sr_count_ena)
begin
	if   (nreset = '0') then sr <= (others => '0');
	elsif rising_edge(clk) then
		if   (sr_load = '1')	  then		sr <= subtraction_result;
		elsif(sr_count_ena = '1') then 	sr <= '0' & sr(31 downto 1);
		else										sr <= sr;
		end if;
	end if;
end process;


-- State Machine define initialization ---------------------------
	sm_reg: process (clk, nreset)	
	begin
		if(nreset = '0')            then CS <= idle;
		elsif(clk'event and clk='1')then CS <= NS;
		end if;
	end process; -- sm_reg
	
	sm_combinatorial: 
	process 
	(CS, get_new_address, address_prev, addr_lower, addr_upper,
	 sr_count, div_exp)
	begin
		-- default signal values
		DAC_Strobe 		<= '0';	
		sr_load 		<= '0';
		sr_count_ena 	<= '0';	
		get_new_data	<= '0';		
		-------------------------
		case CS is
			when idle   =>
				if(get_new_address = '1') then 	NS <= pause1;
				else                      		NS <= idle;
				end if;
			    
			when pause1  =>						NS <= check_addr;
			    
			when check_addr  =>						
				if( address_prev >= addr_lower ) AND
				  ( address_prev <  addr_upper ) then NS <= pause2;
				else								  NS <= idle;
				end if;
			
			when pause2  =>						NS <= load_sr;
			    
			when load_sr =>						NS <= divide;
				sr_load 		<= '1';
				
			when divide =>
				sr_count_ena 	<= '1';
				
				if( sr_count = div_exp ) then 	NS <= get_data;
				else							NS <= divide;
				end if;
				
			when get_data  =>					NS <= strobe_dac;
				get_new_data	<= '1';		

			when strobe_dac  =>					NS <= idle;
				DAC_Strobe 		<= '1';
						
			when others  =>						NS <= idle;
		end case; 
				
	end process;

end;