Library IEEE;
USE IEEE.std_logic_1164.all;
USE IEEE.std_logic_unsigned.all;

-- My Definitions --------------------------------
USE work.pkg_081006.all; 

------------------------------------------------------
-- MODULE DESCRIPTION
-- This module implements the procedures which setup the DDS registers
-- to get the DDS channels up and running.  The general steps in the
-- setup are as follows (1/7/2010)
--		1. A 750us Master Reset is signaled for all channels.
--		2. A 750us pause is implemented after the reset.
--		3. The PLL control register, CFR3(addr = 0x02) is setup.
--		4. Another 750us pause is made to let the PLL settle.
--		5. The multi-chip sync register (addr = 0x0A) is setup for the 
--		   sync master DDS (channel 1).
--		6. The multi-chip sync register (addr = 0x0A) is setup for the 
--		   sync slave DDS's.
--		7. Two main bits are set in control register CFR1(addr = 0x00)
--		   These are bit 13, "Auto-clear phase accumulator" and bit 0
--		   which causes the DDS's to expect the subsequent serial loads
--		   to be performed LSB first.
--		8. The Single Tone Profile 0 register (addr = 0x0E) is setup with 
--			its initial Frequency word and phase word.  It is assumed that 
--		    the profile control pins of the DDS's is set to use Profile 0.
--		9. Step 8. is repeated twice more to ensure all channels' phase 
--			accumulators have sync'd.
--		10. Register CFR1 is written again to release the Auto-clear phase
--			accumulator bit causing the accumulators in the DDS's to be 
--			roughly syncronized.  Also the LSB First bit (bit 0) is cleared
--		    to return to MSB First serial IO mode.
--
-- A ROM memory is established that contains the setup words and setup 
-- control bits to accomplish this initial configuration of the DDS's.
-- This is in a *.mif file pointed to by the Altera Memory LPM Megafunction.
-- The memory initialization file is shown below.
--
--==============================================================
-- -- Quartus II generated Memory Initialization File (.mif)

--	WIDTH=48;
--	DEPTH=32;

--	ADDRESS_RADIX=BIN;
--	DATA_RADIX=HEX;
--	
--	CONTENT BEGIN
--		000000  :   D82000000000;
--		000001  :   9422183F4160;
--		000010  :   982000000000;
--		000011  :   802A0C000000;
--		000100  :   812A08000000;
--		000101  :   902000002001;
--		000110  :   302E05555555;
--		000111  :   302E05555555;
--		001000  :   102000000001;
--		[001001..011011]  :   000000000000;
--		011100  :   0000FFFFFFFC;
--		011101  :   0000FFFFFFFD;
--		011110  :   0000FFFFFFFE;
--		011111  :   0000FFFFFFFF;
--	END;
--==============================================================
-- 
-- CONTROL BITS (not defined by DDS datasheet)
-- The first 8 bits of the 48 bit setup words are control bits and are not
-- written to the DDS's.  These bits are defined below.
--
-- |   47   |   46   |    45     |      44         ||  43  |   42   | 41  | 40  |
-- |MSB_nLSB|Do_Reset|Ena_IORESET|Global_nAddressed||Pause |reserved| CS1 | CS0 |
--
-- Definitions:
-- MSB_nLSB :  Informs the FPGA Logic how to pack the Instruction Byte and data
--             into the serial output shift register for transmission.  A high
--			   	indicates the MSB First mode is to be used.  A low indicates the
--			   	LSB First mode is to be used.
-- Do_Reset :  This signals the FPGA logic that the DDS Master Reset is to be 
--			   	pulsed.
-- Ena_IORESET :  This bit causes the IORESET bit to be enabled during the 
--				  		transmission of this setup word.  The IORESET is a pulse
--				  		that is sent at the end the serial transmission to terminate
--				  		a partial DDS register write.  In the case of the DDS setup
--                logic we note that the DDS CFR configuration registers are 
--                32 bits long.  The Single Tone Profile registers in which we
--				  		can write the Amplitude, Phase, and Frequency words are 64 bits
--				  		long.  The trick used for the setup and later in writing the 
--				 		frequency and phase updates is to write only a portion of the 
--				  		profile register, starting from the LSB of the register.  For
--				  		more details see the datasheet page 49-51.
-- Global_nAddressed :  Indicates to the FPGA logic that either all of the DDS
--						channels should receive the following setup word or only
--						one of the DDS channels addressed by [CS1,CS0] should 
--						receive it.  When high all DDS's are written.
-- Pause :  When set the FPGA logic pauses for 150us before reading the next word
--			from the setup memory.  Note the bits in the current word following 
--			the Pause bit are not transmitted.
-- CS1,CS0 :  These bits are used to address a specific DDS channel for reception of
--			  the setup word when the Global_nAddressed bit is set low.
--
-- INSTRUCTION BYTE
-- The second 8 bits are the Instruction Byte as defined by the AD9910 DDS datasheet.
-- The Instruction Byte always precedes the data being written to or read from the DDS's.
--
-- |    39     |  38 |  37 | 36 | 35 | 34 | 33 | 32 |
-- |Read_nWrite|  x  |  v  | A4 | A3 | A2 | A1 | A0 |
--
-- Read_nWrite :  When high it indicates that the FPGA wishes to read from the 
--				  		addressed register.  When low it indicates that the FPGA is 
--				  		writing the register.  For this setup procedure this bit is
--				  		always low.
-- x : This bit is ignored (don't care).
-- v (valid) :  This bit is actually specified by the AD9910 datasheet as another don't 
--					care bit.  It is used to signal the FPGA logic that the current setup 
--					word is valid when set high.  When a setup memory word is encountered
--					with this bit low, the setup is considered finished.
--
-- A[4..0] :  These bits address the registers inside the DDS's.
--
--

------------------------------------------------------

entity AD9910_Setup_Procedure is 
	port( 
		SCLK				: in std_logic;
		nreset			: in std_logic;
		
		Setup_mem_addr	: in std_logic_vector(4 downto 0);
		Setup_mem_data	: in std_logic_vector(47 downto 0);		
		mem_wr		   : in std_logic;
				
		Trig_Setup		: in std_logic; -- strobe to signal AD9910 Setup Sequence
--		Setup_Sel		: in std_logic; -- Select Setup Group
		Transf_Done		: in std_logic;  -- Strobes when SR is done
		
		Enable_Setup  		: out std_logic; -- mux control bit
		Setup_Data	   	: out std_logic_vector(71 downto 0); -- data to transmit
		Setup_Strb	   	: out std_logic; -- strobe to latch data to transmit
		Setup_NBits	   	: out std_logic_vector(6 downto 0);
		Setup_MSB_nLSB 	: out std_logic; -- send MSB or LSB first indication	
		Setup_Ena_IOReset : out std_logic;
		
		CS1					: out std_logic;
		CS2					: out std_logic;
		CS3					: out std_logic;
		CS4					: out std_logic;
		Setup_DDS_Reset	: out std_logic;
		
		reserved    	: out std_logic;
		reserved2   	: out std_logic;
		DDS_Setup_Done	: out std_logic;
		
		bit0, bit1, bit2, bit3 : out std_logic
		);
	end AD9910_Setup_Procedure ;
	
--########################################################
--========================================================
--				A R C H I T E C T U R E
--========================================================
--########################################################
architecture main of AD9910_Setup_Procedure  is

-- Strobe Sync Signals ------------------------------------
	signal do_setup : std_logic; -- pulse to start setup sequence
	signal tx_done  : std_logic; -- last transmit or receive is complete
	
--	signal sel		: std_logic; -- latched setup selection

-- TRANSMIT_DATA Signals -----------------------------------
	signal xmt_q		: std_logic_vector(71 downto 0);
	signal xmt_q_rev	: std_logic_vector(71 downto 0);
	
	signal tx_strobe	: std_logic; -- transmit strobe signal
	signal dds_reset	: std_logic; -- local variable for Master Reset
	signal msb_first	: std_logic; 
					--	Setup_MSB_nLSB: says which way to shift serial data
	signal valid_bit    : std_logic; -- bit set in the instruction byte
	signal reset_bit    : std_logic;
	signal ioreset_bit	: std_logic;
	
-- ADDR POINTER COUNTER signals ----------------------------
	signal cnt_nrst		: std_logic; -- counter reset signal
	signal cnt_ena		: std_logic; -- counter increment enable strobe
	signal cnt_load		: std_logic; -- count load strobe
	signal addr_done	: std_logic; -- count end flag
	
	signal get_cnt		: std_logic; -- word count buffer enable strobe
	signal word_cnt		: std_logic_vector(2 downto 0); -- flag count
	signal instr_cnt	: std_logic_vector(2 downto 0); -- counter output
	signal setup_cnt_done	: std_logic; -- count end flag
	
	signal pause_bit    : std_logic;
	signal pause_ena    : std_logic;
	signal pause_flag	: std_logic_vector(15 downto 0);
--	signal pause_cnt	: std_logic_vector(13 downto 0);
	signal pause_done   : std_logic;
	
	signal setp_done    : std_logic;
	
	
------------------------------------------------------------
-- ALTERA COMPONENTS ---------------------------------------
-- The following component is the RAM memory
-- It was created using the LPM Megafunction Wizard.
-- The following VHDL file was generated and must be included 
-- in the project.
--                              DDS_Setup_Mem.vhd
--
-- The Megafunction Wizard also generate *.cmp files for each
-- from which I cut and pasted the component declarations below.
----------------------------------------------------------------
	component DDS_Setup_Mem
		PORT
		(
			address	: IN STD_LOGIC_VECTOR (4 DOWNTO 0);
			clock		: IN STD_LOGIC ;
			data		: IN STD_LOGIC_VECTOR (47 DOWNTO 0);
			wren		: IN STD_LOGIC ;
			q			: OUT STD_LOGIC_VECTOR (47 DOWNTO 0)
		);
	end component;

-- SETUP MEMORY Signals ----------------------------------------
	signal mem_ptr		: std_logic_vector(4 downto 0);
	signal mem_addr		: std_logic_vector(4 downto 0);
	signal mem_data_out : std_logic_vector(47 downto 0);
	
-- INTERMEDIATE SETUP OUTPUTS signals --------------------------	
	signal Global_nAddressed : std_logic;
	signal Chip_Sel1 		 : std_logic;
	signal Chip_Sel0 		 : std_logic;

-- SETUP_STATE_MACHINE ---------------------------------------
	type STATE_TYPE is
	(idle, set_mem_adr, ld_mem_adr, get_mem_data, ld_word_cnt, 
	 wait1, get_word, test_bits, strb_xmt, transmit, pause, 
	 inc_adr, done, done2);
	signal CS, NS : STATE_TYPE;
	
--########################################################
--========================================================
--						B E G I N
--========================================================
--########################################################

begin

--**-- --**--   --**--   --**--   --**--   --**--   --**-- 
-- STROBE SYNCHRONIZATION --------------------------------  
SYNC_SETUP: led -- syncronize the signal to start setup sequence.
port map(	sig_in  => Trig_Setup, 
			clk     => SCLK, 
			clrn    => '1', 
			pls_out => do_setup); 
			
SYNC_DONE: led -- syncronize the Transf_Done.
port map(	sig_in  => Transf_Done, 
			clk     => SCLK, 
			clrn    => '1', 
			pls_out => tx_done); 

--**--   --**--   --**--   --**--   --**--   --**--   --**--   
-- SETUP RAM MEMORY ----------------------------------------
SETUP_MEM: DDS_Setup_Mem -- RAM memory that holds setup words
port map(	address  => mem_addr,
			clock	 => SCLK,
			data	 => Setup_mem_data,
			wren	 => mem_wr,
			q		 => mem_data_out);        

--**-- --**--   --**--   --**--   --**--   --**--   --**-- 
--LATCH_SETUP_SEL: vdffe -- latch the setup selection bit
--port map(	d    => Setup_Sel,
--			clk  => SCLK,
--			clrn => nreset,
--			ena  => do_setup, --momentary latch signal
--			prn  => '1',
--			q    => sel ); 	
  
--########################################################
--========================================================
--					STATE MACHINE
--========================================================
--########################################################

SM_REG: process(SCLK, nreset)
begin
	if(nreset='0')				then CS <= idle;
	elsif(rising_edge(SCLK))	then CS <= NS;
	end if;
end process;

SM_NEXT_STATE: 
process(CS, do_setup, tx_done, valid_bit, pause_bit, pause_done, addr_done)
begin
	-- Default Values -----------------------
	Enable_Setup 	<= '1'; -- Active except when idle
	
	case CS is
		when idle =>
			Enable_Setup 	<= '0'; -- give over control of mux when idle
			
			if(do_setup = '1')then 	    NS <= set_mem_adr;
			else                   	    NS <= idle;
			end if;
			
			bit0<='0'; bit1<='0'; bit2<='0'; bit3<='0';
		--................................................
		when set_mem_adr => 			NS <= ld_mem_adr;
			bit0<='1'; bit1<='0'; bit2<='0'; bit3<='0';
		--................................................
		when ld_mem_adr => 				NS <= get_mem_data;
			bit0<='0'; bit1<='1'; bit2<='0'; bit3<='0';
		--................................................
		when get_mem_data => 			NS <= ld_word_cnt;
			bit0<='1'; bit1<='1'; bit2<='0'; bit3<='0';
		--................................................
		when ld_word_cnt => 			NS <= wait1;
			bit0<='0'; bit1<='0'; bit2<='1'; bit3<='0';
		--................................................
		when wait1 =>					NS <= get_word;
			bit0<='1'; bit1<='0'; bit2<='1'; bit3<='0';		--1010  wait1
		--................................................
		when get_word => 				NS <= test_bits;
			bit0<='0'; bit1<='1'; bit2<='1'; bit3<='0';		--0110  get_word
		--................................................
		when test_bits => 		
			if(valid_bit = '0') then	NS <= done; --determines end of setup
			elsif(pause_bit = '1')then 	NS <= pause;
			else                   	    NS <= strb_xmt;
			end if;
			bit0<='1'; bit1<='1'; bit2<='1'; bit3<='0';		--1110  test_bits
		--................................................
		when strb_xmt => 				NS <= transmit;
			bit0<='0'; bit1<='0'; bit2<='0'; bit3<='1';		--0001  strb_xmt
		--................................................
		when transmit =>
			if(tx_done = '1')then 		NS <= inc_adr;
			else                   	    NS <= transmit;
			end if;
			bit0<='1'; bit1<='0'; bit2<='0'; bit3<='1';		--1001 transmit
		--................................................
		when pause =>
			if(pause_done = '1')then 	NS <= inc_adr;
			else                   	    NS <= pause;
			end if;
			bit0<='0'; bit1<='1'; bit2<='0'; bit3<='1';		--0101  pause
		--................................................
		when inc_adr => 				
			if(addr_done = '1')then 	NS <= done;
			else                   	    NS <= wait1;
			end if;
			bit0<='1'; bit1<='1'; bit2<='0'; bit3<='1';		--1101  inc_adr
		--................................................
		when done =>					NS <= done2;
			
			bit0<='0'; bit1<='0'; bit2<='1'; bit3<='1';

		when done2 =>					NS <= idle;
		-- extend the setup done pulse because of SCLK/CLK change of domains
			bit0<='0'; bit1<='0'; bit2<='1'; bit3<='1';

		when others => 					NS <= idle;
			bit0<='1'; bit1<='1'; bit2<='1'; bit3<='1';
	end case;
end process;

-- State Machine Driven Outputs ---------------------------------------
SM_OUTPUTS: process(CS, reset_bit)
begin
	case CS is
		when pause  =>
			pause_ena   <= '1';
			dds_reset 	<= reset_bit;
			tx_strobe   <= '0';
			setp_done	<= '0';
		when get_word  =>		
			pause_ena   <= '0';
			dds_reset 	<= '0';
			tx_strobe   <= '0';
			setp_done	<= '0';
		when test_bits =>		
			pause_ena   <= '0';
			dds_reset 	<= reset_bit;
			tx_strobe   <= '0';
			setp_done	<= '0';
		when strb_xmt =>		
			pause_ena   <= '0';
			dds_reset 	<= '0';
			tx_strobe   <= '1';
			setp_done	<= '0';
		when done =>		
			pause_ena   <= '0';
			dds_reset 	<= '0';
			tx_strobe   <= '0';
			setp_done	<= '1';
		when done2 =>		
			pause_ena   <= '0';
			dds_reset 	<= '0';
			tx_strobe   <= '0';
			setp_done	<= '1';
		when others  =>		
			pause_ena   <= '0';
			dds_reset 	<= '0';
			tx_strobe   <= '0';
			setp_done	<= '0';
	end case;
end process;

--**--   --**--   --**--   --**--   --**--   --**--   --**--
-- PAUSE TIMER  ---------------------------------------------
PAUSE_TMR: counter -- Counter to time a pause
generic map(16)
port map(	clk => SCLK,
			clrn => pause_ena,
			ena  => pause_ena,
			load => '0',
			updn => '1',
			ld_data  => (others => '0'),
			flag_cnt => pause_flag,
			flag_out => pause_done,
			cnt_out  => open );	

		pause_flag <= X"C738"; -- 750us @ 68MHz clock		
	--pause_flag <= "10011000010110"; -- 150us @ 65MHz clock
	--pause_flag <= "00000000000110";   ---short time for simulation

--**--   --**--   --**--   --**--   --**--   --**--   --**--
-- ADDRESS POINTER COUNTER ---------------------------------
ADDR_PTR: counter -- Counter to increment through RAM memory
generic map(5)
port map(	clk => SCLK,
			clrn => cnt_nrst,
			ena  => cnt_ena,
			load => cnt_load,
			updn => '1',
			ld_data  => "00000",
			flag_cnt => "01111",
			flag_out => addr_done,
			cnt_out  => mem_ptr(4 downto 0) );

STATE_MACHINE_CONTROL_OF_ADDR_PTR:
process(CS, Setup_mem_addr, mem_ptr)	
begin
	case CS is
		when idle =>
			mem_addr <= Setup_mem_addr;
			cnt_nrst <= '0'; -- resets counters
			cnt_ena  <= '0'; -- strobe used to increment the counters
			cnt_load <= '0'; -- load the first RAM address into ADDR_PTR count
			
		when set_mem_adr =>
			mem_addr <= mem_ptr;
			cnt_nrst <= '1';
			cnt_ena  <= '0';
			cnt_load <= '1';
						
		when ld_word_cnt =>
			mem_addr <= mem_ptr;
			cnt_nrst <= '1';
			cnt_ena  <= '0';
			cnt_load <= '0';

		when inc_adr =>
			mem_addr <= mem_ptr;
			cnt_nrst <= '1';
			cnt_ena  <= '1';
			cnt_load <= '0';

		when others => 
			mem_addr <= mem_ptr;
			cnt_nrst <= '1';
			cnt_ena  <= '0';
			cnt_load <= '0';

	end case;
end process;	
		
--**--   --**--   --**--   --**--   --**--   --**--   --**-- 
  
	msb_first			<= mem_data_out(47);
	reset_bit 			<= mem_data_out(46);
	ioreset_bit 		<= mem_data_out(45);
	Global_nAddressed 	<= mem_data_out(44);
	pause_bit 			<= mem_data_out(43);
	reserved  			<= mem_data_out(42);
	Chip_Sel1 			<= mem_data_out(41);
	Chip_Sel0 			<= mem_data_out(40);
		
	reserved2  			<= mem_data_out(38);
	valid_bit 			<= mem_data_out(37);

-----------------------------------------------------------------------------
-- Order the bits to load into serial interface according to MSB first or
-- LSB first.  The Instruction Byte will always go at the head of the packet.

process
(mem_data_out, msb_first, Global_nAddressed, Chip_Sel1, Chip_Sel0, xmt_q_rev) 
	variable dd : std_logic_vector(71 downto 0);
begin
	
	xmt_q_rev( 7 downto  0) <= mem_data_out(39 downto 32); -- instruction
	xmt_q_rev(39 downto  8) <= mem_data_out(31 downto  0); -- setup bits
	xmt_q_rev(71 downto 40) <= (others => '0');
	
	if (msb_first = '1') then -- setup for MSB first transmit
		xmt_q(71 downto 64) <= mem_data_out(39 downto 32); -- instruction
		xmt_q(63 downto 32) <= mem_data_out(31 downto  0); -- setup bits
		xmt_q(31 downto  0) <= (others => '0');
	else		
		for i in 71 downto 0 loop dd(i) := xmt_q_rev(71-i);
		end loop;
		
		xmt_q <= dd; -- bits have been reversed for LSB first loading

	end if;
	
	if(Global_nAddressed = '1') then
		CS1 <= '1';
		CS2 <= '1';
		CS3 <= '1';
		CS4 <= '1';
	else
		CS1 <= (NOT Chip_Sel1) AND (NOT Chip_Sel0);
		CS2 <= (NOT Chip_Sel1) AND (    Chip_Sel0);
		CS3 <= (NOT Chip_Sel1) AND (    Chip_Sel0);
		CS4 <= (NOT Chip_Sel1) AND (    Chip_Sel0);
	end if;
end process;

SYNCRONIZED_OUTPUTS: process(SCLK)
begin
	if(rising_edge(SCLK)) then
		Setup_Strb		  	<= tx_strobe;
		Setup_Data		  	<= xmt_q;
		Setup_NBits	      <= "0100111"; -- 39 bits -- includes data bit + 8 instruction bits -1
		Setup_MSB_nLSB    <= msb_first;
		Setup_Ena_IOReset	<= ioreset_bit;
		Setup_DDS_Reset   <= dds_reset;
		DDS_Setup_Done  	<= setp_done;
	end if;
end process;

end main;		
		
			