Library IEEE;
USE IEEE.std_logic_1164.all;
USE IEEE.std_logic_unsigned.all;

-- My Definitions --------------------------------
USE work.pkg_081006.all; 
USE work.DDS_VME_CONSTANTS.all;

----------------------------------------------------------------
-- Placement of these 32 bit registers is controlled by the 
-- constant VME_CTRL_UPPER_FIRST.  This constant is defined in
-- the included file DDS_VME_CONSTANTS.vhd.

entity Debug_Data_Select_v2 is 
	port( 	
			clk					: in std_logic;
			nreset				: in std_logic;
			Select_Data			: in std_logic_vector( 7 downto 0);
			Debug_Offset		: in std_logic_vector(15 downto 0);
			Run_Curves			: in std_logic;
			Rewind_Curves		: in std_logic;
						
			Frequency			: in std_logic_vector(31 downto 0);
			Freq_Strobe			: in std_logic;
			
			Freq_Out			: in std_logic_vector(31 downto 0);
			Freq_Out_Strobe		: in std_logic;
			
			Phase_A				: in std_logic_vector(15 downto 0);
			Phase_A_Strobe		: in std_logic;
			
			Phase_B				: in std_logic_vector(15 downto 0);
			Phase_B_Strobe		: in std_logic;						
			
			Phase_Offset_A			: in std_logic_vector(15 downto 0);
			Phase_Offset_Strobe_A	: in std_logic;						
                                
			Phase_Offset_B			: in std_logic_vector(15 downto 0);
			Phase_Offset_Strobe_B	: in std_logic;						
 
 			ParaPhase			: in std_logic_vector(15 downto 0);
			ParaPhase_Strobe	: in std_logic;
			
			Ref_Phase			: in std_logic_vector(15 downto 0);
			Ref_Phase_Strobe	: in std_logic;
			
			Rpos_Phase		 	: in std_logic_vector(15 downto 0);
			Rpos_Phase_Strobe	: in std_logic;
			
			Phase_Err			: in std_logic_vector(31 downto 0);
			Phase_Err_Strobe	: in std_logic;
			
			FIFO_FULL			: in std_logic;
						
			Data_to_FIFO 		: out std_logic_vector(31 downto 0);
			WR_FIFO				: out std_logic
			
		);
	end Debug_Data_Select_v2;
	
architecture main of Debug_Data_Select_v2 is	

signal sync_start	: std_logic_vector(3 downto 0);
signal start		: std_logic;
signal init			: std_logic;
signal stop			: std_logic;
signal data_out		: std_logic_vector(31 downto 0);
signal strobe_out	: std_logic;

-- Debug Offset, Number of Strobes Before Writting FIFO ----------------
signal stb_count		: std_logic_vector(15 downto 0);
signal ena_stb_count	: std_logic;

-- SETUP_STATE_MACHINE ---------------------------------------
type STATE_TYPE is
(idle, wait1, run);
signal CS, NS : STATE_TYPE;
            
--WWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWW
--MMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMM
begin
------------------------------------------------------------------------
SYNCRONIZE_START: 
process(clk, nreset, Run_Curves, sync_start)
begin
	if   (nreset = '0') then sync_start <= (others => '0');
	elsif rising_edge(clk) then
		sync_start(0) <= Run_Curves;
		sync_start(1) <= sync_start(0);
		sync_start(2) <= sync_start(1);
		sync_start(3) <= sync_start(2);
	end if;
end process;

	start <= sync_start(0) AND NOT sync_start(1);
	init  <= sync_start(2) AND NOT sync_start(3);
-------------------------------------------------------------------------
SYNC_STOP: led -- syncronize the signal to stop update polling sequence.
port map(	sig_in  => NOT Rewind_Curves, 
			clk     => clk, 
			clrn    => '1', 
			pls_out => stop); 


-------------------------------------------------------------------------
STROBE_COUNTER:
process(clk, nreset, ena_stb_count, stb_count)
begin
	if   (nreset = '0') then stb_count <= (others => '0');
	elsif rising_edge(clk) then
		if   (ena_stb_count = '0')	then stb_count <= (others => '0');
		elsif(ena_stb_count = '1')	then
			if (strobe_out  = '1')	then stb_count <= stb_count + 1;
			end if;
		end if;
	end if;
end process;
-------------------------------------------------------------------------
DATA_SELECT:
process( Select_Data, Freq_Out, Freq_Out_Strobe, Frequency, Freq_Strobe, 
		 Phase_A, Phase_A_Strobe, Phase_B, Phase_B_Strobe, ParaPhase, 
		 ParaPhase_Strobe, Ref_Phase, Ref_Phase_Strobe, Phase_Err,
		 Phase_Err_Strobe, Phase_Offset_A, Phase_Offset_Strobe_A,
		 Phase_Offset_B, Phase_Offset_Strobe_B, Rpos_Phase, Rpos_Phase_Strobe )
begin
	case Select_Data is
		when X"01" =>	data_out   <= Freq_Out;
						strobe_out <= Freq_Out_Strobe;
		
		when X"02" =>	data_out   <= Phase_Err;			
						strobe_out <= Phase_Err_Strobe;

		when X"03" =>	data_out   <= Frequency;
						strobe_out <= Freq_Strobe;
		       
		when X"04" =>	data_out   <= X"0000" & Phase_A;
						strobe_out <= Phase_A_Strobe;
			   
		when X"05" =>	data_out   <= X"0000" & Phase_B;
						strobe_out <= Phase_B_Strobe;
						             
		when X"06" =>	data_out   <= X"0000" & Phase_Offset_A;
						strobe_out <= Phase_Offset_Strobe_A;	
			   
		when X"07" =>	data_out   <= X"0000" & Phase_Offset_B;
						strobe_out <= Phase_Offset_Strobe_B;
			   
		when X"08" =>	data_out   <= X"0000" & ParaPhase;			
						strobe_out <= ParaPhase_Strobe;
						
		when X"09" =>	data_out   <= X"0000" & Ref_Phase;			
						strobe_out <= Ref_Phase_Strobe;
						
		when X"0A" =>	data_out   <= X"0000" & Rpos_Phase;			
						strobe_out <= Rpos_Phase_Strobe;
						
		when others =>	data_out   <= X"00000000";
						strobe_out <= '0';
	end case;
end process;

			
--{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{
--                S T A T E   M A C H I N E
--}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}
SM_REG: process(clk, nreset)
begin
	if(nreset='0')			then CS <= idle;
	elsif(rising_edge(clk))	then CS <= NS;
	end if;
end process;

SM_NEXT_STATE: 
process
(CS, start, stop, FIFO_FULL, data_out, strobe_out, Debug_Offset, stb_count, init )
begin
	-- defaults ----------------------
		Data_to_FIFO <= X"00000000";
		WR_FIFO		 <= '0';
		
		ena_stb_count <= '0';
	    
	case CS is
		when idle =>	
			if   (start = '1') AND (FIFO_FULL = '0') then				
												NS <= wait1;
			else								NS <= idle;
			end if;
		
		when wait1 =>
			ena_stb_count <= '1';
			if(stb_count >= Debug_Offset) then	NS <= run;
			else								NS <= wait1;
			end if;
			
		when run =>	    
			if   (stop = '1') then				NS <= idle;
			elsif(FIFO_FULL = '1') then			NS <= idle;
			else								NS <= run;
				Data_to_FIFO <= data_out;
				WR_FIFO		 <= strobe_out OR init;
			end if;
		
		when others =>							NS <= idle;
	end case;
end process;
end;