Library IEEE;
USE IEEE.std_logic_1164.all;
USE IEEE.std_logic_unsigned.all;

-- My Definitions --------------------------------
USE work.pkg_081006.all; 

------------------------------------------------------
-- MODULE DESCRIPTION
-- This module either writes a block of RAM memory to FLASH
-- memory or reads a block of FLASH memory to RAM.  The blocks
-- or memory are defined by the pointer to RAM, the pointer
-- to FLASH and the number of words to transfer variable.
--
-- The FLASH memory can only be written (programmed) after
-- the contents of the block to be written are erased.
--
-- The data to be written to flash is assumed to be 32 bit.
-- The flash itsself is 16 bits wide.  The LSW is stored
-- at the lower address and the MSW is stored at the next 
-- higher address (little-endian).
------------------------------------------------------

entity Flash_Interface_v5 is 
	port( 
		clk				: in std_logic;
		nreset			: in std_logic;
		RAM_Addr_In		: in std_logic_vector(23 downto 0);		
		FLASH_Addr_In	: in std_logic_vector(24 downto 1);		
		NWords			: in std_logic_vector(15 downto 0);
		Unlck_Block		: in std_logic;
		Erase_Block		: in std_logic; 
		Write_Flash		: in std_logic;  -- trigger for Write process
		Read_Flash		: in std_logic;  -- trigger for Read process
		FLASH_Bus_Permit: in std_logic;	
			
		RAM_Data_In		: in std_logic_vector(31 downto 0);
		FLASH_Data_In	: in std_logic_vector(15 downto 0);
		FLASH_Status_7	: in std_logic; --Flash status data bit 7
		FLASH_Wait		: in std_logic; -- Only used with "syncronous read operations"
		
		Lower_Xfr_Done	: in std_logic;
		
		RAM_Data		: out std_logic_vector(31 downto 0);
		RAM_Addr		: out std_logic_vector(23 downto 0);		
		RAM_WR			: out std_logic;
		RAM_RD			: out std_logic;
		FLASH_Ena		: out std_logic; -- RAM accessed by Write FLASH process
										 -- This enable may move to a higher control level <<<<<<
		FLASH_Bus_Req	: out std_logic;
		
		FLASH_Addr		: out std_logic_vector(24 downto 1);
		FLASH_Data		: out std_logic_vector(15 downto 0);
		FLASH_nCE		: out std_logic;
		FLASH_nOE		: out std_logic;
		FLASH_nWE		: out std_logic;
		FLASH_nRST		: out std_logic;
		FLASH_nADV 		: out std_logic;
		FLASH_DCLK		: out std_logic;

		Operation_Done	: out std_logic;
		Transfer_Done	: out std_logic  -- Strobes when SR is done
		);
	end Flash_Interface_v5;
	
---------------------------------------------------------
architecture main of Flash_Interface_v5 is
-- SYNC_WRITE_FLASH Signals -----------------------------
	signal uf		:std_logic;
	signal ef		: std_logic;
	signal wf		: std_logic;
	signal rf		: std_logic;

-- Count writes -----------------------------------------
	signal wr_count 	: std_logic_vector(15 downto 0);
	signal set_wr_count : std_logic;
	signal dec_wr_count : std_logic;
	
	signal N16bitWords	: std_logic_vector(15 downto 0);
	
	signal buf_count	 : std_logic_vector(4 downto 0);
	signal set_buf_count : std_logic;
	signal dec_buf_count : std_logic;

-- RAM Address Pointer ----------------------------------
	signal ram_adr_ptr 	: std_logic_vector(23 downto 1);
	signal set_ram_adr	: std_logic;
	signal inc_ram_adr	: std_logic;

-- FLASH Address Pointer --------------------------------
--	signal converted_FLASH_Addr_In : std_logic_vector(24 downto 1);
	signal flsh_adr_ptr :std_logic_vector(24 downto 1);
	signal set_flsh_adr :std_logic;
	signal inc_flsh_adr :std_logic;
	signal set_original_blk_addr : std_logic;
	
-- FLASH Status Latch -----------------------------------
	signal flsh_stat_latch 	: std_logic;
	signal get_flsh_status 	: std_logic;

-- FLASH Data Latch -------------------------------------
	signal flsh_data_latch	: std_logic_vector(31 downto 0);
	signal get_flsh_data	: std_logic;

-- Pause Timer ------------------------------------
	signal pause_tmr 	: std_logic_vector(2 downto 0);
	signal inc_pause_tmr: std_logic;
	       
-- Pause Timer 2 ------------------------------------
	signal pause_tmr2 	 : std_logic_vector(2 downto 0);
	signal inc_pause_tmr2: std_logic;
	       
-- STATE_MACHINE High Level --------------------------------------
	type STATE_TYPE is
	(idle, wr_unlck_cmd1, wr_unlck_cmd2, chk_unlck_status, setup_buf_cnt,  
	 wr_erase_cmd1, wr_erase_cmd2, chk_erase_status, chk_erase_again,   
	 flash_oper_done, wr_prog_cmd1, chk_wsm_status, wr_prog_cmd2, get_bus_permit,
	 wr_ram_data, inc_pointers, get_ram_data, confirm_prog, chk_prog_status, 
	 chk_prog_again, prog_end, get_bus_permit2, wr_read_cmd1, rd_flsh_data, pause2, 
	 wr_to_ram, inc_pointers2, read_end);	
	signal CS, NS : STATE_TYPE;	
	
-- STATE_MACHINE Low Level --------------------------------------
	type STATE_TYPE2 is
		(idle2, setup_wr, finish_wr, setup_stat, chk_status, signal_ready, 
		 signal_busy, setup_rd, data_read);
	signal CS2, NS2 : STATE_TYPE2;	
	
-- Signals Between the State Machines ------------------------------
	signal wr_flash		: std_logic;
	signal rd_flash		: std_logic;
	signal rd_status	: std_logic;
	
	signal wr_done		: std_logic;
	signal rd_done		: std_logic;
	signal flash_ready	: std_logic;
	signal flash_busy	: std_logic;
	
--========================================================
--========================================================
begin

--**-- --**--   --**--   --**--   --**--   --**--   --**--   
STROBE_UF: led -- syncronize the Read Flash Strobe.
port map(	sig_in  => Unlck_Block, 
			clk     => clk, 
			clrn    => '1', 
			pls_out => uf); 

STROBE_EF: led -- syncronize the Read Flash Strobe.
port map(	sig_in  => Erase_Block, 
			clk     => clk, 
			clrn    => '1', 
			pls_out => ef); 

STROBE_WF: led -- syncronize the Write Flash Strobe.
port map(	sig_in  => Write_Flash, 
			clk     => clk, 
			clrn    => '1', 
			pls_out => wf); 
			
STROBE_RF: led -- syncronize the Read Flash Strobe.
port map(	sig_in  => Read_Flash, 
			clk     => clk, 
			clrn    => '1', 
			pls_out => rf); 

-----------------------------------------------------------------------------
PAUSE_TIMER: -- Timer used to time setup time for FLASH writes
process (clk, inc_pause_tmr)
begin
	if rising_edge(clk) then
		if   (inc_pause_tmr = '0') then pause_tmr <= "000";
		else							pause_tmr <= pause_tmr + 1;
		end if;
	end if;
end process;

-----------------------------------------------------------------------------
PAUSE_TIMER2: -- Timer used to time setup time for RAM reads
process (clk, inc_pause_tmr2)
begin
	if rising_edge(clk) then
		if   (inc_pause_tmr2 = '0') then pause_tmr2 <= "000";
		else							 pause_tmr2 <= pause_tmr2 + 1;
		end if;
	end if;
end process;

-----------------------------------------------------------------------------
-- COUNTING THE WRITES / READS ----------------------------------------------
-- Count the Writes preformed.  wr_count is the number of 
-- 16 bit word writes or reads that need to be performed.  NWords is the
-- number of 32 bit word writes or reads that need to be performed, hence
-- we multiply NWords by 2 when setting up wr_count.  We also subtract 1
-- to make things work out.  The FLASH wants to see N-1 when setting up
-- the Buffered Programming.

-- The Buffered Programming of the FLASH involves writting to a 32 word
-- buffer (16 bit wide buffer) as many times as needed.


N16bitWords(15 downto 0) <= (NWords(14 downto 0) & '0') - 1;

WRITE_COUNT:
process (clk, set_wr_count, dec_wr_count)
begin
	if rising_edge(clk) then
		if   (set_wr_count = '1') then wr_count <= N16bitWords;
					
		elsif(dec_wr_count = '1') then wr_count <= wr_count - 1;
		
		end if;
	end if;
end process;

BUFFER_COUNT:
process(clk, set_buf_count, dec_buf_count)
begin
	if rising_edge(clk) then
		if   (set_buf_count = '1') then 
			if(wr_count < X"0020") then buf_count <= wr_count(4 downto 0);
			else						buf_count <= "11111"; --THAT IS 31;
			end if;
		elsif(dec_buf_count = '1') then buf_count <= buf_count - 1;

		end if;
	end if;
end process;

-----------------------------------------------------------------------------
-- The RAM_ADDR_PTR is a counter that increments from the address that we
-- start at.  Note addresses are assumed to have a byte (8 bit) resolution.
-- 
-- Both this pointer and the flash address, which is a 16 bit pointer, are 
-- incremented at the same time. So we increment the RAM pointer 
-- "ram_adr_ptr(23..1)" 2 bytes (16 bits) at a time.  Since the RAM data being 
-- addressed is 4 bytes (32 bits), we only apply ram_adr_ptr(23..2) to the 
-- RAM_Addr bus.

RAM_ADDR_PTR:
process (clk, set_ram_adr, inc_ram_adr, RAM_Addr_In, ram_adr_ptr)
begin
	if rising_edge(clk) then
		if   (set_ram_adr = '1') then 
			ram_adr_ptr(23 downto 1) <= RAM_Addr_In(23 downto 1);
			
		elsif(inc_ram_adr = '1') then 
			ram_adr_ptr(23 downto 1) <= ram_adr_ptr + 1;
			
		end if;
	end if;
	RAM_Addr(23 downto 2) <= ram_adr_ptr(23 downto 2); -- making 4 byte steps
	RAM_Addr( 1 downto 0) <= "00";					   -- 32 bit data
end process;

-----------------------------------------------------------------------------
FLASH_ADDR_PTR:
process (clk, set_flsh_adr, inc_flsh_adr, FLASH_Addr_In, flsh_adr_ptr, 
		 set_original_blk_addr )
begin
	if rising_edge(clk) then
		if   (set_flsh_adr = '1') then flsh_adr_ptr <= FLASH_Addr_In;
		elsif(inc_flsh_adr = '1') then flsh_adr_ptr <= flsh_adr_ptr + 1;
		end if;
	end if;
	
	if (set_original_blk_addr = '1') then
		FLASH_Addr(24 downto 1) <= FLASH_Addr_In;
	else
		FLASH_Addr(24 downto 1) <= flsh_adr_ptr(24 downto 1);
		
	end if;
end process;

-----------------------------------------------------------------------------
FLASH_STATUS_LATCH:
process (clk, nreset, get_flsh_status)
begin
	if rising_edge(clk) then
		if   (nreset = '0') 	     then flsh_stat_latch <= '0';
		elsif(get_flsh_status = '1') then flsh_stat_latch <= FLASH_Status_7;
		end if;
	end if;
end process;

-----------------------------------------------------------------------------
FLASH_DATA_LATCH:
process (clk, nreset, get_flsh_data, ram_adr_ptr, flsh_data_latch)
begin
	if rising_edge(clk) then
		if   (nreset = '0') 	     then flsh_data_latch <= (others => '0');
		elsif(get_flsh_data = '1') then 
			if( ram_adr_ptr(1) = '0') then 	
					flsh_data_latch(15 downto 0)  <= FLASH_Data_In(15 downto 0);
			else	flsh_data_latch(31 downto 16) <= FLASH_Data_In(15 downto 0);
			end if;
		end if;
	end if;
	RAM_Data(31 downto 0) <= flsh_data_latch(31 downto 0);
	-- here is the gimick. ram_adr_ptr advances in 16 bit address steps
	-- while RAM_Addr advances only half as often in 32 bit address steps.
	-- we will end up writing each RAM_Addr location twice.  first with
	-- only the lower 16 data bits valid and then with both upper and
	-- lower bits valid.
end process;


--**--   --**--   --**--   --**--   --**--   --**--   --**--  
----------------------------------------------------
-- STATE MACHINE  
----------------------------------------------------
SM_REG: process(clk, nreset)
begin
	if(nreset='0')				then CS <= idle;
	elsif(rising_edge(clk))		then CS <= NS;
	end if;
end process;

SM_NEXT_STATE: 
process(CS, uf, ef, wf, rf, wr_count, wr_done, rd_done, ram_adr_ptr, flash_ready, 
		flash_busy, NWords, RAM_Data_In, buf_count, pause_tmr2, Lower_Xfr_Done, 
		FLASH_Bus_Permit)
begin
	-- Defaults -------------------------------------
		FLASH_Data		<= (others => '0');
		RAM_WR 			<= '0';
		RAM_RD			<= '0';
	
		wr_flash		<= '0'; -- signal to FLASH_SM (the second state machine)
		rd_flash		<= '0'; -- signal to FLASH_SM
		rd_status		<= '0'; -- signal to FLASH_SM
		
		set_flsh_adr    <= '0';
		inc_flsh_adr    <= '0';
		set_original_blk_addr <= '0';
		
		set_ram_adr	    <= '0';
		inc_ram_adr	    <= '0';
		
		set_wr_count	<= '0';
		dec_wr_count 	<= '0';
		
		set_buf_count 	<= '0';
		dec_buf_count 	<= '0';
		
		inc_pause_tmr2	<= '0';
		                 
		FLASH_Ena		<= '0'; 		 -- Enable RAM access by Write FLASH process
										 -- This enable may move to a higher control level <<<<<<
		FLASH_Bus_Req	<= '0';
		
		Operation_Done	<= '0';
		Transfer_Done	<= '0';  -- Strobes when procedure is done
	--------------------------------------------------

	case CS is
		when idle =>
			FLASH_Ena		<= '0'; -- FLASH_Ena is inactive during idle
			set_wr_count	<= '1'; -- This gets latched once leaving idle
			set_ram_adr		<= '1'; -- go ahead and setup ram address pointer
			set_flsh_adr 	<= '1'; -- FLASH_Addr <= first address
									-- This gets latched once leaving idle
			if   (uf = '1')   then 			NS <= wr_unlck_cmd1; -- do block unlock
			elsif(ef = '1')   then 			NS <= wr_erase_cmd1; -- do block erase
			elsif(wf = '1')   then			NS <= setup_buf_cnt; -- do block write
			elsif(rf = '1')   then			NS <= get_bus_permit2;  -- do array read
			else                   			NS <= idle;
			end if;
			
		------------------------------------------------------------------
		-- ERASE BLOCK START HERE ---------------------------------------
		when wr_unlck_cmd1 =>
			FLASH_Data		<= X"0060"; -- issue block lock change command
			
			wr_flash		<= '1';
			if (wr_done = '1') then			NS <= wr_unlck_cmd2;
			else							NS <= wr_unlck_cmd1;
			end if;
			
		------------------------------------------------------------------
		when wr_unlck_cmd2 =>			
			FLASH_Data 		<= X"00D0"; 
			wr_flash		<= '1';			
			if (wr_done = '1') then			NS <= chk_unlck_status;
			else							NS <= wr_unlck_cmd2;
			end if;
			
		------------------------------------------------------------------
		when chk_unlck_status =>
			rd_status		<= '1';			
			if   (flash_ready = '1') then	NS <= flash_oper_done;--??????
			elsif(flash_busy  = '1') then	NS <= flash_oper_done;--??????
			else							NS <= chk_unlck_status;
			end if;	
			
		------------------------------------------------------------------			
		------------------------------------------------------------------	
		when wr_erase_cmd1 =>           			
			FLASH_Data 		<= X"0020"; -- issue erase block command
			-- Signal other SM to write the flash with the above address and data
			wr_flash		<= '1'; 
			if (wr_done = '1') then			NS <= wr_erase_cmd2;
			else							NS <= wr_erase_cmd1;
			end if;
			
		------------------------------------------------------------------	
		when wr_erase_cmd2 =>           -- issue erase block confirm command		
			FLASH_Data 		<= X"00D0"; 
			wr_flash		<= '1';			
			if (wr_done = '1') then			NS <= chk_erase_status;
			else							NS <= wr_erase_cmd2;
			end if;
			
		------------------------------------------------------------------
		when chk_erase_status =>		-- watch status reg for erase to complete
			rd_status		<= '1';			
			if   (flash_ready = '1') then	NS <= flash_oper_done;
			elsif(flash_busy  = '1') then	NS <= chk_erase_again; --if not ready
			else							NS <= chk_erase_status;
			end if;	
							
		when chk_erase_again =>				NS <= chk_erase_status;
				-- The FLASH_SM will have returned to idle
				-- Next we re-issue the erase status read request
				
		------------------------------------------------------------------			
		when flash_oper_done =>				NS <= idle;
				Operation_Done	<= '1';
				
		------------------------------------------------------------------			
		------------------------------------------------------------------
		-- START OUTER WRITE FLASH LOOP ----------------------------------	
		when setup_buf_cnt =>				NS <= wr_prog_cmd1;			
				set_buf_count 	<= '1';	
						
		when wr_prog_cmd1 =>           	-- issue the write buffer command		
			FLASH_Data 		<= X"00E8"; 
			wr_flash		<= '1';			
			if (wr_done = '1') then			NS <= chk_wsm_status;
			else									NS <= wr_prog_cmd1;
			end if;
			
		------------------------------------------------------------------	
		when chk_wsm_status =>			-- wait for flash to be ready for writing
			rd_status		<= '1';			
			if   (flash_ready = '1') then	NS <= wr_prog_cmd2;
			elsif(flash_busy  = '1') then	NS <= wr_prog_cmd1; --if not ready
			else									NS <= chk_wsm_status;
			end if;
								
		------------------------------------------------------------------				
		when wr_prog_cmd2 => -- write N-1 where N = number of writes to "buffer"		
			FLASH_Data(15 downto 0) <= "00000000000" & buf_count(4 downto 0);
			
			wr_flash		<= '1';			
			if (wr_done = '1') then			NS <= get_bus_permit;
			else									NS <= wr_prog_cmd2;
			end if;
	 	
		------------------------------------------------------------------	
		------------------------------------------------------------------	
		-- START WRITE FLASH "BUFFER"  INNER LOOP ------------------------
	 	when get_bus_permit =>
			FLASH_Bus_Req 	<= '1';
			if  (FLASH_Bus_Permit = '1') then	NS <= get_ram_data;
			else								NS <= get_bus_permit;
			end if;
			
		when get_ram_data =>				NS <= wr_ram_data; 
			FLASH_Bus_Req 	<= '1';
			FLASH_Ena		<= '1';
			-- pause to allow ram data to setup.
			-- increment the pause timer and wait for timeout
			if   (ram_adr_ptr(23 downto 2) & "00" < X"020000") then 
				inc_pause_tmr2	<= '1'; 			
				if (pause_tmr2 = "011")  then	NS <= wr_ram_data;
				else							NS <= get_ram_data;
				end if;
			else
				RAM_RD		<= '1';
				if(Lower_Xfr_Done = '1') then 	NS <= wr_ram_data;
				else							NS <= get_ram_data;
				end if;
			end if;
		
		------------------------------------------------------------------	
		when wr_ram_data =>           -- write ram data to flash			
			FLASH_Bus_Req 	<= '1';
			FLASH_Ena		<= '1';
			
			if( ram_adr_ptr(1) = '0') then 	
									FLASH_Data <= RAM_Data_In(15 downto 0);
			else					FLASH_Data <= RAM_Data_In(31 downto 16);
			end if;
			
			wr_flash		<= '1';			
			if (wr_done = '1') then		NS <= inc_pointers;
			else								NS <= wr_ram_data;
			end if;
		------------------------------------------------------------------	
		when inc_pointers =>           	-- increment ram and flash adr pointers		
			FLASH_Bus_Req 	<= '1';
			FLASH_Ena		<= '1';
			
			if(buf_count(4 downto 0) = "00000") then	-- end of buffer	
												NS <= confirm_prog; 
			else								NS <= get_ram_data; -- loop back up
				dec_wr_count 	<= '1';
				dec_buf_count 	<= '1';
				inc_ram_adr 	<= '1'; 
				inc_flsh_adr 	<= '1';
			end if;
	
		-- END WRITE FLASH "BUFFER"  INNER LOOP --------------------------
		------------------------------------------------------------------
		------------------------------------------------------------------	
		when confirm_prog =>
			set_original_blk_addr <= '1';
			
			FLASH_Data 		<= X"00D0"; -- issue confirm command
			wr_flash		<= '1';			
			if (wr_done = '1') then	NS <= chk_prog_status;
			else							NS <= confirm_prog;
			end if;
			
		-----------------------------------------------------------------------------
		when chk_prog_status =>		-- watch status reg for buffer write to complete
			rd_status		<= '1';	
					
			if   (flash_ready = '1') then	
				if(wr_count = X"0000") then -- end of write flash process
											NS <= prog_end;
				else
					dec_wr_count 	<= '1';
					inc_ram_adr 	<= '1'; 
					inc_flsh_adr 	<= '1';
											NS <= setup_buf_cnt;
				end if;
				
			elsif(flash_busy  = '1') then	NS <= chk_prog_again; --if not ready
			else							NS <= chk_prog_status;
			end if;	
		
		-- END OUTER FLASH WRITE LOOP ------------------------------------					
		------------------------------------------------------------------	
		------------------------------------------------------------------	
		
		when chk_prog_again =>				NS <= chk_prog_status;
				-- The FLASH_SM will have returned to idle
				-- Next we re-issue the program status read request
				
		------------------------------------------------------------------
		when prog_end =>					NS <= idle;	
			Transfer_Done	<= '1';	
				
		------------------------------------------------------------------
		--		Read Array States				
		------------------------------------------------------------------
		when get_bus_permit2 =>
			FLASH_Bus_Req 	<= '1';
			if  (FLASH_Bus_Permit = '1') then	NS <= wr_read_cmd1;
			else											NS <= get_bus_permit2;
			end if;
									
		when wr_read_cmd1 =>
			FLASH_Bus_Req 	<= '1';
			FLASH_Ena		<= '1';
			
			FLASH_Data 		<= X"00FF"; 
			wr_flash		<= '1';			
			if (wr_done = '1') then			NS <= rd_flsh_data;
			else									NS <= wr_read_cmd1;
			end if;		
		
		------------------------------------------------------------------			
		when rd_flsh_data =>           -- read data from the FLASH
			FLASH_Bus_Req 	<= '1';
			FLASH_Ena		<= '1';
					
			rd_flash		<= '1';			
			if (rd_done = '1') then			NS <= pause2;
			else									NS <= rd_flsh_data;
			end if;
		
		------------------------------------------------------------------
		when pause2 =>						NS <= wr_to_ram; --- new 5/13/2010
			FLASH_Bus_Req 	<= '1';
			FLASH_Ena		<= '1';
		
		when wr_to_ram =>					NS <= inc_pointers2;
			FLASH_Bus_Req 	<= '1';
			FLASH_Ena		<= '1';
			RAM_WR <= '1';
			
		------------------------------------------------------------------	
		when inc_pointers2 =>      -- increment ram and flash adr pointers		
			FLASH_Bus_Req 	<= '1';
			FLASH_Ena		<= '1';
			if(wr_count = X"0000") then		NS <= read_end; 
			else
											NS <= rd_flsh_data;
				dec_wr_count 	<= '1';
				inc_ram_adr 	<= '1'; 
				inc_flsh_adr 	<= '1';
			end if;
			
		------------------------------------------------------------------	
		when read_end =>					NS <= idle;	
			Transfer_Done	<= '1';	
				
		------------------------------------------------------------------
		when others =>              		NS <= idle;
	end case;
end process;

--**--   --**--   --**--   --**--   --**--   --**--   --**--  
----------------------------------------------------
-- FLASH OPERATION STATE MACHINE  
----------------------------------------------------
FLASH_SM: process(clk, nreset)
begin
	if(nreset='0')				then CS2 <= idle2;
	elsif(rising_edge(clk))		then CS2 <= NS2;
	end if;
end process;

FLASH_nADV 		<= '0';
FLASH_DCLK		<= '1'; -- set high for asyncronous reads and writes
		
FLASH_SM_NEXT_STATE: 
process(CS2, wr_flash, rd_flash, rd_status, pause_tmr, flsh_stat_latch)
begin
	-- Defaults -------------------------------------	               
		FLASH_nCE		<= '1';
		FLASH_nOE		<= '1';
		FLASH_nWE		<= '1';
		FLASH_nRST		<= '1';
		
		wr_done			<= '0';
		rd_done			<= '0';
		flash_ready 	<= '0';
        flash_busy  	<= '0';
        
		inc_pause_tmr	<= '0';
		
		get_flsh_status <= '0';
		get_flsh_data	<= '0';
	--------------------------------------------------

	case CS2 is
		when idle2 =>
			if   (wr_flash  = '1')  then 		NS2 <= setup_wr;
			elsif(rd_flash  = '1')  then    	NS2 <= setup_rd;
			elsif(rd_status = '1')  then    	NS2 <= setup_stat;
			else								NS2 <= idle2;
			end if;
					
		when setup_wr =>           			
			FLASH_nCE		<= '0'; 
			FLASH_nWE		<= '0';
			
			-- increment the pause timer and wait for timeout
			inc_pause_tmr	<= '1'; 			
			if (pause_tmr = "011") then			NS2 <= finish_wr;
			else								NS2 <= setup_wr;
			end if;
		
		when finish_wr =>						NS2 <= idle2;
			FLASH_nCE		<= '1'; -- default
			FLASH_nWE		<= '1'; -- default
			wr_done			<= '1';
		
		---------------------------------------------------------
		when setup_stat =>		
			FLASH_nCE		<= '0'; 
			FLASH_nOE		<= '0';
			get_flsh_status <= '1'; -- latches the status reg value

			inc_pause_tmr	<= '1'; -- increment the pause timer		
			if (pause_tmr = "101") then			NS2 <= chk_status;
			else								NS2 <= setup_stat;
			end if;

		when chk_status =>
			FLASH_nCE		<= '1';
			FLASH_nOE		<= '1';
			
			if (flsh_stat_latch = '1') then		NS2 <= signal_ready;
			else								NS2 <= signal_busy;
			end if;
			
		when signal_ready =>           			NS2 <= idle2;
			flash_ready <= '1';
			
		when signal_busy =>           			NS2 <= idle2;
			flash_busy <= '1';
		---------------------------------------------------------
		when setup_rd =>		
			FLASH_nCE		<= '0'; 
			FLASH_nOE		<= '0';
			get_flsh_data	<= '1'; -- latch flash data

			inc_pause_tmr	<= '1'; -- increment the pause timer		
			if (pause_tmr = "101") then			NS2 <= data_read;
			else								NS2 <= setup_rd;
			end if;
			
		---------------------------------------------------------
		when data_read =>						NS2 <= idle2;
				FLASH_nCE		<= '0'; --- new 5/13/2010  more hold time
				FLASH_nOE		<= '0'; --- new 5/13/2010
				rd_done <= '1';
		---------------------------------------------------------
		when others =>              			NS2 <= idle2;
	end case;
end process;

end main;		
		
--November 2007 Datasheet
--Order Number: 314749-004 61
--Intel StrataFlash� Embedded Memory P33

--11.2 Buffered Programming

--The device features a 32-word buffer to enable optimum programming performance.
--For Buffered Programming, data is first written to an on-chip write buffer. Then the
--buffer data is programmed into the flash memory array in buffer-size increments. This
--can improve system programming performance significantly over non-buffered
--programming.
--
--When the Buffered Programming Setup command is issued (see Section 9.2, �Device
--Commands� on page 47), Status Register information is updated and reflects the
--availability of the buffer. SR[7] indicates buffer availability: if set, the buffer is
--available; if cleared, the buffer is not available. To retry, issue the Buffered
--Programming Setup command again, and re-check SR[7]. When SR[7] is set, the
--buffer is ready for loading. (see Figure 41, �Buffer Program Flowchart� on page 85).
--On the next write, a word count is written to the device at the buffer address. This tells
--the device how many data words will be written to the buffer, up to the maximum size
--of the buffer.
--
--On the next write, a device start address is given along with the first data to be written
--to the flash memory array. Subsequent writes provide additional device addresses and
--data. All data addresses must lie within the start address plus the word count.
--Optimum programming performance and lower power usage are obtained by aligning
--the starting address at the beginning of a 32-word boundary (A[4:0] = 0x00). Crossing
--a 32-word boundary during programming will double the total programming time.

--After the last data is written to the buffer, the Buffered Programming Confirm command
--must be issued to the original block address. The WSM begins to program buffer
--contents to the flash memory array. If a command other than the Buffered
--Programming Confirm command is written to the device, a command sequence error
--occurs and SR[7,5,4] are set. If an error occurs while writing to the array, the device
--stops programming, and SR[7,4] are set, indicating a programming failure.
--
--When Buffered Programming has completed, additional buffer writes can be initiated by
--issuing another Buffered Programming Setup command and repeating the buffered
--program sequence. Buffered programming may be performed with VPP = VPPL or VPPH
--(see Section 5.2, �Operating Conditions� on page 28 for limitations when operating the
--device with VPP = VPPH).
--
--If an attempt is made to program past an erase-block boundary using the Buffered
--Program command, the device aborts the operation. This generates a command
--sequence error, and SR[5,4] are set.
--If Buffered programming is attempted while VPP is below VPPLK, SR[4,3] are set. If any
--errors are detected that have set Status Register bits, the Status Register should be
--cleared using the Clear Status Register command.			


--14.1 Read Status Register
--
--To read the Status Register, issue the Read Status Register command at any address.
--Status Register information is available to which the Read Status Register, Word
--Program, or Block Erase command was issued. SRD is automatically made available
--following a Word Program, Block Erase, or Block Lock command sequence. Reads from
--the device after any of these command sequences outputs the device�s status until
--another valid command is written (e.g. the Read Array command).

--The Status Register is read using single asynchronous-mode or synchronous burst
--mode reads. SRD is output on DQ[7:0], while 0x00 is output on DQ[15:8]. In
--asynchronous mode the falling edge of OE#, or CE# (whichever occurs first) updates
--and latches the Status Register contents. However, when reading the Status Register in
--synchronous burst mode, CE# or ADV# must be toggled to update SRD.

--The Device Write Status bit (SR[7]) provides overall status of the device. SR[6:1]
--present status and error information about the program, erase, suspend, VPP, and
--block-locked operations.
