Library IEEE;
USE IEEE.std_logic_1164.all;
USE IEEE.std_logic_unsigned.all;

USE work.pkg_081006.all;
USE work.DDS_VME_CONSTANTS.all;

----------------------------------------------------
-- This module updates the 4 us to 1 us interpolation process

-- Version 4 adds opportunity to write the DACs in the LOWER FPGA

entity Freq_Phase_Controller_v4 is
port(
		clk				: in std_logic;
		nreset			: in std_logic;
		Curve_Data		: in std_logic_vector(31 downto 0);
		Transf_Done		: in std_logic;
		
		Run_Curves		: in std_logic; -- Starts the Freq/Phase Curves
		Reset_Curves 	: in std_logic; -- Resets Curve and Re-Initializes
		Rewind_Curves	: in std_logic; -- Resets Curves for next cycle
		Update_Pulse	: in std_logic;
		
		DAC_3_Data		: in std_logic_vector(15 downto 0);
		DAC_4_Data		: in std_logic_vector(15 downto 0);
		
		Rewind_Done		: in std_logic;

		-----------------------------------------------------

		RUN_Addr_to_Mem	: out std_logic_vector(23 downto 0);
		RUN_Data_to_Mem	: out std_logic_vector(31 downto 0);
		RUN_RD_Pulse		: out std_logic;    
		RUN_WR_Pulse		: out std_logic;    
		RUN_Ena		 		: out std_logic;  
		RUN_Mode_Bus_Req	: out std_logic;
		
		Fk0_Data			: out std_logic_vector(31 downto 0);                          
		Fk1_Data			: out std_logic_vector(31 downto 0);                          
		Fk2_Data			: out std_logic_vector(31 downto 0);                          
		PAk1_Data		: out std_logic_vector(15 downto 0);                          
		PAk2_Data		: out std_logic_vector(15 downto 0);                          
		PBk1_Data		: out std_logic_vector(15 downto 0);                          
		PBk2_Data		: out std_logic_vector(15 downto 0);  
		
		Init_Interpolation 		: out std_logic;
		Advance_Sample			: out std_logic;
		Enable_Updates			: out std_logic;
		Rewind					: out std_logic;
		Rewind_Completed		: out std_logic;
		Disable_Feedback		: out std_logic;
		Error_1					: out std_logic                        
);
end Freq_Phase_Controller_v4;

--########################################################
--				A R C H I T E C T U R E
--########################################################
architecture main of Freq_Phase_Controller_v4 is
-- Synchronized Inputs --------------------------------
	signal run		: std_logic;
	signal rst		: std_logic;
	signal rwd		: std_logic;
	signal update	: std_logic;
	
	signal rwd_latch : std_logic;
	signal clr_rwd	 : std_logic;
	
-- Output Signals ---------------------------------------
	signal rd_mem			: std_logic;
	signal wr_mem			: std_logic;
	signal addr_to_mem	: std_logic_vector(23 downto 0);
	signal data_to_mem	: std_logic_vector(31 downto 0);

-- Count 4 each 1us update ticks ------------------------
	signal upd_count		: std_logic_vector(2 downto 0);
	signal rst_upd_count	: std_logic;
	
	signal count_1, count_2 : std_logic;
	signal count_3, count_4 : std_logic;

-- Buffers ----------------------------------------------
	signal Fk0			: std_logic_vector(31 downto 0);
	signal Fk1			: std_logic_vector(31 downto 0);
	signal Fk2			: std_logic_vector(31 downto 0);
	
	signal PAk1			: std_logic_vector(15 downto 0);
	signal PAk2			: std_logic_vector(15 downto 0);

	signal PBk1			: std_logic_vector(15 downto 0);
	signal PBk2			: std_logic_vector(15 downto 0);
	
-- Buffer Latch Signals ------------------------------------
	signal get_fk0		: std_logic;
	signal get_fk1		: std_logic;
	signal get_fk2		: std_logic;

	signal get_pak1		: std_logic;
	signal get_pak2		: std_logic;
	
	signal get_pbk1		: std_logic;
	signal get_pbk2		: std_logic;
	
-- FREQ_CURVE_POINTER signals ---------------------------------------------
-- This counter will act as the address pointer into the freq curve memory.
-- It is used when retrieving frquency curve points.
	signal fcrv_ptr_inc		: std_logic; -- counter increment enable strobe
	signal fcrv_ptr_load	: std_logic;
	signal fcrv_ptr_dir		: std_logic;
	signal fcrv_ptr_step	: std_logic_vector(3 downto 0);
	signal fcrv_ptr_done1	: std_logic; -- count end flag
	signal fcrv_ptr_out		: std_logic_vector(19 downto 0);

-- PHASE_OFFSET1_CURVE_POINTER signals -------------------------------------
-- This counter will act as the address pointer into the Phase Offset 1 
-- curve memory.  
	signal PAcrv_ptr_inc	: std_logic; -- counter increment enable strobe
	signal PAcrv_ptr_load	: std_logic;
	signal PAcrv_ptr_step	: std_logic_vector(3 downto 0);
	signal PAcrv_ptr_done1	: std_logic; -- count end flag
	signal PAcrv_ptr_out	: std_logic_vector(19 downto 0);

-- PHASE_OFFSET2_CURVE_POINTER signals -------------------------------------
-- This counter will act as the address pointer into the Phase Offset 2 
-- curve memory.  
	signal PBcrv_ptr_inc	: std_logic; -- counter increment enable strobe
	signal PBcrv_ptr_load	: std_logic;
	signal PBcrv_ptr_step	: std_logic_vector(3 downto 0);
	signal PBcrv_ptr_done1	: std_logic; -- count end flag
	signal PBcrv_ptr_out	: std_logic_vector(19 downto 0);

-- Data Access Pause Timer ------------------------------------------------
	signal pause_tmr		: std_logic_vector(2 downto 0);
	signal inc_pause_tmr	: std_logic;

-- SETUP_STATE_MACHINE ---------------------------------------
type STATE_TYPE is
(init, init_data1, init_data2, init_data3, ready, wait1, step_curves, 
 get_next_data, clear_count, rewind_crv, last_update, idle, error1);
signal CS, NS : STATE_TYPE;

--======================================================--
-- State Machine for retrieving Freq and Phase from memory
--======================================================--
type STATE_TYPE2 is
(idle2, addr_nxt_freq, get_nxt_freq, addr_nxt_PhA, get_nxt_PhA, 
 addr_nxt_PhB, get_nxt_PhB, done, write_dacs_1, write_dacs_2, write_dacs_3);
signal CS2, NS2 : STATE_TYPE2;

-- State machine handshake signals --------------------------------
	signal get_next			: std_logic;
	signal new_data_ready	: std_logic;
	
--========================================================
--						B E G I N
--========================================================
begin

SYNC_RESET: led 
	port map(sig_in=>Reset_Curves, clk=>clk, clrn=>nreset, pls_out=>rst); 

SYNC_REWIND: led 
	port map(sig_in=>Rewind_Curves, clk=>clk, clrn=>nreset, pls_out=>rwd); 

SYNC_RUN: led 
	port map(sig_in=>Run_Curves, clk=>clk, clrn=>nreset, pls_out=>run); 

SYNC_UPDATE: led 
	port map(sig_in=>Update_Pulse, clk=>clk, clrn=>nreset, pls_out=>update); 

----------------------------------------------------------
-- Latch the rewind command ------------------------------
-- We latch the rewind command so it is hard to miss.
REWIND_LATCH:
process(clk, nreset, rwd, clr_rwd)
begin
	if  (nreset = '0')       then 	rwd_latch <= '0';
	elsif rising_edge(clk) then
		if   (rwd = '1')     then 	rwd_latch <= '1';
		elsif(clr_rwd = '1') then	rwd_latch <= '0';
		end if;
	end if;
end process;
                
----------------------------------------------------------
-- Count Updates -----------------------------------------
-- We count the updates to know when we need to update the
-- frequency and phase curve points for the interpolation
-- process.
UPDATE_COUNTER:
process(clk, nreset, update, rst_upd_count)
begin
	if(nreset='0' OR rst_upd_count = '1')	then
		upd_count	<= (others => '0');	
	elsif(rising_edge(clk))	then
		if (update = '1') then
			upd_count	<= upd_count + 1;
		else
			upd_count	<= upd_count;
		end if;			   		 		
	end if;
end process;

COUNTER_SIGNALS:
process(clk, nreset, upd_count)
begin
	-- default ----------
	count_1 <= '0';
	count_2 <= '0';
	count_3 <= '0';
	count_4 <= '0';
	case upd_count is
		when "001" =>	count_1 <= '1';
		when "010" =>	count_2 <= '1';
		when "011" =>	count_3 <= '1';
		when "100" =>	count_4 <= '1';
		when others =>  NULL;
	end case;
end process;

-- Bus Request Based on Update Count -----------------------------
-- We use the signal RUN_Mode_Bus_Req to reserve the data bus for
-- the updating of the frequency and phase curve values that are
-- loaded into the interpolation process. The interpolation derives
-- 1us points from the 4us points.
BUS_REQUEST:
process(clk, nreset, upd_count)
begin
	if   (nreset = '0')        then RUN_Mode_Bus_Req <= '0';
	elsif rising_edge(clk)     then
		if   (count_1 = '1') 	then RUN_Mode_Bus_Req <= '0';
		elsif(count_2 = '1') 	then RUN_Mode_Bus_Req <= '0';							
		else						 		  RUN_Mode_Bus_Req <= Run_Curves;--v
		end if;				-- request bus only during Run_Curves Interval
	end if;
end process;

-- Fk0 DATA REGISTER (first data point) -----------------------------
Fk0_DATA_REGISTER: dffr -- used to determine Rewind_Done
generic map(32)
port map( 	d    => Fk2, 
			clk  => clk, 
			clrn => nreset, 
			prn  => '1', 
			ena  => get_fk0, 
			q    => Fk0 
		);
		
-- Fk2 DATA REGISTER (newest data point) ---------------------------
Fk2_DATA_REGISTER: dffr
generic map(32)
port map( 	d    => Curve_Data, 
			clk  => clk, 
			clrn => nreset, 
			prn  => '1', 
			ena  => get_fk2, 
			q    => Fk2 
		);
		
-- Fk1 DATA REGISTER (older data point) -----------------------------
Fk1_DATA_REGISTER: dffr
generic map(32)
port map( 	d    => Fk2, 
			clk  => clk, 
			clrn => nreset, 
			prn  => '1', 
			ena  => get_fk1, 
			q    => Fk1 
		);
		
-- PAk2 DATA REGISTER (newest data point) ---------------------------
PAk2_DATA_REGISTER: dffr
generic map(16)
port map( 	d    => Curve_Data(15 downto 0), 
			clk  => clk, 
			clrn => nreset, 
			prn  => '1', 
			ena  => get_pak2, 
			q    => PAk2 
		);
		
-- PAk1 DATA REGISTER (older data point) -----------------------------
PAk1_DATA_REGISTER: dffr
generic map(16)
port map( 	d    => PAk2, 
			clk  => clk, 
			clrn => nreset, 
			prn  => '1', 
			ena  => get_pak1, 
			q    => PAk1 
		);
		
-- PBk2 DATA REGISTER (newest data point) ---------------------------
PBk2_DATA_REGISTER: dffr
generic map(16)
port map( 	d    => Curve_Data(15 downto 0), 
			clk  => clk, 
			clrn => nreset, 
			prn  => '1', 
			ena  => get_pbk2, 
			q    => PBk2 
		);
		
-- PBk1 DATA REGISTER (older data point) -----------------------------
PBk1_DATA_REGISTER: dffr
generic map(16)
port map( 	d    => PBk2, 
			clk  => clk, 
			clrn => nreset, 
			prn  => '1', 
			ena  => get_pbk1, 
			q    => PBk1 
		);
				
-- FREQ_CURVE_ADDR_POINTER_COUNTER -----------------------------
FREQ_CURVE_ADDR_POINTER: countby_n_VER2 -- Address Pointer to Curve Memory
generic map(20,4)
port map(	clk  	=> clk,
			clrn 	 	=> '1',
			ena  	 	=> fcrv_ptr_inc,
			load 		=> fcrv_ptr_load, 
			ld_data  => FREQUENCY_CURVE1_FIRST(19 downto 0), 
			step	 	=> fcrv_ptr_step, --"0100",-- 32 bit numbers
			flag_cnt	=> FREQUENCY_CURVE1_LAST(19 downto 0),
			flag_out	=> fcrv_ptr_done1,
			cnt_out  	=> fcrv_ptr_out 
		);

-- PA_CURVE_ADDR_POINTER_COUNTER -----------------------------
PA_CURVE_ADDR_POINTER: countby_n_VER2 -- Address Pointer to Curve Memory
generic map(20,4)
port map(	clk  	=> clk,
			clrn 	 	=> '1',
			ena  	 	=> PAcrv_ptr_inc,
			load 		=> PAcrv_ptr_load,
			ld_data  => PHASE_CURVE_A_FIRST(19 downto 0),
			step	 	=> PAcrv_ptr_step, --"0010",-- 16 bit numbers
			flag_cnt	=> PHASE_CURVE_A_LAST(19 downto 0),
			flag_out	=> PAcrv_ptr_done1,					
			cnt_out  => PAcrv_ptr_out			           
		);

-- PB_CURVE_ADDR_POINTER_COUNTER -----------------------------
PB_CURVE_ADDR_POINTER: countby_n_VER2 -- Address Pointer to Curve Memory
generic map(20,4)
port map(	clk  	 => clk,
			clrn 	 	=> '1',
			ena  	 	=> PBcrv_ptr_inc,
			load 		=> PBcrv_ptr_load,
			ld_data  => PHASE_CURVE_B_FIRST(19 downto 0),
			step	 	=> PBcrv_ptr_step, --"0010",-- 16 bit numbers
			flag_cnt => PHASE_CURVE_B_LAST(19 downto 0),
			flag_out	=> PBcrv_ptr_done1,					
			cnt_out 	=> PBcrv_ptr_out			           
		);

----------------------------------------------------------
PAUSE_TIMER: -- Memory Access Timer
process (clk, inc_pause_tmr)
begin
	if rising_edge(clk) then
		if   (inc_pause_tmr = '0') then pause_tmr <= "000";
		elsif(inc_pause_tmr = '1') then pause_tmr <= pause_tmr + 1;
		end if;
	end if;
end process;
							
--{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{
--                S T A T E   M A C H I N E
--}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}
SM_REG: process(clk, nreset, rst)
begin
	if(nreset='0') OR (rst = '1') then CS <= init;
	elsif(rising_edge(clk))	then CS <= NS;
	end if;
end process;

SM_NEXT_STATE: 
process
(CS, run, rwd_latch, new_data_ready, upd_count, pause_tmr, Rewind_Done,
  PAcrv_ptr_done1, PBcrv_ptr_done1, fcrv_ptr_done1 , update, count_4 )
begin
	-- defaults ----------------------
	get_fk0			<= '0';
	get_fk1			<= '0';
	get_pak1		<= '0';
	get_pbk1		<= '0';

	get_next		<= '0';
	fcrv_ptr_load	<= '0';
	PAcrv_ptr_load	<= '0';
	PBcrv_ptr_load	<= '0';
    
	fcrv_ptr_inc	<= '0';
	PAcrv_ptr_inc	<= '0';
	PBcrv_ptr_inc	<= '0';
	                   
	fcrv_ptr_step	<= "0100";
	PAcrv_ptr_step	<= "0010";
	PBcrv_ptr_step	<= "0010";
	
	rst_upd_count	<= '0';
	clr_rwd			<= '0';
      
	Init_Interpolation 		<= '0';
	Enable_Updates			<= '0';
	Rewind					<= '0';
	Rewind_Completed		<= '0';
	Disable_Feedback		<= '0';
	Error_1					<= '0';
	
	case CS is
		when init =>								NS<= init_data1;				
			Disable_Feedback		<= '1';			
			-- load first curve addresses into the pointer/counters
			fcrv_ptr_load	<= '1'; --load first freq curve address
			PAcrv_ptr_load	<= '1'; --load first PhA curve address
			PBcrv_ptr_load	<= '1'; --load first PhB curve address
			
			-- retrieve first value of freq and phase words
			get_next		<= '1'; 
				
		when init_data1 =>	        
			Disable_Feedback		<= '1';	
			-- wait for state machine 2 to finish retreiving data		
			if (new_data_ready = '1') then	NS <= init_data2;
			else										NS <= init_data1;
			end if;	-- wait for freq and phase words to be ready
			
		when init_data2 =>						NS <= init_data3;
			Disable_Feedback		<= '1';			
			get_fk0			<= '1'; -- fk0 gets fk2
			get_fk1			<= '1'; -- fk1 gets fk2
			get_pak1		<= '1'; -- pak1 gets pak2
			get_pbk1		<= '1'; -- pbk1 gets pbk2
			
			-- advance mem pointer for next read
			fcrv_ptr_inc	<= not fcrv_ptr_done1; 
			PAcrv_ptr_inc	<= not PAcrv_ptr_done1; 
			PBcrv_ptr_inc	<= not PBcrv_ptr_done1;
			get_next		<= '1';
				--^-- retrieve second value of freq and phase words
		
		when init_data3 =>	
			Disable_Feedback	<= '1';			
			if (new_data_ready = '1') then	NS <= ready;
			else										NS <= init_data3;
			end if;
						
		when ready =>
			Disable_Feedback	<= '1';			
			Init_Interpolation 	<= '1';
			rst_upd_count		<= '1'; -- keep count reset until run
			
			-- Waiting to Run ...
			if   (run = '1') then 			NS <= wait1;
			else									NS <= ready;
			end if;
			
		----------------------------------------------------------
		-- This loop keeps the curve interpolation processes updated
		--=== Begining of the Run Loop ===--
		when wait1 =>			
			Enable_Updates	<= '1';
			if   (rwd_latch = '1') then		NS <= rewind_crv;
			elsif(run       = '1') then		NS <= error1; 
					--^-- Reset_Curves (rst) should happen before this
			elsif(count_4 = '1') then 			NS <= step_curves;
--				if(fcrv_ptr_done1  = '1' AND
--				   PAcrv_ptr_done1 = '1' AND
--				   PBcrv_ptr_done1 = '1') then 	NS <= wait1; --idle;
--				else							
--				end if;
			else                   	     		NS <= wait1;
			end if;
		
		when step_curves =>						NS <= get_next_data;
			Enable_Updates	<= '1';
			get_fk1			<= '1'; -- fk1 gets fk2
			get_pak1		<= '1'; -- pak1 gets pak2
			get_pbk1		<= '1'; -- pbk1 gets pbk2
			
			-- advance mem pointer for next read, if not done ...
			fcrv_ptr_inc	<= not fcrv_ptr_done1; 
			PAcrv_ptr_inc	<= not PAcrv_ptr_done1; 
			PBcrv_ptr_inc	<= not PBcrv_ptr_done1;
			
			get_next		<= '1'; -- signal the other state machine
		
		when get_next_data =>	
			Enable_Updates	<= '1';
			
			if (new_data_ready = '1') then	NS <= clear_count;
			else										NS <= get_next_data;
			end if;
			
		when clear_count =>						NS <= wait1;
			Enable_Updates	<= '1';
			rst_upd_count	<= '1'; -- reset the update counter
						
		--=== End of the Run Loop ===--
		----------------------------------------------------------
		when rewind_crv  =>					
			rst_upd_count		<= '1'; -- keep count reset during update interval
			Disable_Feedback	<= '1';	
			Rewind				<= '1';
			Enable_Updates		<= '1';
			clr_rwd				<= '1';
					
			if   (Rewind_Done = '1') then		NS <= idle; -- last_update;
			else										NS <= rewind_crv;
			end if;
			
		when last_update =>
			rst_upd_count		<= '1'; -- keep count reset during update interval
			Disable_Feedback	<= '1';			
			Enable_Updates		<= '1';
			if (update = '1') then				NS <= idle;
			else										NS <= last_update;
			end if;
			
		----------------------------------------------------------
		when idle =>						NS <= idle;
			-- Here we will wait for this state machine to be reset
			-- with the Reset_Curves input signal from the DDS_App_Sequencer
			rst_upd_count		<= '1'; -- keep count reset during update interval
			Disable_Feedback	<= '1';
			Rewind_Completed	<= '1'; -- signal sequencer that rewind is complete
						
--			if   (rwd_latch = '1') then		NS <= rewind_crv;
--			elsif(run = '1') then			NS <= init; 
--					--^-- Reset_Curves (rst) should happen before this
--			else							NS <= idle;
--			end if;
			
		when error1 =>						NS <= error1;
			Disable_Feedback	<= '1';
			Error_1				<= '1'; -- Rewind was not requested before next cycle
			
		-----------------------------------------------
		when others => 						NS <= idle;
	end case;
end process;

--======================================================--
-- State Machine for retrieving Freq and Phase from memory
--======================================================--
SM_REG2: process(clk, nreset)
begin
	if(nreset='0')			then CS2 <= idle2;
	elsif(rising_edge(clk))	then CS2 <= NS2;
	end if;
end process;

SM_NEXT_STATE2: 
process
(CS2, get_next, fcrv_ptr_out, PAcrv_ptr_out, PBcrv_ptr_out, 
 pause_tmr, Transf_Done, DAC_4_Data, DAC_3_Data)
begin
	-- defaults ----------------------
	RUN_Ena			<= '1';
	rd_mem			<= '0';
	wr_mem			<= '0';
	                   
	inc_pause_tmr	<= '0';
	addr_to_mem <= "0000" & fcrv_ptr_out(19 downto 0);
	data_to_mem <= DAC_4_Data & DAC_3_Data;
	
	get_fk2			<= '0';	
	get_pak2		<= '0';	        
	get_pbk2		<= '0';
	
	new_data_ready  <= '0';
	
	case CS2 is
		when idle2 =>										
			RUN_Ena			<= '0';
			if (get_next = '1') then		NS2 <= addr_nxt_freq;
			else									NS2 <= idle2;
			end if;

		when addr_nxt_freq =>
			-- Initial Frequency retrieved ---------------------
			addr_to_mem <= "0000" & fcrv_ptr_out(19 downto 0);
			rd_mem			<= '1';
			-- increment the pause timer and wait for timeout
			inc_pause_tmr	<= '1'; 			
			if   (fcrv_ptr_out < X"20000") AND (pause_tmr = "011")
											 then		NS2 <= get_nxt_freq;
			elsif(Transf_Done = '1') then 	NS2 <= get_nxt_freq;
			else										NS2 <= addr_nxt_freq;
			end if;
		
		when get_nxt_freq =>				NS2 <= addr_nxt_PhA;
			addr_to_mem <= "0000" & fcrv_ptr_out(19 downto 0);
			get_fk2			<= '1';
			
		when addr_nxt_PhA =>
			-- Initial Frequency retrieved ---------------------
			addr_to_mem <= "0000" & PAcrv_ptr_out(19 downto 0);
			rd_mem			<= '1';
			-- increment the pause timer and wait for timeout
			inc_pause_tmr	<= '1'; 			
			if   (PAcrv_ptr_out < X"20000") AND (pause_tmr = "011")
									 then	NS2 <= get_nxt_PhA;
			elsif(Transf_Done = '1') then 	NS2 <= get_nxt_PhA;
			else							NS2 <= addr_nxt_PhA;
			end if;
		
		when get_nxt_PhA =>					NS2 <= addr_nxt_PhB;
			addr_to_mem <= "0000" & PAcrv_ptr_out(19 downto 0);
			get_pak2		<= '1';
			
		when addr_nxt_PhB =>
			-- Initial Frequency retrieved ---------------------
			addr_to_mem <= "0000" & PBcrv_ptr_out(19 downto 0);
			rd_mem			<= '1';
			-- increment the pause timer and wait for timeout
			inc_pause_tmr	<= '1'; 			
			if   (PBcrv_ptr_out < X"20000") AND (pause_tmr = "011")
												then	NS2 <= get_nxt_PhB;
			elsif(Transf_Done = '1') then 	NS2 <= get_nxt_PhB;
			else										NS2 <= addr_nxt_PhB;
			end if;
		
		when get_nxt_PhB =>						NS2 <= write_dacs_1;
			addr_to_mem <= "0000" & PBcrv_ptr_out(19 downto 0);
			get_pbk2		<= '1';
			
		when done =>								NS2 <= idle2;
			new_data_ready  <= '1';
			
		-- two 16 bit register writes in one 32 bit transfer
		when write_dacs_1 => 					NS2 <= write_dacs_2;
			addr_to_mem 	<= DAC_3_4_REGISTER;
			data_to_mem 	<= DAC_4_Data & DAC_3_Data;
			
		when write_dacs_2 => 					NS2 <= write_dacs_3;
			addr_to_mem 	<= DAC_3_4_REGISTER;
			data_to_mem 	<= DAC_4_Data & DAC_3_Data;
			wr_mem			<= '1';

		when write_dacs_3 => 
			addr_to_mem 	<= DAC_3_4_REGISTER;
			data_to_mem 	<= DAC_4_Data & DAC_3_Data;
			
			if   (Transf_Done = '1') then 	NS2 <= done;
			else										NS2 <= write_dacs_3;
			end if;
	
		when others => 						NS2 <= idle2;
	end case;
end process;

process (rd_mem, addr_to_mem)
begin
	if   (addr_to_mem < X"20000") then	RUN_RD_Pulse 	<= '0'; 
	else											RUN_RD_Pulse 	<= rd_mem;
	end if;
end process;

	RUN_Addr_to_Mem 	<= addr_to_mem;
	RUN_Data_to_Mem	<= data_to_mem;
	RUN_WR_Pulse		<= wr_mem;
	
	Fk0_Data		<= Fk0;  -- used to determine Rewind_Done                        
	Fk1_Data		<= Fk1;                          
	Fk2_Data		<= Fk2;                          
	PAk1_Data		<= PAk1;	                          
	PAk2_Data		<= PAk2;                          
	PBk1_Data		<= PBk1;	                          
	PBk2_Data		<= PBk2;	
	
	Advance_Sample 	<= count_4;                          
                 
end;             