Library IEEE;
USE IEEE.std_logic_1164.all;
USE IEEE.std_logic_unsigned.all;

-- My Definitions --------------------------------
USE work.pkg_081006.all; 

-------------------------------------------------------------------------------
-- MODULE DESCRIPTION, 
-- First Version: April 12, 2010
-- This logic transfers 32 bits of data and 20 bits of address between the
-- Upper fpga and the Lower fpga.  The data and address lines are serialized
-- so that it requires 4 clocks to transfer the 32 bit data using 8 serial 
-- streams, and 4 clocks to transfer the 20 bit address using 5 serial streams.  
-- The state machine in this applications runs at the clock rate (50MHz), but the
-- shift register clock runs at two times this (100MHz).  At these rates reads and 
-- writes can take place at 100ns intervals.
--
-- The process is controlled from the Upper fpga.  When data is written from
-- the Upper to the Lower, the address and data are transmitted at the same
-- time, in the same cycle.  In order to read data from the Lower fpga, the Upper
-- transmits the address in the first cycle and the next cycle will return the 
-- data from the Lower fpga that was addressed in the previous cycle.

-- The signal lines used to link the Upper and Lower fpga's are the following.
--
--  Signal				Lines	Direction		Description
--  -----------------  ------   --------------	-----------------------
--	Data Bus[7..0]			8		Bi-Directional	The data being read or written

--  Addr Bus[4..0]		5		Upper to Lower	Memory address
--  WR_Lower_mem			1		Upper to Lower	Write pulse to Lower RAM
--	 RD_Lower_mem (OE)	1		Upper to Lower	Databus tri-state ena for Lower
--  LD_Lower_mem			1		Upper to Lower	Data SR Load parallel signal	 
--  RD_Lower_SR			1		Upper to Lower	Clk gate reset & SR latch signal
--  ENA_shift_out			1		Upper to Lower	Initiates SR shifting in Lower
--  RCV_CLK_OUT			1		Upper to Lower	Clk for Lower Rcv SR

--  Rcv_Clk_Lower			1		Lower to Upper	Clk for Upper Rcv SR

------ additional inter-fpga lines are used for the VME interface----

--	VME_A[25..24]					2			Lower to Upper
--  Extended VME Board Select		1			Lower to Upper

-- Currently 23 lines total (May 7, 2010)
-- Note:  WR, write signals are generally used as DFF register enables.
--			Data is loaded when WR is high and latched when it goes low.
-------------------------------------------------------------------------------

-- Version 4:  Added another wait state when reading the Lower FPGA, by adding
--             the state "rd_pause2".

entity InterFPGA_Master_3Port_v4 is 
	port( 
			clk				: in std_logic;
			nreset			: in std_logic;
			XMT_SR_SYNC		: in std_logic;
			XMT_SR_CLK		: in std_logic;
			RCV_SR_SYNC		: in std_logic;
			RCV_SR_CLK		: in std_logic;
			
			Serial_D_in		: in std_logic_vector(7 downto 0);			
			RCV_CLK_IN		: in std_logic;
			
			VME_Data_to_Lower	: in std_logic_vector(31 downto 0);
			VME_Addr_to_Lower	: in std_logic_vector(19 downto 0);
			VME_RD_Lower		: in std_logic;
			VME_WR_Lower		: in std_logic;

			UPD_Data_to_Lower	: in std_logic_vector(31 downto 0);
			UPD_Addr_to_Lower	: in std_logic_vector(19 downto 0);
			UPD_RD_Lower		: in std_logic;
			UPD_WR_Lower		: in std_logic;
			
			Permit_VME			: in std_logic;
			--------------------------------------------------------			
			VME_Data_from_Lower	: out std_logic_vector(31 downto 0);
			DDS_Data_from_Lower	: out std_logic_vector(31 downto 0);
			
			Serial_D_out		: out std_logic_vector(7 downto 0);
			Serial_A_out		: out std_logic_vector(4 downto 0);
			
			WR_lower_mem		: out std_logic;
			RD_lower_mem		: out std_logic;
			LD_lower_mem		: out std_logic;
			RD_lower_sr			: out std_logic;
			ENA_shift_out		: out std_logic;
			
			RCV_CLK_OUT			: out std_logic;
			
			XMT_CLK_TEST    	: out std_logic;
			
			VME_Transfer_Done 	: out std_logic;
			UPD_Transfer_Done 	: out std_logic;
			
			test_of_mode		: out std_logic;
			intf_ready			: out std_logic
		);
	end InterFPGA_Master_3Port_v4 ;
	
architecture main of InterFPGA_Master_3Port_v4  is	

signal Data_B_to_Lower	: std_logic_vector(31 downto 0);
signal Addr_B_to_Lower	: std_logic_vector(19 downto 0);
signal RD_Lower_B		: std_logic;
signal WR_Lower_B		: std_logic;
 
-- Mode Latch -----------------------------------------------------------
signal get_mode			: std_logic;

-- Previous Address Latch -----------------------------------------------
signal prev_Addr		: std_logic_vector(18 downto 0);
signal get_new_addr		: std_logic;

-- Shift Register Signals -----------------------------------------------
signal D_XMT_SR_0 		: std_logic_vector(3 downto 0); 
signal D_XMT_SR_1 		: std_logic_vector(3 downto 0); 
signal D_XMT_SR_2 		: std_logic_vector(3 downto 0); 
signal D_XMT_SR_3 		: std_logic_vector(3 downto 0); 
signal D_XMT_SR_4 		: std_logic_vector(3 downto 0); 
signal D_XMT_SR_5 		: std_logic_vector(3 downto 0); 
signal D_XMT_SR_6 		: std_logic_vector(3 downto 0); 
signal D_XMT_SR_7 		: std_logic_vector(3 downto 0); 

signal D_RCV_SR_0 		: std_logic_vector(3 downto 0);
signal D_RCV_SR_1 		: std_logic_vector(3 downto 0);
signal D_RCV_SR_2 		: std_logic_vector(3 downto 0);
signal D_RCV_SR_3 		: std_logic_vector(3 downto 0);
signal D_RCV_SR_4 		: std_logic_vector(3 downto 0);
signal D_RCV_SR_5 		: std_logic_vector(3 downto 0);
signal D_RCV_SR_6 		: std_logic_vector(3 downto 0);
signal D_RCV_SR_7 		: std_logic_vector(3 downto 0);
                

signal A_XMT_SR_0 		: std_logic_vector(3 downto 0); 
signal A_XMT_SR_1 		: std_logic_vector(3 downto 0); 
signal A_XMT_SR_2 		: std_logic_vector(3 downto 0); 
signal A_XMT_SR_3 		: std_logic_vector(3 downto 0); 
signal A_XMT_SR_4 		: std_logic_vector(3 downto 0); 

-- Shift Register Control Signals ----------------------------------------
signal XMT_CLK		: std_logic;

signal nrst_shift	: std_logic;
signal ena_shift	: std_logic;
signal load_sr		: std_logic;
signal load			: std_logic;
signal read_sr		: std_logic;
signal wr			: std_logic;
signal rd			: std_logic;

-- Input Syncronization --------------------------------------------------
signal RD_A			: std_logic;
signal WR_A			: std_logic;
signal RD_led_A		: std_logic_vector(1 downto 0);
signal WR_led_A		: std_logic_vector(1 downto 0);
signal RD_cmd_A		: std_logic;
signal WR_cmd_A		: std_logic;
signal ltchd_RD_A	: std_logic;
signal ltchd_WR_A	: std_logic;
                
signal RD_B			: std_logic;
signal WR_B			: std_logic;
signal RD_led_B		: std_logic_vector(1 downto 0);
signal WR_led_B		: std_logic_vector(1 downto 0);
signal RD_cmd_B		: std_logic;
signal WR_cmd_B		: std_logic;
signal ltchd_RD_B	: std_logic;
signal ltchd_WR_B	: std_logic;
 
signal latched_mode: std_logic;
signal dds_nvme_mode: std_logic;

signal end_proc		: std_logic;
signal end_vme		: std_logic;
signal end_dds		: std_logic;

-- Shift Clock Counters --------------------------------------------------
signal shift_cnt	: std_logic_vector(4 downto 0);
signal rshift_cnt	: std_logic_vector(4 downto 0);

signal cnt_bit_2	: std_logic;

-- SETUP_STATE_MACHINE ---------------------------------------------------
type STATE_TYPE is
(init, idle, chk_mode_vme, chk_mode_dds, addr_pause, addr_shift, addr_wait, 
 addr_reg, setup_data, rd_pause, rd_pause2, rd_shift, rd_wait, rd_reg, wr_pause, 
 wr_shift, wr_wait, wr_reg);
	  
signal CS, NS : STATE_TYPE;
			
--========================================================================
--========================================================================
--                    B E G I N 
--========================================================================
--========================================================================	
begin
	----------------------------------------------------------------------
	-- This process produces four transmitter shift register clock pulses 
	-- for the transmitter shift register in this FPGA.  The pulses are four 
	-- XMT_SR_CLK clock periods gated by a 4 clock wide LED type pulse 
	-- syncronized by XMT_SR_SYNC.  The trigger for all of this is ena_shift.
	--
	-- Signal ena_shift is produced by the state machine which is clocked by
	-- Logic_clk (clk).  
	--
	-- Four receiver shift register clock pulses are generated in a like manner.
	-- These pulses are used at the other FPGA's receive shift register.  
	--
	-- Receiver shift register pulses are generated in the OTHER FPGA for the 
	-- receiver shift register in THIS FPGA.
	--
	-- The state machine clock and XMT_SR_SYNC, XMT_SR_CLK, RCV_SR_SYNC, RCV_SR_CLK 
	-- are all generated by the same PLL and can be shifted in phase w.r.t. one 
	-- another.  The XMT_SR_SYNC clock is shifted w.r.t. the state machine clock
	-- so that the active high state of the ena_shift signal is registered by the 
	-- rising edge of RCV_SR_SYNC before the rising edge of XMT_SR_SYNC.  The
	-- sync clocks determine the timing of the transmit clock gate and the 
	-- receive clock gate.  The RCV_SR_CLK is delayed w.r.t. the XMT_SR_CLK
	-- so as to optimize the setup and hold times for the data arriving at the
	-- receiver shift register from the transmitter shift register.  Then finally
	-- XMT_SR_CLK and RCV_SR_CLK can be phased shifted w.r.t. their associated
	-- gating signals so as not to clip either the first or last of the four
	-- shift register clock pulses.
--------------------------------------------------------------------------------
--======================================================================
-- Syncronize inputs that start the processes.  Latch the occurance of 
-- a Read or Write command.  
--
-- It is assumed that the process that issued a Read or Write command will be 
-- waiting for a Transfer_Done signal from this module before proceeding further.  
-- The signaling process will be held-off if an inter-fpga transfer is in progress 
-- or a specific permit control signal is not active.  Rule for inter-fpga bus
-- access are summarized below.
--
-- ACCESS RULES (sharing the interFPGA bus)
-- 1. The VME process and the Update process will share the inter-fpga bus.  
--    Each process has many clock cycles on either side of a Lower fpga 
--    access so that pending (latched) Read or Write commands from the process 
--    not being serviced will get in.
--
-- 3. In Run Mode, the VME process must be held off prior to the "scheduled"
--    access of the inter-fpga bus by the Run process.  Scheduling of the
--    Run processes is controlled by the "Process_Timer.vhd" which generates
--    the VME_Permit signal.  The VME_Permit signal will be active for a short
--    interval at the beginning of each 1us DDS Update cycle.


-- ==== VME Read and Write Commands ====================================
VME_COMMAND_LED: -- detects the signals leading edge
	process (clk, nreset, RD_led_A, WR_led_A, VME_RD_Lower,
			 VME_WR_Lower)
	begin -- Leading Edge Detector (LED)
		if(nreset = '0')  then 
			RD_led_A <= "00";
			WR_led_A <= "00";
			
		elsif rising_edge(clk) then
				RD_led_A(0) <= VME_RD_Lower;
				RD_led_A(1) <= RD_led_A(0);

				WR_led_A(0) <= VME_WR_Lower;
				WR_led_A(1) <= WR_led_A(0);
		end if;
		
		--RD_cmd_A <= RD_led_A(0) AND NOT RD_led_A(1); -- Changed July 9,2010
		--WR_cmd_A <= WR_led_A(0) AND NOT WR_led_A(1);
		RD_cmd_A <= VME_RD_Lower AND NOT RD_led_A(1);
		WR_cmd_A <= VME_WR_Lower AND NOT WR_led_A(1);
		
	end process;

VME_COMMAND_LATCH:	-- latches the occurance of the detected leading edge
					-- until the command has been serviced.  It is assumed
					-- the process issueing the command will WAIT for it to
					-- be completed before issueing another.  Separate 
					-- Transfer_Done signals are generated for the three
					-- possible command issueing processes.
	process (clk, nreset, RD_cmd_A, WR_cmd_A, end_vme)
	begin -- Latch
		if ((nreset = '0') OR (end_vme = '1')) then
			ltchd_RD_A <= '0';
			ltchd_WR_A <= '0';
		elsif rising_edge(clk) then
			if(RD_cmd_A = '1') then ltchd_RD_A <= '1'; end if;
			if(WR_cmd_A = '1') then ltchd_WR_A <= '1'; end if;
		end if;
	end process;
	
	-----------------------------------------------------------
	RD_A 		<= '1' 
		when (RD_cmd_A = '1' OR ltchd_RD_A = '1') AND Permit_VME = '1' 
		else '0';
	-----------------------------------------------------------
	WR_A 		<= '1' 
		when (WR_cmd_A = '1' OR ltchd_WR_A = '1') AND Permit_VME = '1' 
		else '0';
	-----------------------------------------------------------
	end_vme 	<= '1' 
		when latched_mode = '0' and end_proc = '1' 
		else '0';
	-----------------------------------------------------------
	
--------------------------------------------------------------------------------
-- ==== UPD or DDS Read and Write Commands ==========================
	Data_B_to_Lower	<= UPD_Data_to_Lower;	
	Addr_B_to_Lower	<= UPD_Addr_to_Lower;	
	RD_Lower_B		<= UPD_RD_Lower;		
	WR_Lower_B		<= UPD_WR_Lower;		

--------------------------------------------------------------------------------		
DDS_COMMAND_LED:
	process (clk, nreset, RD_led_B, WR_led_B, RD_Lower_B, WR_Lower_B)
	begin
		if(nreset = '0')  then 
			RD_led_B <= "00";
			WR_led_B <= "00";
			
		elsif rising_edge(clk) then
				RD_led_B(0) <= RD_Lower_B;
				RD_led_B(1) <= RD_led_B(0);

				WR_led_B(0) <= WR_Lower_B;
				WR_led_B(1) <= WR_led_B(0);
		end if;
		--RD_cmd_B <= RD_led_B(0) AND NOT RD_led_B(1); --Changed July 9, 2010
		--WR_cmd_B <= WR_led_B(0) AND NOT WR_led_B(1);
		RD_cmd_B <= RD_Lower_B AND NOT RD_led_B(1);
		WR_cmd_B <= WR_Lower_B AND NOT WR_led_B(1);
	end process;

DDS_COMMAND_LATCH:	
	process (clk, nreset, RD_cmd_B, WR_cmd_B, end_dds)
	begin
		if ((nreset = '0') OR (end_dds = '1')) then
			ltchd_RD_B <= '0';
			ltchd_WR_B <= '0';
		elsif rising_edge(clk) then
			if(RD_cmd_B = '1') then ltchd_RD_B <= '1'; end if;
			if(WR_cmd_B = '1') then ltchd_WR_B <= '1'; end if;
		end if;
	end process;
	
	-----------------------------------------------------------
	RD_B 		<= '1' 
		when (RD_cmd_B = '1' OR ltchd_RD_B = '1')
		else '0';
	-----------------------------------------------------------
	WR_B 		<= '1' 
		when (WR_cmd_B = '1' OR ltchd_WR_B = '1')
		else '0';
	-----------------------------------------------------------
	end_dds <= '1' 
		when latched_mode = '1' AND end_proc = '1' 
		else '0';
	-----------------------------------------------------------

------------------------------------------------------------------------------
--==== Syncronize SR Load signal and implement a count of the shifts =========
SYNC_LOAD_AND_XMT_SHIFT_COUNT:	
	process (nrst_shift, XMT_SR_SYNC, XMT_SR_CLK, ena_shift, load_sr)
	begin
		if (nrst_shift = '0')  then 
			shift_cnt(4 downto 0) <= "00000";
			load <= '0';
		elsif rising_edge(XMT_SR_SYNC) then
			if(ena_shift = '1') then
				shift_cnt(0)          <= '1';
				shift_cnt(4 downto 1) <= shift_cnt(3 downto 0);
			end if;

			load <= load_sr; --sync load_sr to XMT_SR_SYNC
		end if;
	end process;
		-- create four gated sr clock pulses for the transmitter in this fpga
		XMT_CLK <= XMT_SR_CLK AND ((shift_cnt(0) AND NOT shift_cnt(4)) OR load);	
		
		XMT_CLK_TEST <= XMT_CLK;	
-------------------------------------------------------------------------------	
RECEIVE_SR_SHIFT_COUNT:
	process (nrst_shift, RCV_SR_SYNC, RCV_SR_CLK, ena_shift)
	begin
		if (nrst_shift = '0') then 
			rshift_cnt(4 downto 0) <= "00000";
		elsif rising_edge(RCV_SR_SYNC) then
			if(ena_shift = '1') then
				rshift_cnt(0)          <= '1';
				rshift_cnt(4 downto 1) <= rshift_cnt(3 downto 0);
			end if;
		end if;
	end process;
		-- create four gated sr clock pulses for the receiver in the other fpga
		RCV_CLK_OUT <= RCV_SR_CLK AND (rshift_cnt(0) AND NOT rshift_cnt(4));		

SYNCHRONIZE_RSHIFT_CNT_BIT:
		process(clk, nreset,rshift_cnt)
		begin
			if   (nreset = '0') then cnt_bit_2 <= '0';
			elsif rising_edge(clk) then cnt_bit_2 <= rshift_cnt(2);
			end if;
		end process;
		
-------------------------------------------------------------------------------
-- Eight Data Transmit Shift Registers
	process
	(XMT_CLK, nreset, latched_mode, D_XMT_SR_0, D_XMT_SR_1, D_XMT_SR_2, D_XMT_SR_3, 
		D_XMT_SR_4, D_XMT_SR_5, D_XMT_SR_6, D_XMT_SR_7, VME_Data_to_Lower)
	begin
		if(nreset = '0')  then 
				D_XMT_SR_0 <= "0000";
				D_XMT_SR_1 <= "0000";
				D_XMT_SR_2 <= "0000";
				D_XMT_SR_3 <= "0000";
				D_XMT_SR_4 <= "0000";
				D_XMT_SR_5 <= "0000";
				D_XMT_SR_6 <= "0000";
				D_XMT_SR_7 <= "0000";
				
		elsif rising_edge(XMT_CLK) then
			if (load = '1') then
				if (latched_mode = '0') then
					D_XMT_SR_0(3 downto 0) <= VME_Data_to_Lower(3 downto 0);
					D_XMT_SR_1(3 downto 0) <= VME_Data_to_Lower(7 downto 4);
					D_XMT_SR_2(3 downto 0) <= VME_Data_to_Lower(11 downto 8);
					D_XMT_SR_3(3 downto 0) <= VME_Data_to_Lower(15 downto 12);
					D_XMT_SR_4(3 downto 0) <= VME_Data_to_Lower(19 downto 16);
					D_XMT_SR_5(3 downto 0) <= VME_Data_to_Lower(23 downto 20);
					D_XMT_SR_6(3 downto 0) <= VME_Data_to_Lower(27 downto 24);
					D_XMT_SR_7(3 downto 0) <= VME_Data_to_Lower(31 downto 28);
				else
					D_XMT_SR_0(3 downto 0) <= Data_B_to_Lower(3 downto 0);
					D_XMT_SR_1(3 downto 0) <= Data_B_to_Lower(7 downto 4);
					D_XMT_SR_2(3 downto 0) <= Data_B_to_Lower(11 downto 8);
					D_XMT_SR_3(3 downto 0) <= Data_B_to_Lower(15 downto 12);
					D_XMT_SR_4(3 downto 0) <= Data_B_to_Lower(19 downto 16);
					D_XMT_SR_5(3 downto 0) <= Data_B_to_Lower(23 downto 20);
					D_XMT_SR_6(3 downto 0) <= Data_B_to_Lower(27 downto 24);
					D_XMT_SR_7(3 downto 0) <= Data_B_to_Lower(31 downto 28);				               
				end if;
				               
			else				
				D_XMT_SR_0(3 downto 1) <= D_XMT_SR_0(2 downto 0);
				D_XMT_SR_1(3 downto 1) <= D_XMT_SR_1(2 downto 0);
				D_XMT_SR_2(3 downto 1) <= D_XMT_SR_2(2 downto 0);
				D_XMT_SR_3(3 downto 1) <= D_XMT_SR_3(2 downto 0);
				D_XMT_SR_4(3 downto 1) <= D_XMT_SR_4(2 downto 0);
				D_XMT_SR_5(3 downto 1) <= D_XMT_SR_5(2 downto 0);
				D_XMT_SR_6(3 downto 1) <= D_XMT_SR_6(2 downto 0);
				D_XMT_SR_7(3 downto 1) <= D_XMT_SR_7(2 downto 0);
				
				D_XMT_SR_0(0) <= D_XMT_SR_0(3); -- Data wraps around ...
				D_XMT_SR_1(0) <= D_XMT_SR_1(3);
				D_XMT_SR_2(0) <= D_XMT_SR_2(3);
				D_XMT_SR_3(0) <= D_XMT_SR_3(3);
				D_XMT_SR_4(0) <= D_XMT_SR_4(3);
				D_XMT_SR_5(0) <= D_XMT_SR_5(3);
				D_XMT_SR_6(0) <= D_XMT_SR_6(3);
				D_XMT_SR_7(0) <= D_XMT_SR_7(3);
			end if;             
		end if;
	end process;	
				Serial_D_out(0) <= D_XMT_SR_0(3); -- serial data outputs
				Serial_D_out(1) <= D_XMT_SR_1(3);
				Serial_D_out(2) <= D_XMT_SR_2(3);
				Serial_D_out(3) <= D_XMT_SR_3(3);
				Serial_D_out(4) <= D_XMT_SR_4(3);
				Serial_D_out(5) <= D_XMT_SR_5(3);
				Serial_D_out(6) <= D_XMT_SR_6(3);
				Serial_D_out(7) <= D_XMT_SR_7(3);

-------------------------------------------------------------------------
-- Eight Data Receiver Shift Register

	process
	(RCV_CLK_IN, nreset, D_RCV_SR_0, D_RCV_SR_1, D_RCV_SR_2, D_RCV_SR_3, 
		D_RCV_SR_4, D_RCV_SR_5, D_RCV_SR_6, D_RCV_SR_7, Serial_D_in)
	begin
		if(nreset = '0')  then 
			D_RCV_SR_0 <= "0000";
			D_RCV_SR_1 <= "0000";
			D_RCV_SR_2 <= "0000";
			D_RCV_SR_3 <= "0000";
			D_RCV_SR_4 <= "0000";
			D_RCV_SR_5 <= "0000";
			D_RCV_SR_6 <= "0000";
			D_RCV_SR_7 <= "0000";
				
		elsif rising_edge(RCV_CLK_IN) then
			D_RCV_SR_0(0) <= Serial_D_in(0); -- serial data in
			D_RCV_SR_1(0) <= Serial_D_in(1);
			D_RCV_SR_2(0) <= Serial_D_in(2);
			D_RCV_SR_3(0) <= Serial_D_in(3);
			D_RCV_SR_4(0) <= Serial_D_in(4);
			D_RCV_SR_5(0) <= Serial_D_in(5);
			D_RCV_SR_6(0) <= Serial_D_in(6);
			D_RCV_SR_7(0) <= Serial_D_in(7);
			                
			D_RCV_SR_0(3 downto 1) <= D_RCV_SR_0(2 downto 0);
			D_RCV_SR_1(3 downto 1) <= D_RCV_SR_1(2 downto 0);
			D_RCV_SR_2(3 downto 1) <= D_RCV_SR_2(2 downto 0);
			D_RCV_SR_3(3 downto 1) <= D_RCV_SR_3(2 downto 0);
			D_RCV_SR_4(3 downto 1) <= D_RCV_SR_4(2 downto 0);
			D_RCV_SR_5(3 downto 1) <= D_RCV_SR_5(2 downto 0);
			D_RCV_SR_6(3 downto 1) <= D_RCV_SR_6(2 downto 0);
			D_RCV_SR_7(3 downto 1) <= D_RCV_SR_7(2 downto 0);
			
		end if;
	end process;	
					
-------------------------------------------------------------------------
-- Parallel Data Out of the Data Receiver Shift Registers					  
	process (clk, read_sr, latched_mode, D_RCV_SR_0, D_RCV_SR_1, D_RCV_SR_2, 
			 D_RCV_SR_3, D_RCV_SR_4, D_RCV_SR_5, D_RCV_SR_6, D_RCV_SR_7)
	begin
		if rising_edge(clk) then
			if (read_sr = '1') then
				if(latched_mode = '0') then
					VME_Data_from_Lower(3 downto 0)   <= D_RCV_SR_0(3 downto 0);
					VME_Data_from_Lower(7 downto 4)   <= D_RCV_SR_1(3 downto 0);
					VME_Data_from_Lower(11 downto 8)  <= D_RCV_SR_2(3 downto 0);
					VME_Data_from_Lower(15 downto 12) <= D_RCV_SR_3(3 downto 0);
					VME_Data_from_Lower(19 downto 16) <= D_RCV_SR_4(3 downto 0);
					VME_Data_from_Lower(23 downto 20) <= D_RCV_SR_5(3 downto 0);
					VME_Data_from_Lower(27 downto 24) <= D_RCV_SR_6(3 downto 0);
					VME_Data_from_Lower(31 downto 28) <= D_RCV_SR_7(3 downto 0);
				else
					DDS_Data_from_Lower(3 downto 0)   <= D_RCV_SR_0(3 downto 0);
					DDS_Data_from_Lower(7 downto 4)   <= D_RCV_SR_1(3 downto 0);
					DDS_Data_from_Lower(11 downto 8)  <= D_RCV_SR_2(3 downto 0);
					DDS_Data_from_Lower(15 downto 12) <= D_RCV_SR_3(3 downto 0);
					DDS_Data_from_Lower(19 downto 16) <= D_RCV_SR_4(3 downto 0);
					DDS_Data_from_Lower(23 downto 20) <= D_RCV_SR_5(3 downto 0);
					DDS_Data_from_Lower(27 downto 24) <= D_RCV_SR_6(3 downto 0);
					DDS_Data_from_Lower(31 downto 28) <= D_RCV_SR_7(3 downto 0);
				end if;
				
			end if; 
		end if;		                   
	end process;	
	
--=======================================================================
-------------------------------------------------------------------------
-- Five Address Transmit Shift Registers
	process
	(XMT_CLK, nreset, load, A_XMT_SR_0, A_XMT_SR_1, A_XMT_SR_2, 
		A_XMT_SR_3, A_XMT_SR_4, VME_Addr_to_Lower)
	begin
		if(nreset = '0')  then 
			A_XMT_SR_0 <= "0000";
			A_XMT_SR_1 <= "0000";
			A_XMT_SR_2 <= "0000";
			A_XMT_SR_3 <= "0000";
			A_XMT_SR_4 <= "0000";
		elsif rising_edge(XMT_CLK) then
			if (load = '1') then
				if (latched_mode = '0') then
					A_XMT_SR_0 <= VME_Addr_to_Lower(3 downto 0);
					A_XMT_SR_1 <= VME_Addr_to_Lower(7 downto 4);
					A_XMT_SR_2 <= VME_Addr_to_Lower(11 downto 8);
					A_XMT_SR_3 <= VME_Addr_to_Lower(15 downto 12);
					A_XMT_SR_4 <= VME_Addr_to_Lower(19 downto 16);
				else
					A_XMT_SR_0 <= Addr_B_to_Lower(3 downto 0);
					A_XMT_SR_1 <= Addr_B_to_Lower(7 downto 4);
					A_XMT_SR_2 <= Addr_B_to_Lower(11 downto 8);
					A_XMT_SR_3 <= Addr_B_to_Lower(15 downto 12);
					A_XMT_SR_4 <= Addr_B_to_Lower(19 downto 16);
				end if;
				
			else
				A_XMT_SR_0(3 downto 1) <= A_XMT_SR_0(2 downto 0);
				A_XMT_SR_1(3 downto 1) <= A_XMT_SR_1(2 downto 0);
				A_XMT_SR_2(3 downto 1) <= A_XMT_SR_2(2 downto 0);
				A_XMT_SR_3(3 downto 1) <= A_XMT_SR_3(2 downto 0);
				A_XMT_SR_4(3 downto 1) <= A_XMT_SR_4(2 downto 0);
				                    
				A_XMT_SR_0(0) <= A_XMT_SR_0(3); -- address wraps around
				A_XMT_SR_1(0) <= A_XMT_SR_1(3);
				A_XMT_SR_2(0) <= A_XMT_SR_2(3);
				A_XMT_SR_3(0) <= A_XMT_SR_3(3);
				A_XMT_SR_4(0) <= A_XMT_SR_4(3);			
			end if;
		end if;			
	end process;
-------------------------------------------------------------------------	
-- Serialized Address Out				  
				Serial_A_out(0) <= A_XMT_SR_0(3);
				Serial_A_out(1) <= A_XMT_SR_1(3);
				Serial_A_out(2) <= A_XMT_SR_2(3);
				Serial_A_out(3) <= A_XMT_SR_3(3);
				Serial_A_out(4) <= A_XMT_SR_4(3);
	
--======================================================================
-- Mode Latch -----------------------------------------------------------
	process (clk, nreset, get_mode, dds_nvme_mode)
	begin
		if (nreset = '0') 		then 
			latched_mode <= '1'; -- default to DDS mode
			
		elsif rising_edge(clk) 	then
			if (get_mode = '1') then 
				latched_mode <= dds_nvme_mode; -- depends on who made a request
			end if;								-- dds_nvme_mode is assigned in state mach.
		end if;	
	end process;
		
--======================================================================
-- Address Latches -----------------------------------------------------
	process (clk, nreset, get_new_addr, latched_mode, VME_Addr_to_Lower)
	begin
		if (nreset = '0') 		then 
			prev_Addr <= (others => '0');
		elsif rising_edge(clk) 	then
			if (get_new_addr = '1') then 
				if   (latched_mode = '0') then 
					prev_Addr <= VME_Addr_to_Lower(18 downto 0);
					 
				elsif(latched_mode = '1') then 
					prev_Addr <= Addr_B_to_Lower(18 downto 0);
					 
				end if;
			end if;
		end if;		
	end process;		


--=======================================================================
-- STATE MACHINE  
-------------------------------------------------------------------------
SM_REG: 
process(clk, nreset)
begin
	if(nreset='0')			then CS <= init;
	elsif(rising_edge(clk))	then CS <= NS;
	end if;
end process;

SM_NEXT_STATE: 
process(CS, RD_A, WR_A, RD_B, WR_B, cnt_bit_2, latched_mode, 
		prev_Addr, VME_Addr_to_Lower, Addr_B_to_Lower)
begin
	-- DEFAULT OUTPUT VALUES --------------------------
    ena_shift 	<= '0'; -- enable shifting.
    nrst_shift	<= '1'; -- reset shift counter.
	 load_sr 	<= '0'; -- loads parallel address and data into SR's.
    read_sr 	<= '0'; -- causes parallel output registers to latch SR bits.
    wr			<= '0'; -- becomes a write enable for external memory.
    rd			<= '0'; -- becomes the OE for the Lower fpga during reads.
    
    get_mode	  	<= '0'; 
    dds_nvme_mode <= '1'; -- High if DDS Mode, Low if VME Mode
    
    get_new_addr  <= '0'; -- latches the parallel address in
    
    end_proc		<= '0'; -- signals the end of a transfer for command latches
    
    intf_ready <= '0'; -- becomes inactive one clock after the transfer
                       -- process is triggered and signals 'ready' one
					   -- clock before the process can be triggered again.
    ---------------------------------------------------
        
	case CS is
		when init =>							NS <= idle;
			nrst_shift <= '0'; -- reset shift counter
		
		when idle =>
			intf_ready <= '1';
			load_sr    <= '1'; -- this will latch address and data when '0'
			get_mode   <= '1'; -- modes follow inputs and latch when leaving idle
			
			if   (RD_A = '1') then
				dds_nvme_mode <= '0';
				if( prev_Addr = VME_Addr_to_Lower(18 downto 0)) then	
														NS <= rd_pause; 
					-- THIS ADDRESS SAME AS PREVIOUSLY TRANSMITTED ADDRESS
				else									NS <= addr_pause;
				end if;
				
			elsif(RD_B = '1') then
				dds_nvme_mode <= '1';
				if( prev_Addr = Addr_B_to_Lower(18 downto 0)) then	
														NS <= rd_pause; 
					-- THIS ADDRESS SAME AS PREVIOUSLY TRANSMITTED ADDRESS
				else									NS <= addr_pause;
				end if;
									  
			elsif(WR_A = '1') then
				dds_nvme_mode <= '0';			NS <= wr_pause;
													  
			elsif(WR_B = '1') then 	
				dds_nvme_mode <= '1';			NS <= wr_pause;
															
			else										NS <= idle;
			end if;
			
		-------------------------------------------------------
		-- A new address must be transmitted to complete the read.
		
		when addr_pause =>						NS <= addr_shift;
			rd		  <= '1';
			load_sr    <= '1'; -- this will latch address and data when '0'
			
		when addr_shift =>						
			rd		  <= '1';
			ena_shift <= '1'; -- shift registers are shifting
			if (cnt_bit_2 = '1') then			NS <= addr_wait;
			else										NS <= addr_shift;
			end if;
		
		when addr_wait =>							NS <= addr_reg;
			rd		  		<= '1';
			ena_shift 	<= '1'; -- shift registers are shifting
			get_new_addr <= '1';
			
			read_sr <= '1'; -- ***************************************** added 3/28/2016
				
		when addr_reg =>							NS <= setup_data;
			rd		<= '1';
			read_sr <= '1';
			nrst_shift <= '0'; -- reset shift counter
		
		when setup_data =>						NS <= rd_pause;
			rd		<= '1';  -- ***************************************** added 3/28/2016
			
		--------------------------------------------------------
		-- The desired address is set in the Lower FPGA at this point.
		
		when rd_pause =>							NS <= rd_pause2;
			rd		  	<= '1';
			load_sr   	<= '1'; -- this will latch address and data when '0'
			get_new_addr <= '1';
			
		when rd_pause2 =>							NS <= rd_shift;
			rd		  	<= '1';
			load_sr   	<= '1'; -- this will latch address and data when '0'
			get_new_addr <= '1';
			
		when rd_shift =>						
			rd		  	<= '1';
			ena_shift 	<= '1'; -- shift registers are shifting
			if (cnt_bit_2 = '1') then			NS <= rd_wait;
			else										NS <= rd_shift;
			end if;
		
		when rd_wait =>							NS <= rd_reg;
			rd		  	<= '1';
			ena_shift 	<= '1'; -- shift registers are shifting
			end_proc	<= '1'; -- 1 clock early for registered outputs
					
		when rd_reg =>							NS <= idle;
			rd			<= '1';
			read_sr 	<= '1';
			nrst_shift 	<= '0'; -- reset shift counter
			intf_ready 	<= '1';-- leaning forward a bit is OK
			
		-----------------------------------------------------------------	
		-- Write operation states
		
		when wr_pause =>							NS <= wr_shift;
			get_new_addr <= '1';
			load_sr    <= '1'; -- this will latch address and data when '0'
			
		when wr_shift =>							
			ena_shift 	<= '1'; -- shift registers are shifting
			if (cnt_bit_2 = '1') then			NS <= wr_wait;
			else										NS <= wr_shift;
			end if;

		when wr_wait =>							NS <= wr_reg;
			ena_shift 	<= '1'; -- shift registers are shifting
			end_proc	<= '1'; -- 1 clock early for registered outputs

		when wr_reg =>								NS <= idle;
			wr      	<= '1'; 
			read_sr 	<= '1';
			nrst_shift 	<= '0'; -- reset shift counter
			intf_ready 	<= '1';-- leaning forward a bit is OK
		-----------------------------------------------------------------	
			
		when others =>              			NS <= idle;
	end case;
end process;

-- Outputs ---------------------------------------------------------------------
process (clk, nreset, rd, load_sr, read_sr, ena_shift, wr, end_proc, latched_mode)
begin
	RD_lower_mem 	<= rd; 
	LD_lower_mem 	<= load_sr;	
	RD_lower_sr 	<= read_sr;
	ENA_shift_out	<= ena_shift;	
	
	if(nreset = '0')  then 
		WR_lower_mem 	  <= '0';
		VME_Transfer_Done <= '0';
		UPD_Transfer_Done <= '0';
	elsif rising_edge(clk) then
		WR_lower_mem <= wr;  -- the write is delayed 1 clock w.r.t. RD_lower_sr
		VME_Transfer_Done <= end_proc AND (NOT latched_mode);
		UPD_Transfer_Done <= end_proc AND      latched_mode;
	end if;
	
end process;

test_of_mode <=  latched_mode;


end;