Library IEEE;
USE IEEE.std_logic_1164.all;
USE IEEE.std_logic_arith.all;
USE ieee.std_logic_unsigned.all;
USE ieee.numeric_std.all;
USE work.pkg_081006.all; 

--------------------------------------------------------------
-- Description:
-- This module 

entity LMK01010_INTERFACE_V2 is
	port(clk       : in std_logic; -- global Logic Clock
		 reset_n   : in std_logic; -- global Logic Reset		 
		 clk_uwire  : out std_logic;-- convert signal out to ADC.
		 data_uwire : out std_logic;-- data clock out to ADC.
		 LE_uwire   : out std_logic -- DIAGNOSTIC
		);
end LMK01010_INTERFACE_V2;
	
architecture main of LMK01010_INTERFACE_V2 is

	-- Shift Register declarations -----------------------------------
	constant DDS_0_Delay : std_logic_vector(3 downto 0) := "1000";
	constant DDS_1_Delay : std_logic_vector(3 downto 0) := "1001";
	constant DDS_2_Delay : std_logic_vector(3 downto 0) := "1001";
	constant DDS_3_Delay : std_logic_vector(3 downto 0) := "1000";
	constant DDS_4_Delay : std_logic_vector(3 downto 0) := "0110";
	
	signal delay         : std_logic_vector(3 downto 0);
	
	signal data_out   	: std_logic_vector(31 downto 0); -- serial data out		
	signal shift_reg   	: std_logic_vector(31 downto 0); -- serial data out	
	
	signal xmit				: std_logic;
	signal load 			: std_logic;
	signal shift 			: std_logic;
	signal xmit_done     : std_logic;

	-- Reg Counter declarations --------------------------------------
	signal reg_cnt       : std_logic_vector(3 downto 0);
	signal clr_reg_cnt   : std_logic;
	signal inc_reg_cnt   : std_logic;

	-- Shift Counter declarations --------------------------------------
	signal shift_cnt     : std_logic_vector(5 downto 0);
	signal clr_shift_cnt : std_logic;

	-- State Machine declarations ------------------------------------
	type STATE_TYPE is 
	(init, send_rst, wait1, xmit_reg, wait2, increment_reg_cnt, test_reg_cnt, 
		send_r14, wait_r14, idle);
	signal CS, NS  : STATE_TYPE;

		-- State Machine declarations ------------------------------------
	type STATE_TYPE2 is 
	(idle2, LE_low, LE_setup, clk_hi_1, clk_hi_2, clk_lo_1, clk_lo_2, done);
	signal CS2, NS2  : STATE_TYPE2;

--===begin===begin===begin===begin===begin===begin===begin===begin===
begin -- main --------------------------------------------------------
-- Counters -------------------------------------------------------------
	regcnt: process (clk, reset_n, clr_reg_cnt, inc_reg_cnt) -- count registers written
	begin
		if   (reset_n = '0') 			then reg_cnt <= (others => '0');
		elsif rising_edge(clk) then
			if   (inc_reg_cnt = '1') 	then reg_cnt <= reg_cnt + 1;
			elsif(clr_reg_cnt = '1')   then reg_cnt <= (others => '0');
			end if;
		end if;
	end process;				 

	shiftcnt: process (clk, reset_n, clr_shift_cnt, shift) 
	begin
		if   (reset_n = '0') 			then shift_cnt <= (others => '0');
		elsif rising_edge(clk) then
			if   (shift = '1') 			then shift_cnt <= shift_cnt + 1;
			elsif(clr_shift_cnt = '1') then shift_cnt <= (others => '0');
			end if;
		end if;
	end process;	

-- Delay setting based on address ---------------------------------------
	delay <= DDS_0_Delay when reg_cnt = X"0" else
			 DDS_1_Delay when reg_cnt = X"1" else
			 DDS_2_Delay when reg_cnt = X"2" else
			 DDS_3_Delay when reg_cnt = X"3" else
			 DDS_4_Delay when reg_cnt = X"4" else
			 X"0";
			 	
------------------------------------------------------------------------		
-- State Machine define initialization ---------------------------
	sm_reg: process (clk, reset_n)	
	begin
		if(reset_n = '0')        then CS <= init;
		elsif(clk'event and clk='1')then CS <= NS;
		end if;
	end process; -- sm_reg
		
	sm_combinatorial: process (CS, reg_cnt, xmit_done, delay)
	begin
		-- default signal values
		clr_reg_cnt <= '0';
		inc_reg_cnt <= '0';
		
		data_out		<= (others => '0');
		xmit			<= '0';
		-------------------------

		case CS is
			when init  =>						NS <= send_rst; 
				-- We start here and procede once the global reset is over.
				clr_reg_cnt   <= '1'; -- clear counters
				
			when send_rst  =>					NS <= wait1;
				data_out		<= (others => '1');
				xmit			<= '1';
				
			when wait1		=>					
				data_out		<= (others => '1'); -- hold data		
				
				if(xmit_done = '1') then 	NS <= xmit_reg;
				else				     			NS <= wait1;
				end if;
			
			-- start reg loop -----------------------------
			when xmit_reg  =>					NS <= wait2;
				data_out(31 downto 8) 	<= X"000501";
				data_out( 7 downto 4) 	<= delay;
				data_out( 3 downto 0) 	<= reg_cnt;
				xmit							<= '1';
				
			when wait2		=>					
				data_out(31 downto 8) 	<= X"000501";
				data_out( 7 downto 4) 	<= delay;
				data_out( 3 downto 0) 	<= reg_cnt;
				
				if(xmit_done = '1') then 	NS <= increment_reg_cnt;
				else				     			NS <= wait2;
				end if;
			
			when increment_reg_cnt =>		NS <= test_reg_cnt;
				inc_reg_cnt   <= '1';  -- increment counter
			
			when test_reg_cnt =>
			
				if   (reg_cnt = X"5") then NS <= send_r14;
				else								NS <= xmit_reg;
				end if;
			
		  -- end loop -----------------------------------	
				
			when send_r14  => 				NS <= wait_r14;
				data_out		<= X"4800000E";
				xmit			<= '1';
				
			when wait_r14		=>					
				data_out		<= X"4800000E"; -- hold data	
				
				if(xmit_done = '1') then 	NS <= idle;
				else				     			NS <= wait_r14;
				end if;
		
			when idle  => -- Note: may want to update LMK in the future.
				NS <= idle; -- dead stop here until next system reset.
										
		end case; 
				
	end process; -- sm_combinatorial
	
	---------------------------------------------------------------------------------------
	SHIFT_REGISTER: process (clk, reset_n, load, shift, data_out, shift_reg)
	begin
		if   (reset_n = '0') then shift_reg <= (others => '1'); -- clear to reg0 reset word
		elsif rising_edge(clk) then
			if   (load = '1') then 
				shift_reg <= data_out;
			
			elsif(shift = '1') then 
				shift_reg(31 downto 1) <= shift_reg(30 downto 0);
				shift_reg(0) <= '0';
			end if;
		end if;
	end process;				 

------------------------------------------------------------------------		
-- SR State Machine define initialization ---------------------------
	sm_reg2: process (clk, reset_n)	
	begin
		if(reset_n = '0')        then CS2 <= idle2;
		elsif(clk'event and clk='1')then CS2 <= NS2;
		end if;
	end process; -- sm_reg
		
	sm_combinatorial2: process (CS2, xmit, shift_reg, shift_cnt)
	begin
		-- default signal values
		clr_shift_cnt 	<= '0';
		shift 		  	<= '0';
		load			  	<= '0';
		xmit_done	  	<= '0';
		
		-- OUTPUTS ----------------------
		LE_uwire			<= '0';
		clk_uwire		<= '0';
		data_uwire 		<= shift_reg(31);
		---------------------------------
-- 	(idle2, LE_low, LE_setup, clk_hi_1, clk_hi_2, clk_lo_1, clk_lo_2, done);
		case CS2 is
			when idle2  =>
				clr_shift_cnt 	<= '1';
				LE_uwire			<= '1';
				load				<= '1';
				
				if   (xmit = '1') then 		NS2 <= LE_low;
				else								NS2 <= idle2;
				end if;
						
			when LE_low		=>					NS2 <= LE_setup;
				LE_uwire			<= '0'; -- default value
				
			when LE_setup	=>					NS2 <= clk_hi_1;
			
			-- start loop ---------------------------------
			when clk_hi_1	=>					NS2 <= clk_hi_2;
				clk_uwire		<= '1';
			
			when clk_hi_2	=>					NS2 <= clk_lo_1;
				clk_uwire		<= '1';
				shift 			<= '1'; -- trigger shift register shift
											  -- here when clock goes Hi to Low
			
			when clk_lo_1	=>				
				clk_uwire		<= '0';
				if   (shift_cnt = "100000")	then 	NS2 <= done;
				else										NS2 <= clk_lo_2;
				end if;
			
			when clk_lo_2	=>					NS2 <= clk_hi_1;
				clk_uwire		<= '0';
			-- end loop -----------------------------------
			
			when done		=>					NS2 <= idle2;
				xmit_done	  	<= '1'; -- signal higher level
				clk_uwire		<= '0';
				LE_uwire			<= '0'; -- default value
		
		end case; 
				
	end process; -- sm_combinatorial
end main;

