Library IEEE;
USE IEEE.std_logic_1164.all;
USE IEEE.std_logic_unsigned.all;

USE work.pkg_081006.all;
USE work.DDS_VME_CONSTANTS.all;

---------------------------------------------------------
-- version 4 expanded the disable paraphase to include the PC1, PC2 and BR
--  offsets.

-- version 5 inverts (x (-1)) the RPOS phase drive signal

-- version 6 is modified to have a continuous update pulse to keep the
-- phase updates to the DDS continually following the RPOS phase drive.
-- Also, the Disable_Paraphase signal disables the Transition Trigger.
-- Also, reversed DataA and DataB in the subtraction in PARAPHASE_MINUS_REF_PHASE:
-- since the last version.

entity Phase_Controller_v6 is
port(
		clk						: in std_logic;
		nreset					: in std_logic;
		
		Reset_Curves			: in std_logic;
		Run_Curves				: in std_logic;
		Update_Pulse			: in std_logic;
		Get_Next_Freq_Phase	: in std_logic;
		
		Transition_Trig		: in std_logic;
		Trans_Time_Intvl		: in std_logic_vector(4 downto 0);
		
		RPOS_Phase_Drive		: in std_logic_vector(15 downto 0);
		RPOS_Strobe				: in std_logic;
		
		Phase_Offset_A			: in std_logic_vector(15 downto 0);
		Phase_Offset_B			: in std_logic_vector(15 downto 0);
		
		ParaPh_Mem_Data		: in std_logic_vector(15 downto 0);
		ParaPh_PC1OFF_OB		: in std_logic_vector(15 downto 0);
		ParaPh_PC2OFF_OB		: in std_logic_vector(15 downto 0);
		ParaPh_PC2CNT			: in std_logic_vector(15 downto 0);
		ParaPh_PBROFF_OB		: in std_logic_vector(15 downto 0);
		ParaPh_PBRCNT			: in std_logic_vector(15 downto 0);
		 
		Disable_RPOS_Phase	: in std_logic; -- VME Control for Debugging
		Disable_Offsets		: in std_logic; -- VME Control for Debugging
		Disable_Paraphase		: in std_logic; -- VME Control for Debugging
		BR_Enable				: in std_logic; -- Paraphase Bunch Rotation Enable
		------------------------------------------------------------
		ParaPh_Mem_Addr		: out std_logic_vector(10 downto 1);
		ParaPh_Mem_Ena			: out std_logic;
		
		Phase_A					: out std_logic_vector(15 downto 0);
		Phase_B					: out std_logic_vector(15 downto 0);
		Paraphase				: out std_logic_vector(15 downto 0);
		Reference_Phase		: out std_logic_vector(15 downto 0);
		RPOS_Phase				: out std_logic_vector(15 downto 0);
		Paraphase_Sum			: out std_logic_vector(15 downto 0);
		Strobe_Out				: out std_logic;
		
		c1, c2, c3, c4, c5, c6, c7	: out std_logic
	);
end Phase_Controller_v6;

--########################################################
--				A R C H I T E C T U R E
--########################################################
architecture main of Phase_Controller_v6 is
-- Synchronized Inputs --------------------------------
	signal rst			: std_logic;
	signal get_next	: std_logic;
	signal trans_trig	: std_logic;
	
	signal get_paraphase: std_logic;
	signal strobe		: std_logic;
	signal update_ref	: std_logic;
	
	
-- Buffers ----------------------------------------------
	signal ParaPh_in				: std_logic_vector(15 downto 0);
	signal pp_upd_cnt				: std_logic_vector(15 downto 0);

	signal pc1off_2s				: std_logic_vector(15 downto 0);
	signal pc2off_2s				: std_logic_vector(15 downto 0);
	signal pbroff_2s				: std_logic_vector(15 downto 0);
	
	signal pp_r1					: std_logic_vector(15 downto 0);
	signal pp_ovf1					: std_logic;
	signal pp_r1_result			: std_logic_vector(15 downto 0);

	signal pp_r2					: std_logic_vector(15 downto 0);
	signal pp_ovf2					: std_logic;
	signal pp_r2_result			: std_logic_vector(15 downto 0);

	signal pp_r3					: std_logic_vector(15 downto 0);
	signal pp_ovf3					: std_logic;
	signal pp_r3_result			: std_logic_vector(15 downto 0);
	
	signal RPOS_Phase_Drive_in : std_logic_vector(15 downto 0);
	signal Phase_Drive_in		: std_logic_vector(15 downto 0);
	signal pd_inverse				: std_logic_vector(16 downto 0);
	signal phase_drive			: std_logic_vector(16 downto 0);

	signal Phase_Off_A_in		: std_logic_vector(15 downto 0);
	signal Phase_Off_B_in		: std_logic_vector(15 downto 0);
	
	signal Phase_A_buf			: std_logic_vector(15 downto 0);
	signal Phase_B_buf			: std_logic_vector(15 downto 0);
          
-- Altera Math Functions ------------------------------
component signed_adsub_16a
	PORT
	(
		add_sub		: IN STD_LOGIC ;
		dataa			: IN STD_LOGIC_VECTOR (15 DOWNTO 0);
		datab			: IN STD_LOGIC_VECTOR (15 DOWNTO 0);
		overflow		: OUT STD_LOGIC ;
		result		: OUT STD_LOGIC_VECTOR (15 DOWNTO 0)
	);
end component;

component unsigned_sub17
	PORT
	(
		datab		: IN STD_LOGIC_VECTOR (16 DOWNTO 0);
		result		: OUT STD_LOGIC_VECTOR (16 DOWNTO 0)
	);
end component;


--	signal c1, c2, c3, c4, c5, c6, c7	: std_logic; -- carry out signals
	
	signal r1	: std_logic_vector(15 downto 0); -- result value
	signal r2	: std_logic_vector(15 downto 0); -- result value
	signal r3	: std_logic_vector(15 downto 0); -- result value
	signal r4	: std_logic_vector(15 downto 0); -- result value
	
	signal reg_r1	: std_logic_vector(15 downto 0); -- register for result value
	signal reg_r2	: std_logic_vector(15 downto 0); -- register for result value
	signal reg_r3	: std_logic_vector(15 downto 0); -- register for result value
	signal reg_r4	: std_logic_vector(15 downto 0); -- register for result value
	
	signal r5	: std_logic_vector(15 downto 0); -- result value
	signal r6	: std_logic_vector(15 downto 0); -- result value
	signal r7	: std_logic_vector(15 downto 0); -- result value
	

component phase_step_LUT -- initialized by Phase_Step_LUT.mif
	PORT
	(
		address	: IN STD_LOGIC_VECTOR (4 DOWNTO 0);
		clock		: IN STD_LOGIC ;
		q			: OUT STD_LOGIC_VECTOR (15 DOWNTO 0)
	);
end component;

-- Reference Phase Signals -------------------------------------
	signal ref_phase		: std_logic_vector(15 downto 0);
	signal ref_phase_step	: std_logic_vector(15 downto 0);
	signal ref_step_count	: std_logic_vector( 4 downto 0);
	signal rst_ref_phase	: std_logic;
	signal dec_ref_phase	: std_logic;
	signal set_ref_phase	: std_logic;
	
-- Paraphase Memory Address Pointer -----------------------------
	signal paraph_ptr		:std_logic_vector(10 downto 1);
	signal rst_paraph_ptr	:std_logic;
	signal inc_paraph_ptr	:std_logic;
	
-- Pause Timer -----------------------------------------------------
	signal pause_tmr 	: std_logic_vector(2 downto 0);
	signal inc_pause_tmr: std_logic;

-- SETUP_STATE_MACHINE ---------------------------------------
type STATE_TYPE is
(init, init_data1, idle, pause1, inc_pointer, send_strobe, 
 get_next_paraph);
signal CS, NS : STATE_TYPE;

-- SETUP_STATE_MACHINE 2---------------------------------------
-- Used to manage the phase change in the reference phase at
-- transition.
type STATE_TYPE2 is
(init2, idle2, wait_for_update, step_ref_phase);
signal CS2, NS2 : STATE_TYPE2;

--========================================================
--						B E G I N
--========================================================
begin

get_next <= (Get_Next_Freq_Phase AND Run_Curves);

------------------------------------------------------------
SYNC_RESET: led 
	port map(sig_in=>Reset_Curves, clk=>clk, clrn=>nreset, pls_out=>rst); 


SYNC_TRANS_TRIG: led 
	port map(sig_in=>Transition_Trig AND NOT Disable_Paraphase,
				clk=>clk, 
				clrn=>nreset, 
				pls_out=>trans_trig); 

----------------------------------------------------------
-- Look Up Table for the Reference Phase Step
-- This is the step in phase taken each 1us when the 
-- Reference Phase is rotating 180 deg at Booster Transition
------------------------------------------------------------
REFERENCE_PHASE_STEP_LUT_MEMORY:
--CHOOSE THE PHASE STEP GIVEN THE NUMBER OF 1us UPDATES IN INTERVAL
phase_step_LUT port map( 	address 	=> Trans_Time_Intvl,
									clock		=> clk,
									q			=> ref_phase_step );
							
------------------------------------------------------------
-- Reference Phase
-- The reference phase is setup to be adjust at transition.
------------------------------------------------------------
REFERENCE_PHASE_VALUE:
process(clk, nreset, ref_phase, ref_phase_step, rst_ref_phase,
		dec_ref_phase, set_ref_phase)
begin
	if   (nreset = '0') then	ref_phase <= (others => '0');
	elsif rising_edge(clk) then
		if   (rst_ref_phase = '1') then 
					ref_phase <= X"4000"; -- +90 is the nominal ref phase
		elsif(dec_ref_phase = '1') then 
					ref_phase <= ref_phase - ref_phase_step;
		elsif(set_ref_phase = '1') then 
					ref_phase <= X"C000"; -- -90 is the post trans refence
		end if;
	end if;
end process;

------------------------------------------------------------
REFERENCE_PHASE_STEP_COUNT: -- COUNT UPDATES DURING TRANSITION INTERVAL
process(clk, nreset, ref_step_count)
begin
	if   (nreset = '0') then ref_step_count <= (others => '0');
	elsif rising_edge(clk) then
		if   (rst_ref_phase = '1') then 
					ref_step_count <= (others => '0');
		elsif(dec_ref_phase = '1') then 
					ref_step_count <= ref_step_count + 1;
		end if;
	end if;
end process;			

------------------------------------------------------------
PARAPHASE_UPDATE_COUNT: -- COUNT UPDATES THROUGHOUT THE CYCLE
process(clk, nreset, get_next)
begin
	if   (nreset = '0') then 		pp_upd_cnt <= (others => '0');
	elsif rising_edge(clk) then
		if(get_next = '1')  then 	pp_upd_cnt <= pp_upd_cnt + 1;
		else   							pp_upd_cnt <= (others => '0');
		end if;
	end if;
end process;			

------------------------------------------------------------
BUFFERS:
process(clk, nreset, get_paraphase,	RPOS_Strobe,
		ParaPh_Mem_Data, RPOS_Phase_Drive, Phase_Offset_A, Phase_Offset_B,
		ParaPh_PC1OFF_OB, ParaPh_PC2OFF_OB, ParaPh_PBROFF_OB)
begin
	if   (nreset = '0') then
		ParaPh_in		<= (others => '0');
		Phase_Drive_in	<= (others => '0');	 
		Phase_Off_A_in	<= (others => '0');	
		Phase_Off_B_in	<= (others => '0');
		pc1off_2s 		<= (others => '0');	-- set always see above	
		pc2off_2s 		<= (others => '0');		
		pbroff_2s 		<= (others => '0');		
		
	elsif rising_edge(clk) then
	
		if   (Disable_Paraphase = '1') then	
			ParaPh_in <= (others => '0');			
			pc1off_2s <= (others => '0');		
			pc2off_2s <= (others => '0');		
			pbroff_2s <= (others => '0');		
		elsif(get_paraphase = '1') then		
			ParaPh_in	<= ParaPh_Mem_Data;
			pc1off_2s <= not ParaPh_PC1OFF_OB(15) & ParaPh_PC1OFF_OB(14 downto 0);	
			pc2off_2s <= not ParaPh_PC2OFF_OB(15) & ParaPh_PC2OFF_OB(14 downto 0);	
			pbroff_2s <= not ParaPh_PBROFF_OB(15) & ParaPh_PBROFF_OB(14 downto 0);	
		end if;
		
		-------------------------------------
		if   (Disable_Offsets = '1') then	Phase_Off_A_in	<= (others => '0');
														Phase_Off_B_in	<= (others => '0');	
														
		else											Phase_Off_A_in	<= Phase_Offset_A;	
														Phase_Off_B_in	<= Phase_Offset_B;	
		end if;
		
		-------------------------------------
		if  (Disable_RPOS_Phase = '1') then	Phase_Drive_in	<= (others => '0');
		elsif(RPOS_Strobe = '1') then			Phase_Drive_in <= RPOS_Phase_Drive;
		end if;
		
	end if;
end process;

------------------------------------------------------------
-- Paraphase Memory Address Pointer
PARAPHASE_MEMORY_POINTER:
process(clk, nreset, rst_paraph_ptr, inc_paraph_ptr)
begin
	if   (nreset = '0') then paraph_ptr <= (others => '0');
	elsif rising_edge(clk) then
		if   (rst_paraph_ptr = '1') then paraph_ptr <= (others => '0');
		elsif(inc_paraph_ptr = '1') then paraph_ptr <= paraph_ptr + 1;
		end if;
	end if;
end process;
	
PARAPHASE_PLUS_PC1OFF:
signed_adsub_16a    port map(	add_sub => '1', -- add
								dataa		=> pc1off_2s,
								datab		=> ParaPh_in,
								overflow => pp_ovf1,
								result	=> pp_r1 );
								
PARAPHASE_OVERFLOW_TEST_1:
process(clk, nreset, pp_ovf1, pp_r1)
begin
	if   (nreset = '0') 		then 		pp_r1_result <= X"7FFF";
	elsif rising_edge(clk) 	then
		if   (pp_ovf1 = '1') then 
			if (pp_r1(15) = '0') then 	pp_r1_result <= X"8000"; -- most negative
			else								pp_r1_result <= X"7FFF"; -- most positive
			end if;
		else									pp_r1_result <= pp_r1;
		end if;
	end if;
end process;

PARAPHASE_PLUS_PC2OFF:
signed_adsub_16a    port map(	add_sub => '1', -- add
								dataa		=> pc2off_2s,
								datab		=> pp_r1_result,
								overflow => pp_ovf2,
								result	=> pp_r2 );
								
PARAPHASE_OVERFLOW_TEST_2:
process(clk, nreset, pp_ovf2, pp_r2, pp_r1_result, pp_upd_cnt, ParaPh_PC2CNT)
begin
	if   (nreset = '0') 		then 		pp_r2_result <= X"7FFF";
	elsif rising_edge(clk) 	then
		if   (pp_upd_cnt > ParaPh_PC2CNT) then
			if   (pp_ovf2 = '1') then 
				if (pp_r2(15) = '0') then 	pp_r2_result <= X"8000"; -- most negative
				else								pp_r2_result <= X"7FFF"; -- most positive
				end if;
			else									pp_r2_result <= pp_r2; -- pass on sum of offset and pp_r1_result
			end if;
		else										pp_r2_result <= pp_r1_result; -- otherwise bypass this offset
		end if;
	end if;
end process;

PARAPHASE_PLUS_PBROFF:
signed_adsub_16a    port map(	add_sub => '1', -- add
								dataa		=> pBRoff_2s,
								datab		=> pp_r2_result,
								overflow => pp_ovf3,
								result	=> pp_r3 );
								
PARAPHASE_OVERFLOW_TEST_3:
process(clk, nreset, pp_ovf3, pp_r3, pp_r2_result, pp_upd_cnt, ParaPh_PBRCNT, BR_Enable)
begin
	if   (nreset = '0') 		then 		pp_r3_result <= X"7FFF";
	elsif rising_edge(clk) 	then
		if   ((pp_upd_cnt >= ParaPh_PBRCNT) AND (BR_Enable = '1')) then
			if   (pp_ovf3 = '1') then 
				if (pp_r3(15) = '0') then 	pp_r3_result <= X"8000"; -- most negative
				else								pp_r3_result <= X"7FFF"; -- most positive
				end if;
			else									pp_r3_result <= pp_r3;-- pass on sum of offset and pp_r2_result
			end if;
		else										pp_r3_result <= pp_r2_result; -- otherwise bypass this offset
		end if;
	end if;
end process;

								
--PARAPHASE_MINUS_REF_PHASE:   --- OLD BROKEN VERSION
--signed_adsub_16a    port map(	add_sub => '0', -- subtract
--								dataa		=> pp_r3_result,
--								datab		=> ref_phase,
--								overflow	=> c1,
--								result	=> r1 );
REF_PHASE_MINUS_PARAPHASE:
signed_adsub_16a    port map(	add_sub => '0', -- subtract
								dataa		=> ref_phase,
								datab		=> pp_r3_result,
								overflow	=> c1,
								result	=> r1 );

PARAPHASE_PLUS_REF_PHASE:
signed_adsub_16a    port map(	add_sub => '1', -- add
								dataa		=> ref_phase,
								datab		=> pp_r3_result,
								overflow	=> c2,
								result	=> r2 );

--------------------------------------------------------								
REGISTER_R1_R2: process(clk, nreset, r1, r2)
begin
	if   (nreset = '0') then 
						reg_r1 <= (others => '0');
						reg_r2 <= (others => '0');
	elsif rising_edge(clk) then
						reg_r1 <= r1;
						reg_r2 <= r2;
	end if;
end process;
--------------------------------------------------------
								
R1_PLUS_OFFSET_A:
signed_adsub_16a    port map(	add_sub => '1', -- add
								dataa		=> reg_r1,
								datab		=> Phase_Off_A_in,
								overflow	=> c3,
								result	=> r3 );
R2_PLUS_OFFSET_B:
signed_adsub_16a    port map(	add_sub => '1', -- add
								dataa		=> reg_r2,
								datab		=> Phase_Off_B_in,
								overflow	=> c4,
								result	=> r4 );
								
--------------------------------------------------------								
REGISTER_R3_R4: process(clk, nreset, Get_Next_Freq_Phase, r3, r4)
begin
	if   (nreset = '0') then 
						reg_r3 <= (others => '0');
						reg_r4 <= (others => '0');
	elsif rising_edge(clk) then
		if   (Get_Next_Freq_Phase = '1') then
						reg_r3 <= r3;
						reg_r4 <= r4;
		end if;
	end if;
end process;
--------------------------------------------------------
INVERT_PHASE_DRIVE : 
unsigned_sub17 PORT MAP (
		datab	 => '0' & Phase_Drive_in,
		result	 => pd_inverse(16 downto 0)
	);

UPDATE_PHASE_DRIVE: 
process(clk, nreset, Update_Pulse, pd_inverse, phase_drive)
begin
	if   (nreset = '0') then 				phase_drive <= (others => '0');
	elsif rising_edge(clk) then
		if   (Update_Pulse = '1') then	phase_drive <= pd_inverse;
		end if;
	end if;
end process;
--------------------------------------------------------


PHASE_DRIVE_PLUS_OFFSET_A:
signed_adsub_16a    port map(	add_sub => '1', -- add
								dataa		=> reg_r3,
								datab		=> phase_drive(15 downto 0),
								overflow	=> c6,
								result	=> r6 );
PHASE_B_PLUS_180:
signed_adsub_16a    port map(	add_sub => '1', -- add
								dataa		=> reg_r4,
								datab		=> X"8000",
								overflow	=> c5,
								result	=> r5 );
PHASE_DRIVE_PLUS_OFFSET_B:
signed_adsub_16a    port map(	add_sub => '1', -- add
								dataa		=> r5,
								datab		=> phase_drive(15 downto 0),
								overflow	=> c7,
								result	=> r7 );

------------------------------------------------------------------------------		
PAUSE_TIMER: -- Timer used to generate setup time for Paraphase memory Reads
process (clk, inc_pause_tmr)
begin
	if rising_edge(clk) then
		if   (inc_pause_tmr = '0') then pause_tmr <= "000";
		elsif(inc_pause_tmr = '1') then pause_tmr <= pause_tmr + 1;
		end if;
	end if;
end process;

--{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{
--                S T A T E   M A C H I N E
--}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}
SM_REG: process(clk, nreset, rst)
begin
	if(nreset='0') OR (rst = '1') then CS <= init;
	elsif(rising_edge(clk))	then CS <= NS;
	end if;
end process;

SM_NEXT_STATE: 
process
(CS, pause_tmr, paraph_ptr, get_next, pp_upd_cnt)
begin
	-- defaults ----------------------
	rst_paraph_ptr 	<= '0';
	inc_paraph_ptr 	<= '0';
	
	ParaPh_Mem_Ena	<= '0';

	get_paraphase 	<= '0';
	
	inc_pause_tmr 	<= '0';
	
	Strobe_Out 		<= '0';
	update_ref 		<= '0';
	
	----------------------------------
	
	case CS is
		when init =>						NS<= init_data1;	
			-- We start and return here after a Reset_Curves signal
			rst_paraph_ptr 	<= '1';
			ParaPh_Mem_Ena		<= '1';
			
		when init_data1 =>	        
			-- latch first paraphase value
			ParaPh_Mem_Ena	<= '1';
			get_paraphase 	<= '1';
			
			-- increment the pause timer and wait for timeout
			inc_pause_tmr	<= '1'; 			
			if   (pause_tmr = "011") then	NS <= idle;
			else									NS <= init_data1;
			end if;
		
		when idle => 			
			if   (get_next = '1') then		NS <= pause1;
			else									NS <= idle;
			end if;
		
		when pause1 =>							NS <= inc_pointer;
			
		when inc_pointer =>					NS <= send_strobe;			
			if  (paraph_ptr < N_PARAPHASE_PTS) then
				inc_paraph_ptr 	<= '1';
			end if;	
						
		when send_strobe =>					NS <= get_next_paraph;
			Strobe_Out 		<= '1';			
			ParaPh_Mem_Ena	<= '1'; -- get next paraphase value
			update_ref 		<= '1'; -- signal to second state machine (SM2)
			
		when get_next_paraph =>				
			ParaPh_Mem_Ena	<= '1'; -- get next paraphase value
			get_paraphase 	<= '1';
			
			inc_pause_tmr	<= '1'; 			
			if   (pause_tmr = "011") then	NS <= idle;
			else									NS <= get_next_paraph;
			end if;
		
		when others => 						NS <= idle;
	end case;
end process;
--{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{{
--           2 -- S T A T E   M A C H I N E -- 2
--}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}}
-- Used to control the reference phase change at transition

SM_REG2: process(clk, nreset, rst)
begin
	if(nreset='0') OR (rst = '1') then CS2 <= init2;
	elsif(rising_edge(clk))	then CS2 <= NS2;
	end if;
end process;

SM_NEXT_STATE2: 
process
(CS2, trans_trig, update_ref, ref_step_count, Trans_Time_Intvl)
begin
	-- defaults ----------------------	
	rst_ref_phase 	<= '0';
	dec_ref_phase 	<= '0';
	set_ref_phase 	<= '0';
	
	case CS2 is
		when init2 =>							NS2 <= idle2;	
			rst_ref_phase 	<= '1'; -- start and return here with Reset_Curves
		
		when idle2 => 		
			if   (trans_trig = '1') then	NS2 <= wait_for_update;
			else									NS2 <= idle2;
			end if;
		
		when wait_for_update =>  -- update_ref is a signal from the first SM
			if   (update_ref = '1') then	NS2 <= step_ref_phase;
			else									NS2 <= wait_for_update;
			end if;
		
		when step_ref_phase =>				NS2 <= wait_for_update;
			if   (ref_step_count < Trans_Time_Intvl) then
				dec_ref_phase <= '1';
			else
				set_ref_phase <= '1';
			end if;
		
		when others => 						NS2 <= idle2;
	end case;
end process;
	
-- OUTPUT --------------------------------
--~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
		Phase_A <= r6;
		Phase_B <= r7;	
	
	ParaPh_Mem_Addr	<= paraph_ptr;	
	Paraphase			<= ParaPh_in;		
	Reference_Phase	<= ref_phase;	
	RPOS_Phase			<= Phase_Drive_in;
	Paraphase_Sum		<= pp_r3_result;
	
end;             