Library IEEE;
USE IEEE.std_logic_1164.all;
USE IEEE.std_logic_unsigned.all;

-- My Definitions --------------------------------
USE work.pkg_081006.all; 
USE work.DDS_VME_CONSTANTS.all;

-------------------------------------------------------------------------------
-- MODULE DESCRIPTION, May 27, 2010
--
-- This module is a programmable timer with multiple output control signals.
-- The timer is syncronized to the Frequency/Phase Curve start time.  This
-- start time is the Booster Cycle Trigger plus a programmable delay.

-- Version 4,  April 19,2017
-- Added a pause interval between run and update to separate the bus accesses.
-- ADDED A dedicated Rewind_Curves was added since Rewind is what stops the Run Bus
-- accesses.

-- Versoion 6, June 21, 2017
-- In this version we let the DDS_1us_Update run continuously 
-- after the module start up. When triggered to run we resync the 1us update and 
-- the Run_Curves signal goes active. 

entity Process_Timer_v6 is 
	port( 
			clk						: in std_logic;
			nreset					: in std_logic;
			Curve_Delay				: in std_logic_vector( 31 downto 0);
			DDS_Ready				: in std_logic; -- enabled after power up setup.
			Booster_Cycle_Trig	: in std_logic;

			DDS_1us_Update		: out std_logic;
			Run_Curves			: out std_logic; -- High after start delay (time = 0.0)
												 -- maximum delay from trigger = 10,000 us
												 
			Rewind_Curves			: out std_logic;
			Start_Update_Period	: out std_logic; -- High after 38,000 us
			End_Update_Period		: out std_logic; -- High after 60,000 us
			
			Start_Initialize		: out std_logic; -- High after 61,000 us
			Cycle_Done				: out std_logic  -- Pulses at the end of cycle 55,000 us
			        
		);
	end Process_Timer_v6;
	
architecture main of Process_Timer_v6 is	

-- Syncronize the Trigger input -------------------------------------
	signal module_ready : std_logic;
	signal start_cycle : std_logic;
	
-- Update Timer ----------------------------------------------------
	signal timer			: std_logic_vector(5 downto 0);
	signal timer_ena		: std_logic; -- counter increment enable strobe

-- Cycle Timer, Count 1 us Updates ----------------------------------
	signal cyc_timer		: std_logic_vector(23 downto 0);
	signal cyc_timer_ena	: std_logic;
	
-- TIMER_STATE_MACHINE ----------------------------------------------
	type STATE_TYPE is
(init, idle, wait_on_delay, sync_1us_pulse, active_interval, pause_interval, 
update_interval, end_update_interval, init_interval, done);		
 	signal CS, NS : STATE_TYPE;

begin	

SYNC_READY:
process(clk, nreset, DDS_Ready)
begin
	if   (nreset = '0')    then 	module_ready <= '0'; 
	elsif rising_edge(clk) then	module_ready <= DDS_Ready; 
	end if;
end process;  

--====================================================
-- Syncronize the Trigger input.
SYNC_START: led 
port map(	sig_in  => Booster_Cycle_Trig AND module_ready, 
				clk     => clk, 
				clrn    => '1', 
				pls_out => start_cycle);

--====================================================
-- 1us Frequency/Phase Update Timer
UPDATE_TIMER_1US: -- Timer to define sample interval
process(clk, nreset, timer_ena)
begin
	if   (nreset = '0') then timer <= (others => '0');
	elsif rising_edge(clk) then
		if   (timer_ena = '0')   then timer <= (others => '0');
		elsif(timer >= "110001") then timer <= (others => '0');									  
		else						  			timer <= timer + 1;
		end if;
	end if;
end process;

UPDATE_PULSE_LOGIC:
process(clk, nreset, timer)
begin
	if   (nreset = '0')    then DDS_1us_Update <= '0'; 
	elsif rising_edge(clk) then
		if   (timer = "110001") then 	DDS_1us_Update <= '1'; 
		else						 			DDS_1us_Update <= '0';
		end if;
	end if;
end process;                        

--====================================================
-- Measures time from the Booster_Cycle_Trig
CYCLE_TIMER:  
process(clk, nreset, cyc_timer_ena)
begin
	if   (nreset = '0') then cyc_timer <= (others => '0');
	elsif rising_edge(clk) then
		if   (cyc_timer_ena = '0') then 	cyc_timer <= (others => '0');
		else						    			cyc_timer <= cyc_timer + 1;
		end if;
	end if;	
end process;

--====================================================
-- STATE MACHINE TIMER CONTROL
--**--   --**--   --**--   --**--   --**--   --**--   --**--  
SM_REG: process(clk, nreset)
begin
	if(nreset='0')					then CS <= init;
	elsif(rising_edge(clk))		then CS <= NS;
	end if;
end process;

SM_NEXT_STATE: 
process (CS, module_ready, start_cycle, cyc_timer, Curve_Delay)
begin
	-- Defaults -------------------------------------
		timer_ena		<= '1'; -- enabled by default
		cyc_timer_ena	<= '1'; -- enabled by default

		Run_Curves 				<= '0';
		Rewind_Curves			<= '0';
		Start_Update_Period	<= '0';
		End_Update_Period		<= '0';
		Start_Initialize    	<= '0';
		Cycle_Done				<= '0';
						
	--------------------------------------------------

	case CS is
		when init =>
			timer_ena		<= '0';
			cyc_timer_ena	<= '0';
			
			if   (module_ready = '1') then 	NS <= idle;
			else                   				NS <= init;
			end if;

		when idle =>
			cyc_timer_ena	<= '0';
			
			if   (start_cycle = '1') then 	NS <= sync_1us_pulse;
			else                   				NS <= idle;
			end if;
		
		when sync_1us_pulse =>					NS <= wait_on_delay;
			timer_ena		<= '0'; -- reset 1us timer
			
		when wait_on_delay =>
			cyc_timer_ena	<= '1'; -- default
			
			if(cyc_timer = "0000" & Curve_Delay(19 downto 0)) then 	
														NS <= active_interval;
			else                   				NS <= wait_on_delay;
			end if;
		
		when active_interval => 
		-- This interval ends 36 ms from "start_cycle" -------------------
			Run_Curves 			<= '1';
		
			if   (cyc_timer = X"1B7740") then 	NS <= pause_interval;
			else                   					NS <= active_interval;
			end if;
		
		when pause_interval	=>			-- pause for 4 us
			Run_Curves 			<= '0'; -- stop run curves (default value)
			Rewind_Curves		<= '1';

			if   (cyc_timer = X"1B7808") then 	NS <= update_interval;
			else                   					NS <= pause_interval;
			end if;
		
		
		when update_interval =>					
		-- This interval ends 60 ms from "start_cycle" -------------------
			Start_Update_Period	<= '1';
			Rewind_Curves			<= '1';
		
			if   (cyc_timer = X"2DC6C0") then 	NS <= end_update_interval;
			else                   					NS <= update_interval;
			end if;
			
		when end_update_interval =>
		-- This interval is a pause to ensure updates finish up (61 ms)----------
			End_Update_Period	<= '1';
			if   (cyc_timer = X"2E8A10") then 	NS <= init_interval;
			else                   					NS <= end_update_interval;
			end if;
					
		when init_interval => 					
		-- This interval ends 64 ms from "start_cycle" -------------------
			Start_Initialize	<= '1';
		
			if   (cyc_timer = X"30D400") then 	NS <= done;
			else                   					NS <= init_interval;
			end if;
		
		when done => 									NS <= idle;
			Cycle_Done			<= '1';
							
		when others =>              				NS <= idle;
	end case;
end process;
                      
end;