Library IEEE;
USE IEEE.std_logic_1164.all;
USE IEEE.std_logic_unsigned.all;

-- My Definitions --------------------------------
USE work.pkg_081006.all; 
USE work.DDS_VME_CONSTANTS.all;
-------------------------------------------------------------------------------
-- MODULE DESCRIPTION, May 11, 2010
--
-- The purpose of the Update process is to keep the Run Mode parameter registers 
-- in-sync with the shared memory used by the higher level processes in the crate 
-- controller and beyond in the ACNET control system.
--
-- The parameters used by the module are typically scale factors, offset, time
-- delays, and process flow control switch settings.  Three words in shared
-- memory are associated with each parameter.  There is a "Setting", a "Reading"
-- and a "Update Flag".  When the control system changes one of the modules 
-- parameters it writes the new value into the Setting of the parameter and sets
-- the parameters Update Flag to a value greater than zero.  The Update process
-- polls through the flags.  When it finds a set flag it will test the new
-- parameter value against upper and lower limits set in ROM Memory.  If the
-- new value is within the limits we write the associated Run Mode parameter 
-- register with the new value.  Then the parameter register is read and the
-- value is written into the "Reading", and then the Flag is reset to zero.
--
-- This module is also used to test every parameter Setting value and setup
-- all of the parameter registers after a boot, after the shared memory is
-- initialized from the Flash.  This is done by setting the Verify_All signal
-- and then triggering an Update.  If the Setting is not within its limits,
-- the default value defined for the parameter is written to the parameter
-- and the parameters Reading.

-- ########## A future revision will add an error signal out if a Setting is 
-- ########## found beyond its limits during a Verify_All.

entity Update_SM_Process_v2 is 
	port( 
		clk					: in std_logic;
		nreset				: in std_logic;
		UPD_Data_From_Mem 	: in std_logic_vector(31 downto 0);
		Transfer_Done		: in std_logic;

		Update_Trigger  	: in std_logic;
		Verify_All			: in std_logic;
		
		End_Update_Period : in std_logic; -- added 7/22/2016 to kill hung update process
		VME_Active			: in std_logic;
		
		UPD_Data_to_Mem		: out std_logic_vector(31 downto 0);
		UPD_Addr_to_Mem		: out std_logic_vector(23 downto 0);
		UPD_WR_Pulse		: out std_logic;
		UPD_RD_Pulse		: out std_logic;
		UPD_Bus_Ena			: out std_logic;
		
		Parameter_Changed	: out std_logic;
		Update_Proc_Done	: out std_logic;  -- Strobes when SR is done
		UPD_Permit_VME		: out std_logic;
		Update_Proc_State	: out std_logic_vector(3 downto 0)
			        
		);
	end Update_SM_Process_v2;
	
architecture main of Update_SM_Process_v2 is	

-- Tigger inputs ------------------------------------------------------
	signal do_sm_update			: std_logic;

-- SM_POINTER signals -------------------------------------------------
	signal sm_ptr_nrst	: std_logic; -- counter reset signal
	signal sm_ptr_inc	: std_logic; -- counter increment enable strobe
	signal sm_ptr_load	: std_logic;
	signal sm_ptr_step	: std_logic_vector(3 downto 0);
	signal sm_ptr_first	: std_logic_vector(19 downto 0);
	signal sm_ptr_last	: std_logic_vector(19 downto 0);
	signal sm_ptr_done	: std_logic; -- count end flag
	signal sm_ptr_out	: std_logic_vector(19 downto 0);
	
-- LIMIT_POINTER signals ------------------------------------------------
	signal lim_ptr_nrst	: std_logic; -- counter reset signal
	signal lim_ptr_inc	: std_logic; -- counter increment enable strobe
	signal lim_ptr_load	: std_logic;
	signal lim_ptr_step	: std_logic_vector(3 downto 0);
	signal lim_ptr_first: std_logic_vector(19 downto 0);
	signal lim_ptr_last	: std_logic_vector(19 downto 0);
	--signal lim_ptr_done	: std_logic; -- count end flag
	signal lim_ptr_out	: std_logic_vector(19 downto 0);

-- PARAM_POINTER signals -------------------------------------------------
	signal pr_ptr_nrst	: std_logic; -- counter reset signal
	signal pr_ptr_inc	: std_logic; -- counter increment enable strobe
	signal pr_ptr_load	: std_logic;
	signal pr_ptr_step		: std_logic_vector(3 downto 0);
	signal pr_ptr_first	: std_logic_vector(19 downto 0);
	signal pr_ptr_last	: std_logic_vector(19 downto 0);
	--signal pr_ptr_done	: std_logic; -- count end flag
	signal pr_ptr_out	: std_logic_vector(19 downto 0);

-- PARAM ADDRESS LATCH signals --------------------------------------
	signal get_param_ptr : std_logic;
	signal param_addr    : std_logic_vector(19 downto 0);
		   
-- Data Register 1 --------------------------------------------------
	signal get_data1	: std_logic;
	signal data_reg1_d	: std_logic_vector(31 downto 0);
	signal data_reg1_q	: std_logic_vector(31 downto 0);

-- Data Register 2 (LIMITS) -----------------------------------------
	signal get_data2	: std_logic;
	signal data_reg2_d	: std_logic_vector(31 downto 0);
	signal data_reg2_q	: std_logic_vector(31 downto 0);

-- Flag Address Latch -----------------------------------------------
signal get_flag_addr	: std_logic;
signal flag_addr 		: std_logic_vector(19 downto 0);

-- Time Out Timer ---------------------------------------------------
signal time_out		: std_logic_vector(15 downto 0);
signal ena_time_out	: std_logic;
	   
-- WR_STATE_MACHINE High Level --------------------------------------
	type STATE_TYPE is
(
 idle, init_sm_poll, check_vme, address_flag, setup_flag, get_sm_flag, 
 check_flag, addr_param_ptr, setup_param_ptr, get_param_addr,  
 addr_setting, setup_setting, get_setting, 
 addr_ul, setup_ul, get_ul, check_ul, chk_verify, next_lim, 
 addr_ll, setup_ll, get_ll, check_ll, chk_verify2, 
 addr_default, setup_default, get_default, 
 addr_param, write_param, wait1, 
 read_param, wait2, get_reading, upd_reading, 
 addr_reading, wr_reading, 
 addr_flag_again, clear_flag, next_flag, done
 );		
 	signal CS, NS : STATE_TYPE;	


--########################################################
--========================================================
--						B E G I N
--========================================================
--########################################################

begin

--**-- --**--   --**--   --**--   --**--   --**--   --**-- 
-- STROBE SYNCHRONIZATION --------------------------------  
SYNC_UPDATE: led -- syncronize the signal to start update polling sequence.
port map(	sig_in  => Update_Trigger, 
			clk     => clk, 
			clrn    => '1', 
			pls_out => do_sm_update); 

	
-- ADDR_POINTER_COUNTER --------------------------------------------
SM_ADDR_POINTER: countby_n -- Address Pointer into Shared Memory
generic map(20,4)
port map(	clk  	 => clk,
			clrn 	 => sm_ptr_nrst,
			ena  	 => sm_ptr_inc,
			load 	 => sm_ptr_load,
			ld_data  => sm_ptr_first,
			step	 => sm_ptr_step,
			flag_cnt => sm_ptr_last,
			flag_out => sm_ptr_done,
			cnt_out  => sm_ptr_out 
		);

-- ADDR_POINTER_COUNTER 2-------------------------------------------
RAM_ADDR_POINTER: countby_n -- Address Pointer into RAM Memory
generic map(20,4)
port map(	clk  	 => clk,
			clrn 	 => lim_ptr_nrst,
			ena  	 => lim_ptr_inc,
			load 	 => lim_ptr_load,
			step	 => lim_ptr_step,
			ld_data  => lim_ptr_first,
			flag_cnt => lim_ptr_last,
			flag_out => open, --lim_ptr_done,
			cnt_out  => lim_ptr_out 
		);

-- ADDR_POINTER_COUNTER 3--------------------------------------------
PR_ADDR_POINTER: countby_n -- Address Pointer to the Parameter Register
generic map(20,4)
port map(	clk  	 => clk,
				clrn 	 => pr_ptr_nrst,
				ena  	 => pr_ptr_inc,
				load 	 => pr_ptr_load,
				ld_data 	=> pr_ptr_first,
				step	 	=> pr_ptr_step,
				flag_cnt => pr_ptr_last,
				flag_out => open, --pr_ptr_done,
				cnt_out  => pr_ptr_out 
		);

-- PARAM PTR LATCH ----------------------------------------------
PARAM_ADDRESS_LATCH: dffr
generic map(20)
port map( 	d    => UPD_Data_From_Mem(19 downto 0), 
				clk  => clk, 
				clrn => nreset, 
				prn  => '1', 
				ena  => get_param_ptr, 
				q    => param_addr(19 downto 0)    
		);
		
-- FLAG ADDRESS LATCH ----------------------------------------------
FLAG_ADDRESS: dffr
generic map(20)
port map( 	d    => sm_ptr_out, 
			clk  => clk, 
			clrn => nreset, 
			prn  => '1', 
			ena  => get_flag_addr, 
			q    => flag_addr 
		);
		
-- DATA REGISTER 1 ------------------------------------------------
DATA_REGISTER_1: dffr
generic map(32)
port map( 	d    => data_reg1_d, 
			clk  => clk, 
			clrn => nreset, 
			prn  => '1', 
			ena  => get_data1, 
			q    => data_reg1_q 
		);
	
-- DATA REGISTER 2 ------------------------------------------------
DATA_REGISTER_2: dffr
generic map(32)
port map( 	d    => data_reg2_d, 
			clk  => clk, 
			clrn => nreset, 
			prn  => '1', 
			ena  => get_data2, 
			q    => data_reg2_q 
		);
	
TIME_OUT_TIMER: process(clk, nreset, ena_time_out)
begin
	if   (nreset='0') 		then time_out <= (others => '0');
	elsif(rising_edge(clk))	then
		if   (ena_time_out = '1') then 	time_out <= time_out + '1';
		else										time_out <= (others => '0');
		end if;
	end if;
end process;	
	
--**--   --**--   --**--   --**--   --**--   --**--   --**--  
----------------------------------------------------
-- STATE MACHINE  
----------------------------------------------------
SM_REG: process(clk, nreset, End_Update_Period)
begin
	if((nreset='0') or (End_Update_Period = '1'))then CS <= done; -- CS <= idle;
	elsif(rising_edge(clk))		then CS <= NS;
	end if;
end process;

SM_NEXT_STATE: 
process
(CS, do_sm_update, UPD_Data_From_Mem, data_reg1_q, data_reg2_q, 
 sm_ptr_out, sm_ptr_done, param_addr, VME_Active, time_out,
 flag_addr, pr_ptr_out, lim_ptr_out, Verify_All, Transfer_Done)
begin
	-- Defaults -------------------------------------
		sm_ptr_nrst		<= '1';
		sm_ptr_inc		<= '0';
		sm_ptr_load		<= '0';
		sm_ptr_step		<= "0100";
		sm_ptr_first	<= SM_PARAM_BLOCK_FIRST(19 downto 0);
		sm_ptr_last		<= SM_PARAM_BLOCK_LAST(19 downto 0);
		
		get_flag_addr	<= '0';

		lim_ptr_nrst 	<= '1';
      lim_ptr_inc 	<= '0';
      lim_ptr_load 	<= '0';
		lim_ptr_step	<= "0100";
      lim_ptr_first	<= SM_PARAM_LIMITS_FIRST(19 downto 0);
		lim_ptr_last	<= SM_PARAM_LIMITS_LAST(19 downto 0);
		
		pr_ptr_nrst		<= '1';
		pr_ptr_inc		<= '0';
		pr_ptr_load		<= '0';
		pr_ptr_step		<= "0100";
		pr_ptr_first	<= SM_PARAM_PTR_FIRST(19 downto 0);
		pr_ptr_last		<= SM_PARAM_PTR_LAST(19 downto 0);
		
		get_param_ptr   <= '0';

		get_data1		<= '0';  -- Strobe to latch data into register 1
		data_reg1_d		<= UPD_Data_From_Mem;
		
		get_data2		<= '0';  -- Strobe to latch data into register 2
		data_reg2_d		<= UPD_Data_From_Mem;
		
		UPD_Data_to_Mem <= X"00000444";--data_reg1_q; --- ^$%^$^%#^%#^%$%^$#%$^%&%%^$&^*^*&^*&^
		UPD_Addr_to_Mem <= X"001000";--"0000" & sm_ptr_out(19 downto 0);
		UPD_WR_Pulse 	<= '0';
		UPD_RD_Pulse 	<= '0';
		UPD_Bus_Ena		<= '1';  -- enabled except in the idle state
		
		Parameter_Changed 	<= '0'; -- used to indicate need for backup
		Update_Proc_Done 	<= '0';
		Update_Proc_State <= X"0";
		
		UPD_Permit_VME <= '0';  -- permit only in idle, init_sm_poll, check_vme and done
		ena_time_out 	<= '0';		
	--------------------------------------------------

	case CS is
		when idle =>					Update_Proc_State <= X"1";
			UPD_Permit_VME <= '1';
			UPD_Bus_Ena		<= '0'; -- enabled except in the idle state
			sm_ptr_nrst		<= '0';
			pr_ptr_nrst		<= '0';
			lim_ptr_nrst	<= '0';
			
			if   (do_sm_update = '1') then 	NS <= init_sm_poll;
			else                   				NS <= idle;
			end if;
		
		when init_sm_poll => 					NS <= check_vme; --NS <= address_flag;
			UPD_Permit_VME <= '1';
			Update_Proc_State <= X"1";
			UPD_Addr_to_Mem <= "0000" & sm_ptr_out(19 downto 0);
			sm_ptr_load		<= '1';
			pr_ptr_load		<= '1';
			lim_ptr_load 	<= '1';

-- added 7/22/16 -------------------------------------------			
		when check_vme =>
			UPD_Permit_VME <= '1';
			Update_Proc_State <= X"8";
			ena_time_out <= '1';
			if   (time_out > X"2710") then 	NS <= done; -- give up if VME is hung > 20us ######
			elsif(VME_Active = '1') then   	NS <= check_vme; -- wait if active
			else										NS <= address_flag;
			end if;
						
		------------------------------------------------------
-- start loop --
		when address_flag =>					
			Update_Proc_State <= X"1";
			UPD_Addr_to_Mem <= "0000" & sm_ptr_out(19 downto 0);
			if (sm_ptr_done = '1') then		NS <= done;
			else										NS <= setup_flag;
			end if;
		
		when setup_flag =>					NS <= get_sm_flag;
			Update_Proc_State <= X"1";
			UPD_Addr_to_Mem <= "0000" & sm_ptr_out(19 downto 0);
			
		when get_sm_flag =>					NS <= check_flag;
			Update_Proc_State <= X"1";
			UPD_Addr_to_Mem <= "0000" & sm_ptr_out(19 downto 0);
			get_flag_addr	<= '1'; -- store flag address to use later
			get_data1		<= '1';
		
		when check_flag =>					
			Update_Proc_State <= X"1";
			UPD_Addr_to_Mem <= "0000" & sm_ptr_out(19 downto 0);
			
			if((data_reg1_q > X"00000000") OR (Verify_All = '1')) then	
													NS <= addr_param_ptr;
				sm_ptr_inc 	<= '1'; -- step to setting
				sm_ptr_step	<= "0100";
				
			else								
													NS <= address_flag;
				pr_ptr_inc 	<= '1';
				pr_ptr_step	<= "0100";

				sm_ptr_inc 	<= '1';  	-- step 12 bytes to
				sm_ptr_step	<= "1100";	-- next flag.      
				
				lim_ptr_inc 	<= '1';     -- step 12 bytes to
				lim_ptr_step	<= "1100"; 	-- next set of limits.      
				                     
			end if;
-- end loop, return to start of loop if flag is not set --
						
		--======================================================
		-- The flag was set.  We now test the new setting to ensure
		-- it is within the limits.  If the value passes we will
		-- use the address pointer to the parameter register to go 
		-- and write the new value into this register.  We will then 
		-- read the value from the register and into the shared memory 
		-- reading variable for this parameter, and then reset the 
		-- update flag.  If the value was outside the limits we will 
		-- reset the flag without changing the parameter or the reading.
		
		---------------------------------------------------------
		-- latch the parameter's address 
		
		when addr_param_ptr =>				NS <= setup_param_ptr;
			Update_Proc_State <= X"2";
			UPD_Addr_to_Mem <= "0000" & pr_ptr_out(19 downto 0);
		
		when setup_param_ptr =>				NS <= get_param_addr;
			Update_Proc_State <= X"2";
			UPD_Addr_to_Mem <= "0000" & pr_ptr_out(19 downto 0);
		
		when get_param_addr =>				NS <= addr_setting;
			Update_Proc_State <= X"2";
			UPD_Addr_to_Mem <= "0000" & pr_ptr_out(19 downto 0);
			get_param_ptr   <= '1';
			
		---------------------------------------------------------
		-- latch setting into data register 1
		
		when addr_setting =>				NS <= setup_setting;
			Update_Proc_State <= X"2";
			UPD_Addr_to_Mem <= "0000" & sm_ptr_out(19 downto 0);
	
		when setup_setting =>				NS <= get_setting;
			Update_Proc_State <= X"2";
			UPD_Addr_to_Mem <= "0000" & sm_ptr_out(19 downto 0);

		when get_setting =>					NS <= addr_ul;
			Update_Proc_State <= X"2";
			UPD_Addr_to_Mem <= "0000" & sm_ptr_out(19 downto 0);
			get_data1		<= '1';
		
		---------------------------------------------------------
		-- get upper limit and compare to setting
	
		when addr_ul =>						NS <= setup_ul;
			Update_Proc_State <= X"2";
			UPD_Addr_to_Mem <= "0000" & lim_ptr_out(19 downto 0);
			
			-- CONVENIENT PLACE TO ADVANCE SM POINTER --
			sm_ptr_inc 		<= '1'; -- advance to reading
			sm_ptr_step		<= "0100";
			--========================================--
	
		when setup_ul =>					NS <= get_ul;
			Update_Proc_State <= X"2";
			UPD_Addr_to_Mem <= "0000" & lim_ptr_out(19 downto 0);

		when get_ul =>						NS <= check_ul;
			Update_Proc_State <= X"2";
			UPD_Addr_to_Mem <= "0000" & lim_ptr_out(19 downto 0);
			get_data2		<= '1';
		
		when check_ul =>					
			Update_Proc_State <= X"2";
			UPD_Addr_to_Mem <= "0000" & lim_ptr_out(19 downto 0);
			
			if  (data_reg1_q > data_reg2_q) then  -- exceeds limt
													NS <= chk_verify;
			else								  -- within limit
													NS <= next_lim;
			end if;
		
		when chk_verify =>
			Update_Proc_State <= X"2";
			lim_ptr_inc 	<= '1';    -- advances limit pointer
			lim_ptr_step	<= "1000"; -- to default value
			
			if (Verify_All = '1') then  	NS <= addr_default;
			else									NS <= addr_flag_again;
			-- If Verify_All is active then we will write the 
			-- default value to the parameter register, otherwise
			-- we will not change the register and move on to
			-- check the other flags.
			end if;
													  
		----------------------------------------------------------
		-- get lower limit and compare to setting
		
		when next_lim =>					NS <= addr_ll;
			Update_Proc_State <= X"3";
			UPD_Addr_to_Mem <= "0000" & lim_ptr_out(19 downto 0);
			lim_ptr_inc 	<= '1'; -- step to lower limit
			lim_ptr_step	<= "0100";
		
		when addr_ll =>						NS <= setup_ll;
			Update_Proc_State <= X"3";
			UPD_Addr_to_Mem <= "0000" & lim_ptr_out(19 downto 0);

		when setup_ll =>					NS <= get_ll;
			Update_Proc_State <= X"3";
			UPD_Addr_to_Mem <= "0000" & lim_ptr_out(19 downto 0);
		
		when get_ll =>						NS <= check_ll;
			Update_Proc_State <= X"3";
			UPD_Addr_to_Mem <= "0000" & lim_ptr_out(19 downto 0);
			get_data2		<= '1';
		
		when check_ll =>					
			Update_Proc_State <= X"3";
			UPD_Addr_to_Mem <= "0000" & lim_ptr_out(19 downto 0);
			lim_ptr_inc 	<= '1';    -- advances limit pointer
			lim_ptr_step	<= "0100"; -- to default value
			
			if  (data_reg1_q < data_reg2_q) then  -- exceeds limt
													NS <= chk_verify2;
			else								  -- within limit
													NS <= addr_param;
			end if;
			
		when chk_verify2 =>
			Update_Proc_State <= X"3";
			
			if (Verify_All = '1') then  	NS <= addr_default;
			else									NS <= addr_flag_again;
			-- If Verify_All is active then we will write the 
			-- default value to the parameter register, otherwise
			-- we will not change the register and move on to
			-- check the other flags.
			end if;
													  
		-----------------------------------------------------------	
		-- Load the default value into the parameter since the 
		-- value in the Setting was not in the limits during a 
		-- Verify_All test.
			
		when addr_default =>				NS <= setup_default;
			Update_Proc_State <= X"4";
			UPD_Addr_to_Mem <= "0000" & lim_ptr_out(19 downto 0);
		
		when setup_default =>				NS <= get_default;
			Update_Proc_State <= X"4";
			UPD_Addr_to_Mem <= "0000" & lim_ptr_out(19 downto 0);
			
		when get_default =>					NS <= addr_param;
			Update_Proc_State <= X"4";
			UPD_Addr_to_Mem <= "0000" & lim_ptr_out(19 downto 0);
			get_data1		<= '1';

		-----------------------------------------------------------
		-- Here we procede to write the parameter with either the
		-- Setting or the default value.
			
		when addr_param =>					NS <= write_param;
			Update_Proc_State <= X"5";
			UPD_Addr_to_Mem <= "0000" & param_addr(19 downto 0);
			UPD_Data_to_Mem <= data_reg1_q;
			
		when write_param => 				NS <= wait1;
			Update_Proc_State <= X"5";
			UPD_Addr_to_Mem <= "0000" & param_addr(19 downto 0);
			UPD_Data_to_Mem <= data_reg1_q;
			UPD_WR_Pulse <= '1'; -- << The write operation
									   		
		when wait1 =>
			Update_Proc_State <= X"8";
			UPD_Addr_to_Mem <= "0000" & param_addr(19 downto 0);
			UPD_Data_to_Mem <= data_reg1_q;
			
			if (param_addr < X"20000") then NS <= read_param;
			elsif(Transfer_Done = '1') then NS <= read_param;
			else							NS <= wait1;
			end if;
		
		-----------------------------------------------------------
		-- Here we are reading the parameter we just wrote back into
		-- Data Register 1.
			
		when read_param =>					NS <= wait2;	
			Update_Proc_State <= X"6";
			UPD_Addr_to_Mem <= "0000" & param_addr(19 downto 0);
			UPD_RD_Pulse <= '1'; -- << The read operation
			
		when wait2 =>
			Update_Proc_State <= X"8";
			UPD_Addr_to_Mem <= "0000" & param_addr(19 downto 0);
			
			if (param_addr < X"20000") then NS <= get_reading;
			elsif(Transfer_Done = '1') then NS <= get_reading;
			else							NS <= wait2;
			end if;
		
		when get_reading =>					NS <= upd_reading;
			Update_Proc_State <= X"7";
			UPD_Addr_to_Mem <= "0000" & param_addr(19 downto 0);
			get_data1		<= '1';

-- MOVED EARLIER IN THE PROCESS			
--			sm_ptr_inc 		<= '1'; -- advance to reading
--			sm_ptr_step		<= "0100";
	
		-------------------------------------------------------------
		-- Now we will write the parameter just written then read, into
		-- the parameter's "Reading".
			
		when upd_reading =>					NS <= addr_reading;
			Update_Proc_State <= X"7";
			UPD_Addr_to_Mem <= "0000" & sm_ptr_out(19 downto 0);
			UPD_Data_to_Mem <= data_reg1_q;
			
		when addr_reading =>				NS <= wr_reading;
			Update_Proc_State <= X"7";
			UPD_Addr_to_Mem <= "0000" & sm_ptr_out(19 downto 0);
			UPD_Data_to_Mem <= data_reg1_q;

		when wr_reading =>					NS <= addr_flag_again;
			Update_Proc_State <= X"7";
			UPD_Addr_to_Mem <= "0000" & sm_ptr_out(19 downto 0);
			UPD_Data_to_Mem <= data_reg1_q;
			UPD_WR_Pulse 	<= '1'; -- << The write to the reading
			
			-- Since Shared Memory has now been changed we will signal
			-- the need for backing up this new value in the Flash 
			-- memory.  There is no need to backup the Flash if this
			-- was a Verify_All (Bootup) operation.
			Parameter_Changed 	<= NOT Verify_All;
		-------------------------------------------------------
		-- We will now clear the "Flag" and continue to the next.
		
		when addr_flag_again =>				NS <= clear_flag;
			Update_Proc_State <= X"0";
			UPD_Addr_to_Mem <= "0000" & flag_addr(19 downto 0);
			UPD_Data_to_Mem <= X"00000000"; 
		
		when clear_flag =>					NS <= next_flag;
			Update_Proc_State <= X"0";
			UPD_Addr_to_Mem <= "0000" & flag_addr(19 downto 0);
			UPD_Data_to_Mem <= X"00000000"; 
			UPD_WR_Pulse 	<= '1';
		------------------------------------------------------	
		when next_flag =>					NS <= address_flag;
			Update_Proc_State <= X"0";
			UPD_Addr_to_Mem <= "0000" & flag_addr(19 downto 0);
			UPD_Data_to_Mem <= X"00000000"; 
			
			--increment to next flag -------------
			sm_ptr_inc 		<= '1'; 
			sm_ptr_step		<= "0100"; 
			
			pr_ptr_inc 		<= '1';
			pr_ptr_step		<= "0100";
			
			lim_ptr_inc 	<= '1';
			lim_ptr_step	<= "0100"; 
			
-- end loop if flag was set --	
				
		when done =>						NS <= idle;
			UPD_Permit_VME <= '1';
			Update_Proc_State <= X"0";
			Update_Proc_Done <= '1';
			UPD_Bus_Ena		<= '0'; -- enabled except in the idle and done state

		when others =>              		NS <= idle;
	end case;
end process;
	
end;