Library IEEE;
USE IEEE.std_logic_1164.all;
USE IEEE.std_logic_unsigned.all;

-- My Definitions --------------------------------
USE work.pkg_081006.all; 
USE work.DDS_VME_CONSTANTS.all;

-------------------------------------------------------------------------------
-- DDS_Update_Sequence
--
-- Periodic Update Sequence:
-- After each 36 ms Booster acceleration cycle there is 30 ms before the
-- next cycle in which we can rewind the frequency curve, update parameters, 
-- service curve buffer requests and updates, do curve interpolations from curve 
-- buffers to curve memory, and do backups of the Shared Memory to the Flash.
--
--		Frequency Curve Rewind: 
--		In order to avoid sparking in the accelerating cavities the LLRF frequency
--		is ramped back from 52.8MHz to 37MHz.  This process is triggered at the end
--		of the "Active" part of the cycle and proceeds on its own without need for
--		the data bus.  The Phase and Frequency curves are reinitialized just before
--		the next "Active" cycle.
--
--		Updating Parameters:  
--		The update process takes no more than 4 us to update as many as 6 parameters.
--
--		Curve Requests and Updates: 
--		The polling of semaphores and the transfer of one entire buffer takes 430 us.
--		Only one curve request will be processed in any "Non-Active" cycle interval.
--
--		Curve Interpolation: 
--		Interpolating a 1000 point curve buffer into a 10,000 value curve takes 2.5 ms.
--		Only one curve will be interpolated	in any "Non-Active" cycle interval.				 
--
-- **None of these processes take much time.  We could do all three in 3 ms.
--
--		Backing Up SM to Flash: 
--		Backups, or store, operations are done in two steps.  The first step unlocks 
--		the Flash block and initiates the erasure of the block we wish to store the 
--		Shared Memory into.  This sequence process is signal by the Backup process 
--		to indicate that the 1.2 second Flash erase process has begun (Backup_Setup_Done)
--		and that this process will no longer need the data bus until the Flash erase
--		process is complete.						
--		  
--		The unlocking and erase initiation takes 1.8 us.
--
--		Erasure of the block takes 800 ms.
--
--		Once the Flash erase process has completed, the Flash will begin the process 
--		of writing the 32 word flash programming buffer.  This buffer will need to be 
--		written 1000 or so times before the store operation is complete.  It takes 
--		7.6 us to write the buffer and then it takes 440 us for the buffer to be programmed 
--		into the flash memory block.  
--
--		When the Flash_Interface process wishes to control the UPD Databus for Reading or 
--		Writing the Shared Memory it will set the Flash_Bus_Request signal high and hold it
--		high until the buffer has been written and the Flash devices independently begins 
--		programming the buffer into the flash memory block.  
--
--		The Flash Interface will not proceed to take control of the bus until it receives 
--		a high on signal Flash_Bus_Permit.  The Flash Interface is expected to need
--		the bus for only 7.6 us.  It will request the bus again 440 us later.
--
--		The Flash_Bus_Permit will only be given during the non-active
--		portion of the 15Hz booster acceleration cycle, and then only
--		after the Shared Memory parameters, and curves have been 
--		checked for updates or requests and then serviced.  
--
--		The permit will not be given after getting 15 ms into the non-active interval.
--		Between this 15 ms mark and the beginning of the next Active interval, the
--		Flash will finish any buffer writes in progress and then the Phase and Frequency
--		curve reinitialization will be triggered.
--		
--		This approach to managing the storing of Shared Memory to the
--		Flash could take approximately 4 seconds.
--
--
-- 		Policy for Backing Up SM to Flash: ** MODIFIED -- SEE VERSION 2 COMMENT BELOW **
-- 		Backing up the Shared Memory to Flash will occur on an as needed basis and no more often
-- 		than the "2 minutes" it takes to complete a store operation.  A Backup_Update bit will be
-- 		set any time a parameter or curve is changed.  When found set it is cleared and a store 
-- 		to Flash operation will be initiated.  Once the store operation is completed the Backup_
-- 		Update bit will be tested again.
-- 		
-- 	Version 2:  Added a bit that comes from a VME writable control register that enables
--                backups only when set.

entity dds_update_seq_v2 is 
port( 
		clk						: in std_logic;
		nreset					: in std_logic;
		
		DDS_Ready 				: in std_logic; -- Indicates the Startup Process is completed
		
		Start_Update_Period	: in std_logic;  -- output of the Process Timer
		End_Update_Period		: in std_logic;  -- output of the Process Timer
		Start_Initialize		: in std_logic;  -- output of the Process Timer
		
		Update_Proc_Done		: in std_logic; -- Parameter update has completed
		Crv_Req_Proc_Done		: in std_logic; -- Curve buffer process has completed
		Crv_Intrp_Proc_Done 	: in std_logic; -- Curve Interpolation process has completed
		Backup_Setup_Done		: in std_logic; -- Flash has been Unlocked 
														 -- and Block Erase has begun
		Bootup_Backup_Done	: in std_logic; -- A Flash store/restore has completed
		
		Flash_Bus_Request		: in std_logic; -- Flash ready for a store operation
		
		Parameter_Changed		: in std_logic;  -- indicates need for backup
		Curve_Changed			: in std_logic;  -- indicates need for backup
		Backup_Enable			: in std_logic;  -- detect rising edge to enable a backup
		
		----------------------------------------------------------------------------------		
		
		Update_Trigger			: out std_logic; -- Trigger the Parameter Update Process
		Curve_Req_Trigger		: out std_logic; -- Trigger the Curve Request/Update Process
		Curve_Intrp_Trigger 	: out std_logic; -- Trigger the Curve Interpolation Process
		Backup_Trigger			: out std_logic; -- Trigger the Flash Store Process
		Flash_Bus_Permit		: out std_logic; -- Signal Flash to continue the Store Process
		Reset_Run_Curves		: out std_logic  -- Initializes Run Mode DDS Curves
		
	);
end dds_update_seq_v2;
	
architecture main of dds_update_seq_v2 is	
    
-- Backup Status Registers ------------------------------------------
	signal backup_ena_pulse	: std_logic;
	signal backup_enabled	: std_logic; 
	signal backup_needed		: std_logic; -- register
	signal backup_started	: std_logic; -- signal
	signal backup_in_prog	: std_logic; -- register

-- TIMER_STATE_MACHINE ----------------------------------------------
	type STATE_TYPE is
	(init, initialize_run, wait_for_start_update, check_update_params, 
	 check_crv_buffer, check_crv_intrp, check_backup_request, initiate_backup, 
	 check_bus_requests, permit_flash, wait_for_initialize );
 		
	signal CS, NS : STATE_TYPE;

--===================================================================
--			B E G I N
--===================================================================
begin	
-- Syncronize the Trigger input.
SYNC_START: led 
port map( sig_in  => Backup_Enable,	clk=>clk, clrn=>'1', 
			pls_out => backup_ena_pulse);
			
process(clk, nreset, backup_ena_pulse, backup_started)
begin
	if   (nreset = '0')    then				backup_enabled <= '0';
	elsif rising_edge(clk) then
		if   (backup_ena_pulse = '1') then	backup_enabled	<= '1';
		elsif(backup_started = '1') 	then	backup_enabled	<= '0';
		end if;
	end if;
end process;
			
--====================================================
BACKUP_STATUS_REGISTERS:
-- This process will set the backup_needed register if
-- a parameter or curve is changed.  It clears this register
-- when a backup is started.  It is possible that a 
-- parameter or curve will change during the backup.

-- We also set the backup in progress register when a
-- backup is started.  This is cleared when a bootup
-- or backup completes (Bootup_Backup_Done).
process(clk, nreset, Parameter_Changed, Curve_Changed, backup_started)
begin
	if   (nreset = '0')    then				backup_needed 	<= '0';
	elsif rising_edge(clk) then
		if   (Parameter_Changed = '1') OR 
			 (Curve_Changed = '1') 		then	backup_needed	<= '1';
		elsif(backup_started = '1') 	then	backup_needed	<= '0';
		end if;
	end if;
end process;

process(clk, nreset, backup_started, Bootup_Backup_Done)
begin
	if   (nreset = '0')    then					backup_in_prog	<= '0';
	elsif rising_edge(clk) then
		if   (backup_started = '1') 		then	backup_in_prog	<= '1';
		elsif(Bootup_Backup_Done = '1') 	then 	backup_in_prog	<= '0';
		end if;
	end if;
end process;

--====================================================
-- STATE MACHINE 
--**--   --**--   --**--   --**--   --**--   --**--   --**--  
SM_REG: process(clk, nreset, DDS_Ready)
begin
	if(nreset = '0') OR (DDS_Ready = '0') then CS <= init;
	elsif(rising_edge(clk))		then CS <= NS;
	end if;
end process;

SM_NEXT_STATE: 
process
(CS, Update_Proc_Done, Crv_Req_Proc_Done, Crv_Intrp_Proc_Done,
Backup_Setup_Done, Flash_Bus_Request, Start_Update_Period, DDS_Ready,
backup_needed, backup_in_prog, Start_Initialize, End_Update_Period,
backup_enabled )

begin
	-- Defaults -------------------------------------
	Reset_Run_Curves	<= '0';

	Update_Trigger		<= '0';
	Curve_Req_Trigger	<= '0';
	Curve_Intrp_Trigger <= '0';
	Backup_Trigger		<= '0';
	Flash_Bus_Permit	<= '0';
	
	backup_started 		<= '0';
	--------------------------------------------------

	case CS is
		when init =>						
			if   (DDS_Ready = '1') then		NS <= initialize_run;
			else										NS <= init;
			end if;
			
		-----------------------------------------------------------------------
		-- PREPARE FOR RUNNING DDS CURVES -------------------------------------
		when initialize_run =>						NS <= wait_for_start_update;
			-- Trigger Initialization of Run Mode DDS Curves
			Reset_Run_Curves 	<= '1';
			
		---------------------------------------------------------------------
		-- Beginning of Periodic Running of Update/Backup Processes
		--...................................................................
		when wait_for_start_update =>					
			if (Start_Update_Period = '1') then	NS <= check_update_params;
			else											NS <= wait_for_start_update;
			end if;
			
		when check_update_params =>
			Update_Trigger		<= '1';
			if (Update_Proc_Done = '1') then		NS <= check_crv_buffer;
			else											NS <= check_update_params;
			end if;
			
		when check_crv_buffer =>
			Curve_Req_Trigger	<= '1';
			if (Crv_Req_Proc_Done = '1') then	NS <= check_crv_intrp;
			else											NS <= check_crv_buffer;
			end if;
			
		when check_crv_intrp =>
			Curve_Intrp_Trigger	<= '1';
			if (Crv_Intrp_Proc_Done = '1') then	NS <= check_backup_request;
			else											NS <= check_crv_intrp;
			end if;
			
		when check_backup_request =>
			if   (backup_in_prog = '1') then		NS <= check_bus_requests;
			elsif(backup_needed  = '1') AND (backup_enabled = '1') then	
															NS <= initiate_backup;
			else											NS <= wait_for_initialize;
			end if;
		
		when initiate_backup =>					
			Backup_Trigger		<= '1'; -- Start Flash Backup process
			backup_started		<= '1'; -- signal to BACKUP_STATUS_REGISTERS
			
			Flash_Bus_Permit	<= '1'; -- Give the FLASH bus access and wait
			if (Backup_Setup_Done = '1') then	NS <= check_bus_requests;
			else											NS <= initiate_backup;		
			end if;
			
				
		when check_bus_requests =>
			-- If we have not reached the Initialization Interval yet, we
			-- will allow the FLASH backup process to write one FLASH buffers
			-- worth of data.  Writing this buffer takes only 8 us. Then the
			-- FLASH backup process will release the Flash_Bus_Request while
			-- it waits for the FLASH to finish programming the buffer into
			-- the FLASH memory.  Programming a buffer to memory can take 880 us
			-- or typically 440 us.  The backup process will proceed as it is
			-- permited.  Backup of the entire shared memory takes 1280, 32 word
			-- buffer writes.  This takes as much as 1.14 seconds.  Given 20 ms
			-- of time during each 15 Hz cycle, a backup could take 57 cycles.
			
			if   (End_Update_Period = '1') then NS <= wait_for_initialize; 
			elsif(Flash_Bus_Request = '1') then	NS <= permit_flash;
			else											NS <= check_bus_requests;
			end if;
		
		when permit_flash =>
			Flash_Bus_Permit	<= '1';
			if   (Flash_Bus_Request = '0') then	NS <= check_bus_requests;
			else											NS <= permit_flash;
			end if;
		
		-----------------------------------------------------------------	
		when wait_for_initialize =>				
			if   (Start_Initialize = '1')  then NS <= initialize_run;
			else											NS <= wait_for_initialize; 
			end if;
		
		when others =>              			NS <= init;
	end case;
end process;

end;