Library IEEE;
USE IEEE.std_logic_1164.all;
USE IEEE.std_logic_unsigned.all;

USE work.pkg_081006.all;
USE work.DDS_VME_CONSTANTS.all;

---------------------------------------------------------
-- This module performs an interpolation of input data
-- spaced 4us apart into output data spaced 1us apart.
-- It was written to work with 16 bit data.

entity interpolate_1us_16bit is
port(
		clk						: in std_logic;
		nreset					: in std_logic;
		Update_Pulse			: in std_logic;
		X_k1					: in std_logic_vector(15 downto 0);
		X_k2					: in std_logic_vector(15 downto 0);
		Init_Interp 			: in std_logic;
		Advance_Sample			: in std_logic;
		Enable_Updates			: in std_logic;

		Result					: out std_logic_vector(15 downto 0);
		Get_Next_Value			: out std_logic;
		carry_sum				: out std_logic
);
end interpolate_1us_16bit;

--########################################################
--				A R C H I T E C T U R E
--########################################################
architecture main of interpolate_1us_16bit is
-- Synchronized Inputs --------------------------------
	signal init			: std_logic;
	signal update_in	: std_logic;
	signal update		: std_logic;
	signal bypass		: std_logic;
	
-- Buffers ----------------------------------------------
	signal Y			: std_logic_vector(15 downto 0);
	signal Y_out		: std_logic_vector(15 downto 0);
	
	signal delay		: std_logic_vector(1 downto 0);
	signal get_next		: std_logic;


-- Altera Math Functions ------------------------------
component signed_adsub_16
	PORT
	(
		add_sub		: IN STD_LOGIC ;
		dataa		: IN STD_LOGIC_VECTOR (15 DOWNTO 0);
		datab		: IN STD_LOGIC_VECTOR (15 DOWNTO 0);
		cout		: OUT STD_LOGIC ;
		result		: OUT STD_LOGIC_VECTOR (15 DOWNTO 0)
	);
end component;
	signal c1			: std_logic;
	signal delta_X	: std_logic_vector(15 downto 0);
	
	signal c2			: std_logic;
	signal Y_k1_plus_step	: std_logic_vector(15 downto 0);
	           

--========================================================
--						B E G I N
--========================================================
begin

SYNC_INIT: led 
	port map(sig_in=>Init_Interp, clk=>clk,	clrn=>nreset, pls_out=>init); 
----
SYNC_UPDATE: led 
	port map(sig_in=>Update_Pulse, clk=>clk,	clrn=>nreset, pls_out=>update_in); 
----
	update <= (update_in AND Enable_Updates) OR init;
----
----------------------------------------------------------

------------------------------------------------------------
-- Interpolation from 4us points to 1us points
DELTA_FREQ: 
signed_adsub_16
port map(	add_sub => '0', -- subtract
			dataa	=> X_k2,
			datab	=> X_k1,
			cout	=> c1,
			result	=> delta_X );    -- delta X	
								
FREQ_INCREMENT:
signed_adsub_16    
port map(	add_sub => '1', -- add
			dataa	=> delta_X(15) & delta_X(15) & delta_X(15 downto 2),
			datab	=> Y_out,					--^-- delta X divided by 4
			cout	=> c2,
			result	=> Y );							
														
DELAYED_UPDATE_PULSE:
process(clk, nreset)
begin
	if(nreset='0')			then 
		get_next 	<= '0';
		delay     	<= (others => '0');
	elsif(rising_edge(clk))	then 
		delay(0)   	<= update;
		delay(1)   	<= delay(0);
		get_next 	<= delay(1);
		
	end if;
end process;

-- Y DATA REGISTER (accumulation register) ---------------
-- Note that we must initialize Y_out at the beginning and
-- we need to set Y_out to each given point, every 4us, 
-- because the step or increment will often have round-off
-- errors that would accumulate.


Y_DATA_REGISTER: 
process(clk, nreset, Init_Interp, Advance_Sample, X_k1)
begin
	if   (nreset = '0') then Y_out <= (others => '0');
	elsif rising_edge(clk) then
		if (get_next = '1') then
			if   (Init_Interp    = '1') then	Y_out <= X_k1;
			elsif(Advance_Sample = '1') then	Y_out <= X_k1;
			else								Y_out <= Y;
			end if;
		end if;
	end if;
end process;

-- OUTPUT --------------------------------
--~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Result 	   <= Y_out;

carry_sum  <= c1 OR c2;
Get_Next_Value	<= get_next;

end;             