Library IEEE;
USE IEEE.std_logic_1164.all;
USE IEEE.std_logic_unsigned.all;

USE work.pkg_081006.all;
USE work.DDS_VME_CONSTANTS.all;

---------------------------------------------------------
-- 

entity Interpolate_1us_32bit_v3 is
port(
		clk						: in std_logic;
		nreset					: in std_logic;
		Run_Curves				: in std_logic;
		Update_Pulse			: in std_logic;
		X_k0					: in std_logic_vector(31 downto 0);
		X_k1					: in std_logic_vector(31 downto 0);
		X_k2					: in std_logic_vector(31 downto 0);
		Reverse_Step			: in std_logic_vector(31 downto 0);
		Init_Interpolation 		: in std_logic;
		Advance_Sample			: in std_logic;
		Enable_Updates			: in std_logic;
		Rewind_Curve			: in std_logic;                         

		Result					: out std_logic_vector(31 downto 0);
		Rewind_Done				: out std_logic;
		Init_DDS					: out std_logic;
		Get_Next_Freq			: out std_logic; -- diagnostic output
		in_lt_out				: out std_logic; -- diagnostic output
		error						: out std_logic -- diagnostic output
);
end Interpolate_1us_32bit_v3;

--########################################################
--				A R C H I T E C T U R E
--########################################################
architecture main of Interpolate_1us_32bit_v3 is
-- Synchronized Inputs --------------------------------
	signal init			: std_logic;
	signal update_in	: std_logic;
	signal update		: std_logic;
	signal initialize	: std_logic;
	signal bypass		: std_logic;
	
-- Buffers ----------------------------------------------
	signal Y_in			: std_logic_vector(31 downto 0);
	signal Y_out		: std_logic_vector(31 downto 0);
	
	signal delay		: std_logic_vector(4 downto 0);
	signal get_next		: std_logic;

-- Altera Math Functions ------------------------------
component signed_sub32
	PORT
	(
		dataa		: IN STD_LOGIC_VECTOR (31 DOWNTO 0);
		datab		: IN STD_LOGIC_VECTOR (31 DOWNTO 0);
		overflow	: OUT STD_LOGIC ;
		result		: OUT STD_LOGIC_VECTOR (31 DOWNTO 0)
	);
end component;
	signal ovf1			: std_logic;
	signal delta	: std_logic_vector(31 downto 0);
	
component signed_add32
	PORT
	(
		dataa		: IN STD_LOGIC_VECTOR (31 DOWNTO 0);
		datab		: IN STD_LOGIC_VECTOR (31 DOWNTO 0);
		overflow	: OUT STD_LOGIC ;
		result		: OUT STD_LOGIC_VECTOR (31 DOWNTO 0)
	);
end component;
	signal ovf2			: std_logic;
	signal ovf3			: std_logic;
	signal Y_plus_step	: std_logic_vector(31 downto 0);
	signal rev_result	: std_logic_vector(31 downto 0);
	           

--========================================================
--						B E G I N
--========================================================
begin

SYNC_INIT: led 
	port map(sig_in=>Init_Interpolation, clk=>clk,	clrn=>nreset, pls_out=>init); 
----
SYNC_UPDATE: led 
	port map(sig_in=>Update_Pulse, clk=>clk,	clrn=>nreset, pls_out=>update_in); 
----
	update <= (update_in AND Enable_Updates) OR (update_in AND Rewind_Curve) OR init;
----
	initialize 	<= Init_Interpolation;
----------------------------------------------------------

------------------------------------------------------------
-- Interpolation from 4us points to 1us points
DELTA_FREQ: 
	signed_sub32 port map(	dataa	=>	X_k2,
							datab	=>	X_k1,
							overflow => ovf1,
							result	=> 	delta );
							
FREQ_INCREMENT:
	signed_add32 port map
		(	dataa		=>	delta(31) & delta(31) & delta(31 downto 2),
			datab		=>	Y_out,
			overflow	=>	ovf2,
			result	=>	Y_plus_step );

FREQ_DECREMENT: 
	signed_sub32 port map(	dataa		=>	Y_out,
									datab		=>	Reverse_Step,
									overflow => ovf3,
									result	=> 	rev_result );
														
DELAYED_UPDATE_PULSE:
process(clk, nreset, update, delay, initialize)
begin
	if(nreset='0')			then 
		get_next 	<= '0';
		Init_DDS 	<= '0';
		delay     	<= (others => '0');
	elsif(rising_edge(clk))	then 
		delay(0)   	<= update;
		delay(1)   	<= delay(0);
		get_next 	<= delay(1);

		delay(2)   	<= delay(1);
		delay(3)   	<= delay(2);
		delay(4)   	<= delay(3);
		Init_DDS	<= delay(4) AND initialize;
	end if;
end process;

-- Y DATA REGISTER (accumulation register) ---------------
Y_DATA_REGISTER: dffr
generic map(32)
port map( 	d    => Y_in, 
			clk  => clk, 
			clrn => nreset, 
			prn  => '1', 
			ena  => get_next, 
			q    => Y_out 
		);

SELECT_Y_IN:
process(Rewind_Curve, initialize, X_k1, Y_out, Y_plus_step, rev_result,
		Advance_Sample)
begin
	if   (Rewind_Curve = '1')  then -- test that result is positive
		if  (rev_result(31) = '0') then 	Y_in <= rev_result;
		else				  						Y_in <= Y_out;
		end if;
	elsif(initialize    = '1') then 	Y_in <= X_k1;
	elsif(Advance_Sample = '1') then	Y_in <= X_k1;
	else						    			Y_in <= Y_plus_step;
	end if;
end process;



-- OUTPUT --------------------------------
--~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Result 	   <= Y_out;

error	   <= ovf1 OR ovf2 OR ovf3;
Get_Next_Freq 	<= get_next;

IS_REWIND_DONE:
process(X_k0, rev_result, Y_out)
begin
	if   (Y_out <= X_k0 OR rev_result(31) = '1') then 	Rewind_Done <= '1';
	else					 							Rewind_Done <= '0';
	end if;
end process;

IS_YIN_LESS_THAN_YOUT:
process(Y_in, Y_out)
begin
	if   (Y_in < Y_out) then 	in_lt_out <= '1';
	else								in_lt_out <= '0';
	end if;
end process;



end;             